/*
 client.c : iauth client handling

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "iauth.h"
#include "irc.h"
#include "irc-limits.h"
#include "numeric-strings.h"

Pool client_pool;

static Client *clients;
static Timeout timeout;

static void client_parse_line(Client *client, char *line)
{
        const char *cmd, *args[COMMAND_MAX_ARGS+1];
        char buf[256];
        int argc;

	DEBUG((client, "input: %s", line));

	cmd = irc_parse_command(line, &argc, args);
	if (cmd == NULL)
		return;

	if (!module_client_read_line(client, cmd, argc, args)) {
		/* command not handled by any module */
		i_snprintf(buf, sizeof(buf), s_ERR_NOTREGISTERED,
			   client->info.nick);
                client_send_data(client, buf);
	}
}

static void client_read(Client *client)
{
	unsigned char *line;

	switch (tbuffer_read(client->inbuf)) {
	case -1:
                client_auth_abort(client, "disconnected");
		break;
	case -2:
                client_auth_abort(client, "input buffer full");
		break;
	default:
		while ((line = tbuffer_next_line(client->inbuf)) != NULL)
			client_parse_line(client, (char *) line);
                break;
	}
}

Client *client_create(int handle, IPADDR *ip, int port,
		      IAuthListener *listener)
{
	Client *client;

	client = p_new(client_pool, Client, 1);
        client->created = ioloop_time;
	strcpy(client->info.nick, "*");
	strcpy(client->info.user, "<unknown>");
	client->port = port;
	client->handle = handle;
	client->listener = listener;
        iauth_listener_ref(client->listener);

	memcpy(&client->ip, ip, sizeof(IPADDR));
	net_ip2host(&client->ip, client->info.ip);
	net_ip2host(&client->ip, client->info.host);

	client->inbuf = tbuffer_create(handle, client_pool,
				       IO_PRIORITY_DEFAULT, 1024);
	client->outbuf = tbuffer_create(handle, client_pool,
					IO_PRIORITY_DEFAULT, 256);
	client->io = io_add(handle, IO_READ, (IOFunc) client_read, client);

	if (modules_max_refnum >= 0) {
		client->module_data_size = modules_max_refnum+1;
		client->module_data = p_new(client_pool, void *,
					    client->module_data_size);
	}

	if (clients != NULL)
                clients->prev = client;
	client->next = clients;
	clients = client;

        return client;
}

static void client_remove(Client *client)
{
	if (client->closed)
		return;
	client->closed = TRUE;

	DEBUG((client, "closed"));

	if (client->prev == NULL)
		clients = client->next;
	else
		client->prev->next = client->next;

	if (client->next != NULL)
		client->next->prev = client->prev;
}

static void client_destroy(Client *client)
{
        client_remove(client);
	DEBUG((client, "destroyed"));

	iauth_listener_unref(client->listener);

        p_free(client_pool, client->module_data);
	tbuffer_destroy(client->inbuf);
	tbuffer_destroy(client->outbuf);

        if (client->io != NULL)
		io_remove(client->io);

	p_free(client_pool, client);
}

void client_auth_ref(Client *client)
{
        client->auth_refcount++;
}

void client_auth_unref(Client *client)
{
	if (--client->auth_refcount == 0) {
		if (!client->disconnected && !client->accepted)
			iauth_client_authenticated(client);

		client_destroy(client);
	}
}

void client_register_ref(Client *client)
{
	client_auth_ref(client);
        client->register_refcount++;
}

void client_register_unref(Client *client)
{
	if (--client->register_refcount == 0) {
                if (!client->disconnected)
			iauth_client_registered(client);
	}

        client_auth_unref(client);
}

void client_input_ref(Client *client)
{
        client->input_refcount++;
}

void client_input_unref(Client *client)
{
	if (client->input_refcount == 0 &&
	    !client->inbuf->closed) {
		client->extra_input_data =
			tbuffer_get_data(client->inbuf,
					 &client->extra_input_size);

		tbuffer_close(client->inbuf);
		tbuffer_reset(client->inbuf);

		io_remove(client->io);
		client->io = NULL;
	}
}

void client_auth_abort(Client *client, const char *msg)
{
	if (client->disconnected)
		return;

	client->disconnected = TRUE;

	iauth_client_rejected(client, msg);
	net_disconnect(client->handle);
}

void client_send_data(Client *client, const char *line)
{
	unsigned char *pos;
        int linelen;

        linelen = strlen(line);
	pos = tbuffer_get_space(client->outbuf, linelen+2);
	if (pos == NULL) {
		client_auth_abort(client, "output buffer full");
                return;
	}

	memcpy(pos, line, linelen);
        pos[linelen] = '\r';
        pos[linelen+1] = '\n';

	tbuffer_send_buffer(client->outbuf, linelen+2);
}

static int check_timeouts(void)
{
	Client *client, *next;
	time_t listen_timeout, connect_timeout;
        int connect;

	listen_timeout = ioloop_time - IAUTH_LOGIN_TIMEOUT;
	connect_timeout = ioloop_time - IAUTH_CONNECT_TIMEOUT;

	for (client = clients; client != NULL; client = next) {
		next = client->next;

                connect = client->info.connect;
		if ((connect && client->created < connect_timeout) ||
		    (!connect && client->created < listen_timeout)) {
			DEBUG((client, "timeout"));

			module_client_timeout(client);
                        client_remove(client);
		}
	}
        return TRUE;
}

void clients_init(void)
{
	client_pool = pool_create("Clients", 10240, TRUE);
        timeout = timeout_add(1000, (TimeoutFunc) check_timeouts, NULL);
}

void clients_deinit(void)
{
        timeout_remove(timeout);
	pool_unref(client_pool);
}
