/*
 dns.c : DNS lookup wrapper for adns

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "network.h"

#include "dns.h"
#include "../adns/internal.h"

#include <stdlib.h> /* we need to free() adns allocated query */

struct _DnsQuery {
	void (*func)();
	const void *data;

	adns_query query;
	unsigned int destroying:1;
	unsigned int reverse:1;
};

static adns_state dns_state;

static Pool pool;
static Timeout adns_to;
static IO poll_readio[MAXFD_POLL];
static IO poll_writeio[MAXFD_POLL];

static void dns_update_io(void);

DnsQuery *dns_lookup_ip(const char *host, DnsLookupFunc func, const void *data)
{
	DnsQuery *rec;

	rec = p_new(pool, DnsQuery, 1);
	rec->func = func;
	rec->data = data;

	if (adns_submit(dns_state, host, adns_r_a, 0, rec, &rec->query) != 0) {
                /* failed */
		p_free(pool, rec);
		return NULL;
	}

	dns_update_io();
        return rec;
}

DnsQuery *dns_lookup_name(IPADDR *ip, DnsReverseFunc func, const void *data)
{
	DnsQuery *rec;
        struct sockaddr_in sin;

	rec = p_new(pool, DnsQuery, 1);
	rec->func = func;
	rec->data = data;
        rec->reverse = TRUE;

	sin.sin_family = ip->family;
        memcpy(&sin.sin_addr.s_addr, &ip->ip, 4);

	if (adns_submit_reverse(dns_state, (struct sockaddr *) &sin,
				adns_r_ptr, 0, rec, &rec->query) != 0) {
                p_free(pool, rec);
		return NULL;
	}

	dns_update_io();
        return rec;
}

void dns_lookup_cancel(DnsQuery *query)
{
	if (!query->destroying) {
		adns_cancel(query->query);
		p_free(pool, query);
	}
}

/* Based on code by Aaron Sethman <androsyn@ratbox.org> */
static void dns_do_callbacks(void)
{
        DnsQuery *query;
	DnsLookupFunc lookup_func;
        DnsReverseFunc reverse_func;
	IPADDR ip;
	const char *host;

	adns_query q, r;
	adns_answer *answer;

	adns_forallqueries_begin(dns_state);
	while ((q = adns_forallqueries_next(dns_state, (void **)&r)) != NULL) {
		switch (adns_check(dns_state, &q, &answer, (void **) &query)) {
		case 0:
			/* found it */
                        query->destroying = TRUE;

			if (query->reverse) {
				host = answer->rrs.str == NULL ? NULL :
                                        *answer->rrs.str;
				reverse_func = query->func;
				reverse_func(query->data, host);
			} else {
				ip.family = AF_INET;
				memcpy(&ip.ip, &answer->rrs.inaddr->s_addr, 4);
				lookup_func = query->func;
                                lookup_func(query->data, &ip);
			}
			free(answer);
			p_free(pool, query);
			break;
		case EAGAIN:
			/* Go into the queue again */
			break;
		default:
			/* lookup failed - don't bother checking if it was
			   reverse or lookup or not, the parameters are same
			   anyway */
                        query->destroying = TRUE;

			lookup_func = query->func;
			lookup_func(query->data, NULL);
			p_free(pool, query);
			break;
		}
	}
}

static void dns_writeable(void *data, int fd)
{
	adns_processwriteable(dns_state, fd, &ioloop_timeval);
	dns_update_io();
}

static void dns_readable(void *data, int fd)
{
	adns_processreadable(dns_state, fd, &ioloop_timeval);
	dns_do_callbacks();
	dns_update_io();
}

static void dns_update_io(void)
{
	struct adns_pollfd pollfds[MAXFD_POLL];
	int npollfds, i, fd;

	adns__consistency(dns_state, 0, cc_entex);

	npollfds = adns__pollfds(dns_state, pollfds);
	for (i = 0; i < MAXFD_POLL; i++) {
		fd = pollfds[i].fd;

		if (i >= npollfds || (pollfds[i].events & ADNS_POLLIN) == 0) {
			if (poll_readio[i] != NULL) {
				io_remove(poll_readio[i]);
                                poll_readio[i] = NULL;
			}
		} else if (poll_readio[i] == NULL) {
			poll_readio[i] = io_add(fd, IO_READ,
						(IOFunc) dns_readable, NULL);
		}

		if (i >= npollfds || (pollfds[i].events & ADNS_POLLOUT) == 0) {
			if (poll_writeio[i] != NULL) {
				io_remove(poll_writeio[i]);
                                poll_writeio[i] = NULL;
			}
		} else if (poll_writeio[i] == NULL) {
			poll_writeio[i] = io_add(fd, IO_WRITE,
						(IOFunc) dns_writeable, NULL);
		}
	}
}

static void timeout_adns(void)
{
	adns_processtimeouts(dns_state, &ioloop_timeval);
}

int dns_lookups_init(void)
{
	pool = pool_create("DNS lookups", 1024, TRUE);

	memset(poll_readio, 0, sizeof(poll_readio));
	memset(poll_writeio, 0, sizeof(poll_writeio));

        if (adns_init(&dns_state, adns_if_noautosys|adns_if_nosigpipe, 0) != 0) {
		fprintf(stderr, "Error initializing dns_state\n");
                return FALSE;
	}

	adns_to = timeout_add(2000, (TimeoutFunc) timeout_adns, NULL);
        dns_update_io();
	return TRUE;
}

void dns_lookups_deinit(void)
{
	int i;

	for (i = 0; i < MAXFD_POLL; i++) {
                if (poll_readio[i] != NULL)
			io_remove(poll_readio[i]);
                if (poll_writeio[i] != NULL)
			io_remove(poll_writeio[i]);
	}
	timeout_remove(adns_to);

	adns_finish(dns_state);
        pool_unref(pool);
}
