/*
 m_dns.c : DNS module for iauth

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "dns.h"

static int init(void);
static void deinit(void);

static void client_connected(Client *client, void **data);
static void client_rejected(Client *client, void *data);
static void client_accepted(Client *client, void *data);

IAuthModule m_dns = {
	"dns",

	init,
	deinit,

	NULL, /* connection_accepted */
        NULL, /* client_registered */

	client_connected,
	client_rejected,
        client_accepted,

        NULL, /* client_timeout */
	NULL /* client_read_line */
};

typedef struct {
	Client *client;
	char host[IAUTH_HOST_LENGTH+1];
        DnsQuery *query;

	unsigned int destroyed:1;
} DnsModuleData;

static Pool pool;

static void dns_close(DnsModuleData *data)
{
	if (data->destroyed)
		return;
	data->destroyed = TRUE;

	if (data->query != NULL)
                dns_lookup_cancel(data->query);

	DEBUG((data->client, "DNS done"));
        client_register_unref(data->client);
}

static void dns_destroy(DnsModuleData *data)
{
        dns_close(data);
	p_free(pool, data);
}

static void dns_lookup_callback(const void *_data, IPADDR *ip)
{
	DnsModuleData *data = (void *) _data;

	if (ip != NULL) {
		if (net_ip_compare(ip, &data->client->ip)) {
			/* NOTE: max. host lengths are the same */
			strcpy(data->client->info.host, data->host);
		}
	}

	dns_close(data);
}

static void dns_reverse_callback(const void *_data, const char *host)
{
	DnsModuleData *data = (void *) _data;

	if (host == NULL) {
                dns_close(data);
                return;
	}

	if (strlen(host) >= sizeof(data->host)) {
                /* too long hostname */
		dns_close(data);
		return;
	}
        strcpy(data->host, host);

        /* and check the A record */
        data->query = dns_lookup_ip(data->host, dns_lookup_callback, data);
	if (data->query == NULL)
                dns_close(data);
}

static void client_connected(Client *client, void **data)
{
        DnsModuleData *dnsdata;

        dnsdata = p_new(pool, DnsModuleData, 1);
        dnsdata->client = client;
	dnsdata->query = dns_lookup_name(&client->ip, dns_reverse_callback,
					 dnsdata);

	if (dnsdata->query == NULL)
                p_free(pool, dnsdata);
	else {
		*data = dnsdata;
		client_register_ref(client);
	}
}

static void client_rejected(Client *client, void *data)
{
        dns_destroy(data);
}

static void client_accepted(Client *client, void *data)
{
        dns_destroy(data);
}

static int init(void)
{
        pool = pool_create("DNS", 1024, TRUE);
        return TRUE;
}

static void deinit(void)
{
        pool_unref(pool);
}
