/*
 iauth.c : iauth main()

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "lib-signals.h"
#include "network.h"
#include "fdpass.h"
#include "tbuffer.h"

#include "dns.h"
#include "modules.h"
#include "client.h"
#include "ircd-commands.h"

#include <unistd.h>

static const int data_handle = 3; /* used for passing fds */

Ircd *ircd;

static IOLoop ioloop;
static Pool pool;
static IAuthListener *listeners;
static unsigned int iauth_uid;

static void iauth_send(const char *format, ...)
{
        const char *data;
	va_list va;

	va_start(va, format);
        data = t_strdup_vprintf(format, va);
        va_end(va);

	write(1, data, strlen(data));
        write(1, "\n", 1);
}

/* send error message to ircd about client */
static void imsg_client_error(Client *client, const char *msg)
{
	fprintf(stderr, "%s@%s: %s\n", client->info.user,
		client->info.host, msg);
}

/* pass file handle to ircd */
static void iauth_pass_client(Client *client)
{
        client->info.id = iauth_uid++;
        client->info.extra_input_size = client->extra_input_size;

	if (fd_send(data_handle, client->handle,
		    &client->info, sizeof(client->info)) == -1) {
		imsg_client_error(client, t_strconcat("Error sending fd: ",
						      strerror(errno), NULL));
	} else if (client->extra_input_size > 0) {
                /* send also the extra input from client we didn't use */
		if (net_transmit(data_handle, client->extra_input_data,
				 client->extra_input_size) <= 0) {
			imsg_client_error(client,
					  t_strconcat("Error sending data: ",
						      strerror(errno), NULL));
		}
	}
}

void debug_log(Client *client, const char *format, ...)
{
	va_list va;

	va_start(va, format);
	if (client != NULL) {
		fprintf(stderr, "%s!%s@%s: ", client->info.nick,
			client->info.user, client->info.host);
	}
	vfprintf(stderr, format, va);
        fputc('\n', stderr);
        va_end(va);
}

void iauth_client_registered(Client *client)
{
	if (client->registered)
		return;
        client->registered = TRUE;

	DEBUG((client, "client registered"));

	module_client_registered(client);
}

void iauth_client_authenticated(Client *client)
{
	if (client->disconnected || client->accepted)
		return;

	DEBUG((client, "client authenticated"));

	module_client_accepted(client);

	DEBUG((client, "accepted"));

	iauth_pass_client(client);
	net_disconnect(client->handle);

	client->accepted = TRUE;

	DEBUG((client, "done"));
}

void iauth_client_rejected(Client *client, const char *msg)
{
	char str[512];
        int len;

	if (client->rejected)
                return;
	client->rejected = TRUE;

	if (client->accepted) {
		/* client was already sent to ircd, we'll need to kill it now */
		DEBUG((client, "rejected already accepted client"));
		iauth_send("K %u :%s", client->info.id, msg);
		return;
	}

	DEBUG((client, "client rejected"));

	if (msg != NULL && *msg != '\0') {
		len = i_snprintf(str, sizeof(str), "ERROR :%s\r\n", msg);
		net_transmit(client->handle, str, len);
	}

	if (client->info.connect) {
		fprintf(stderr, "Connection to %s:%d failed: %s\n",
			client->info.host, client->info.connect_port,
			msg == NULL ? "" : msg);
	} else {
		iauth_send("R %c %c %s@%s :%s", IAUTH_REPLY_REJECT,
			   client->banned ? 'K' : 'R',
			   client->info.user, client->info.host,
			   msg == NULL ? "" : msg);
	}

        module_client_rejected(client);
}

static void io_connect(IAuthListener *listener)
{
        Client *client;
	IPADDR ip;
        int handle, port;

	handle = net_accept(listener->handle, &ip, &port);
	if (handle == -1)
		return;

        client = client_create(handle, &ip, port, listener);

	DEBUG((client, "connected"));

	module_connection_accepted(client);
        if (!client->disconnected)
		module_client_connected(client);
}

void iauth_quit(void)
{
        io_loop_stop(ioloop);
}

void iauth_set_ircd_info(Ircd *ircd, const char *server_name,
			 const char *version, const char *flags,
			 const char *sid, const char *info)
{
	p_strdup_replace(ircd->pool, &ircd->name, server_name);
	p_strdup_replace(ircd->pool, &ircd->version, version);
	p_strdup_replace(ircd->pool, &ircd->flags, flags);
	p_strdup_replace(ircd->pool, &ircd->sid, sid);
	p_strdup_replace(ircd->pool, &ircd->info, info);
}

IAuthListener *iauth_create_listener(void)
{
	IAuthListener *listener;

	listener = p_new(pool, IAuthListener, 1);
	listener->refcount = 1;

        return listener;
}

void iauth_listener_ref(IAuthListener *listener)
{
        listener->refcount++;
}

void iauth_listener_unref(IAuthListener *listener)
{
        if (--listener->refcount == 0)
		p_free(pool, listener);
}

void iauth_add_listener(int port, IPADDR *ip)
{
	IAuthListener *listener;
	char ip_name[MAX_IP_LEN];
	int handle;

        /* listen for incoming IRC connections */
	handle = net_listen(ip, &port);
	if (handle == -1) {
		if (ip == NULL)
			strcpy(ip_name, "<any>");
		else
			net_ip2host(ip, ip_name);

		fprintf(stderr, "can't listen in port %d (%s): %s\n",
			port, strerror(errno), ip_name);
                return;
	}

	listener = iauth_create_listener();
	listener->port = port;
	if (ip != NULL)
                memcpy(&listener->ip, ip, sizeof(IPADDR));
	listener->handle = handle;
	listener->io = io_add(listener->handle, IO_READ,
			      (IOFunc) io_connect, listener);

	listener->next = listeners;
        listeners = listener;
}

static void iauth_listener_close(IAuthListener *listener)
{
	IAuthListener **tmp;

	for (tmp = &listeners; *tmp != NULL; tmp = &(*tmp)->next) {
		if (*tmp == listener) {
			*tmp = listener->next;
                        break;
		}
	}

	if (listener->handle != -1) {
		close(listener->handle);
                listener->handle = -1;
	}

	if (listener->io != NULL) {
		io_remove(listener->io);
                listener->io = NULL;
	}

	iauth_listener_unref(listener);
}

void iauth_remove_listener(int port, IPADDR *ip)
{
	IAuthListener *l, *next;

	for (l = listeners; l != NULL; l = next) {
		next = l->next;

		if ((port <= 0 || l->port == port) &&
		    (ip == NULL || net_ip_compare(&l->ip, ip)))
			iauth_listener_close(l);
	}
}

static void io_ircd_input(Ircd *ircd)
{
	char *line;

	switch (tbuffer_read(ircd->inbuf)) {
	case -1:
		fprintf(stderr, "ircd died? %s\n", strerror(errno));
		exit(1);
                return;
	case -2:
                /* shouldn't happen.. */
                i_panic("Input buffer full");
                return;
	}

	while ((line = (char *) tbuffer_next_line(ircd->inbuf)) != NULL)
                ircd_command_parse(ircd, line);
}

static void sig_quit(int signo)
{
        io_loop_stop(ioloop);
}

static int iauth_init(void)
{
	lib_init_signals(sig_quit);

	listeners = NULL;
        iauth_uid = 0;

	if (!dns_lookups_init())
                return FALSE;

	pool = pool_create("iauth", 4096, TRUE);
	modules_init();
        clients_init();
        ircd_commands_init();

	/* listen for input from ircd */
        ircd = p_new(pool, Ircd, 1);
        ircd->pool = pool_create("ircd connection", 4096, TRUE);
	ircd->inbuf = tbuffer_create_file(0, ircd->pool, 0);
	ircd->io = io_add(0, IO_READ, (IOFunc) io_ircd_input, ircd);

        return TRUE;
}

static void ircd_destroy(Ircd *ircd)
{
	tbuffer_destroy(ircd->inbuf);
        io_remove(ircd->io);

        p_free_clean(ircd->pool, ircd->name);
        p_free_clean(ircd->pool, ircd->version);
        p_free_clean(ircd->pool, ircd->flags);
        p_free_clean(ircd->pool, ircd->sid);
        p_free_clean(ircd->pool, ircd->info);
	p_free(pool, ircd);

	pool_unref(ircd->pool);
}

static void iauth_deinit(void)
{
        if (lib_signal_kill != 0)
		i_warning("Killed with signal %d\n", lib_signal_kill);

	iauth_remove_listener(-1, NULL);

        ircd_destroy(ircd);

        ircd_commands_deinit();
        clients_deinit();
	modules_deinit();
	dns_lookups_deinit();

        pool_unref(pool);
}

int main(int argc, char *argv[])
{
	int ret;

        lib_init();
	ioloop = io_loop_create();

	if (!iauth_init())
		ret = 1;
	else {
		io_loop_run(ioloop);
		iauth_deinit();
                ret = 0;
	}

	io_loop_destroy(ioloop);
        lib_deinit();

        return ret;
}
