/*
 m_ident.c : ident module for iauth

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"

/*
   ^  I line with OTHER type ident
   ~  I line, no ident
   +  i line with ident
   =  i line with OTHER type ident
   -  i line, no ident
*/

#define IDENT_USER_MAX_LENGTH 10

static int init(void);
static void deinit(void);

static void client_connected(Client *client, void **data);
static void client_rejected(Client *client, void *data);
static void client_accepted(Client *client, void *data);

IAuthModule m_ident = {
	"ident",

	init,
	deinit,

	NULL, /* connection_accepted */
	NULL, /* client_registered */

	client_connected,
	client_rejected,
        client_accepted,

        NULL, /* client_timeout */
	NULL /* client_read_line */
};

typedef struct {
        Client *client;

	int handle;
	IO io;
	TBuffer *inbuf;

	char user[IDENT_USER_MAX_LENGTH+1];
	unsigned int other:1;
	unsigned int closed:1;
} IdentModuleData;

static Pool pool;

static void ident_close(IdentModuleData *data)
{
	if (data->closed)
		return;

        data->closed = TRUE;
	net_disconnect(data->handle);
	io_remove(data->io);
	tbuffer_destroy(data->inbuf);

	DEBUG((data->client, "ident done"));
	client_register_unref(data->client);
}

static char *ident_strtok(char **line)
{
	char *ret;

	if (*line == NULL || **line == '\0')
                return NULL;

	while (**line == ' ')
		(*line)++;
	ret = *line;

	while (**line != ' ' || (*line)[1] != ':') {
		if (**line == '\0')
                        break;
		(*line)++;
	}

	if (**line != '\0') {
		**line = '\0';
                *line += 2;
	}
        return ret;
}

static void ident_parse_line(IdentModuleData *data, char *line)
{
	char *ports, *resp_type, *os, *user;
        char *local_port, *p;

	ports = ident_strtok(&line);
	resp_type = ident_strtok(&line);
	os = ident_strtok(&line);
	user = ident_strtok(&line);

	/* we want to handle only USERID replies */
	if (user == NULL || strcasecmp(resp_type, "USERID") != 0)
		return;

        /* split remote/local ports */
	p = local_port = strchr(ports, ',');
	if (local_port == NULL)
		return;

	while (p[-1] == ' ' && p > ports) p--;
        *p = '\0';

	local_port++;
	while (*local_port == ' ') local_port++;

        /* make sure the local/remote ports match */
	if (atoi(local_port) != data->client->listener->port ||
	    atoi(ports) != data->client->port)
                return;

        /* don't allow spaces in user */
	p = strchr(user, ' ');
        if (p != NULL) *p = '\0';

	if (strcasecmp(os, "OTHER") == 0) {
		/* OTHER type ident, skip given user */
		data->other = TRUE;
	} else {
		/* save the reply */
		strncpy(data->user, user, sizeof(data->user)-1);
	}
}

static void io_read(IdentModuleData *data)
{
	unsigned char *line;

	if (data->inbuf == NULL) {
		data->inbuf = tbuffer_create(data->handle, pool,
					     IO_PRIORITY_DEFAULT, 128);
	}

	if (tbuffer_read(data->inbuf) < 0)
		ident_close(data);
	else {
		line = tbuffer_next_line(data->inbuf);
		if (line != NULL) {
			ident_parse_line(data, (char *) line);
                        ident_close(data);
		}
	}
}

static void io_connected(IdentModuleData *data)
{
        char buf[128];
	int ret, len;

        /* switch to read-only IO waiting */
	io_remove(data->io);
	data->io = io_add(data->handle, IO_READ, (IOFunc) io_read, data);

	/* send the query */
	len = i_snprintf(buf, sizeof(buf), "%d, %d\r\n",
			 data->client->port,
			 data->client->listener->port);
	ret = net_transmit(data->handle, buf, len);
	if (ret != len)
                ident_close(data);
}

static void client_connected(Client *client, void **data)
{
	IdentModuleData *ident_data;
        IPADDR *ip;
	int handle;

	ip = client->listener->ip.family == 0 ? NULL : &client->listener->ip;
	handle = net_connect_ip(&client->ip, 113, ip);
	if (handle == -1)
                return;

	client_register_ref(client);

	*data = ident_data = p_new(pool, IdentModuleData, 1);
        ident_data->client = client;
        ident_data->handle = handle;
	ident_data->io = io_add(ident_data->handle, IO_READ|IO_WRITE,
				(IOFunc) io_connected, ident_data);
}

static void ident_data_destroy(IdentModuleData *data)
{
        ident_close(data);
        p_free(pool, data);
}

static void client_rejected(Client *client, void *data)
{
        ident_data_destroy(data);
}

static void client_accepted(Client *client, void *data)
{
	IdentModuleData *ident_data = data;
	int restricted;

	/* change the user value */
	restricted = client->info.restricted;
	if (ident_data->user[0] == '\0') {
		/* no ident */
		memmove(client->info.user+1, client->info.user,
			sizeof(client->info.user)-2);
		client->info.user[0] = restricted ? '-' : '~';
	} else if (ident_data->other) {
		/* OTHER ident */
		memmove(client->info.user+1, client->info.user,
			sizeof(client->info.user)-2);
		client->info.user[0] = restricted ? '=' : '^';
	} else if (restricted) {
                /* normal ident, restricted */
		client->info.user[0] = '+';
		strncpy(client->info.user+1, ident_data->user,
			sizeof(client->info.user)-2);
	} else {
		/* normal ident, non-restricted */
		strncpy(client->info.user, ident_data->user,
			sizeof(client->info.user)-1);
	}

	ident_data_destroy(data);
}

static int init(void)
{
        pool = pool_create("Ident", 1024, TRUE);
        return TRUE;
}

static void deinit(void)
{
        pool_unref(pool);
}
