/*
 ircd-commands.c : parser for commands sent to iauth from ircd

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "network.h"
#include "dns.h"
#include "ircd-commands.h"

#include "irc.h"
#include "irc-limits.h"

#include <unistd.h>

typedef struct {
        Ircd *ircd;

	char *host, *pass;
	int port;

        IPADDR ip;
	int handle;

	IO io;
} ConnectData;

static Pool pool;

static void connect_destroy(ConnectData *data)
{
	if (data->handle != -1)
                close(data->handle);
	if (data->io != NULL)
		io_remove(data->io);
        p_free(pool, data->host);
        p_free(pool, data->pass);
        p_free(pool, data);
}

static void io_server_connected(ConnectData *data)
{
	IAuthListener *listener;
        Client *client;
	char line[513];

	io_remove(data->io);
	data->io = NULL;

        /* create a fake listener for us */
	listener = iauth_create_listener();
	net_getsockname(data->handle, &listener->ip, &listener->port);

        /* create the client */
	client = client_create(data->handle, &data->ip, data->port, listener);
	client->info.connect = TRUE;
        client->info.connect_port = data->port;

	DEBUG((client, "connected to port %d", data->port));

        /* send PASS */
	i_snprintf(line, sizeof(line), "PASS %s %s IRC|%s P\r\n", data->pass,
		   data->ircd->version, data->ircd->flags);
	tbuffer_send(client->outbuf, line, strlen(line));

        /* send SERVER */
	i_snprintf(line, sizeof(line), "SERVER %s 1 %s :%s\r\n",
		   data->ircd->name, data->ircd->sid, data->ircd->info);
	tbuffer_send(client->outbuf, line, strlen(line));

	data->handle = -1;
	connect_destroy(data);

        /* start calling the hooks */
	module_connection_accepted(client);
	if (!client->disconnected)
		module_client_connected(client);
}

static int server_connect(ConnectData *data)
{
	data->handle = net_connect_ip(&data->ip, data->port, NULL);
	if (data->handle == -1) {
		DEBUG((NULL, "Connect failed"));
                connect_destroy(data);
		return FALSE;
	}

	DEBUG((NULL, "handle = %d", data->handle));
	data->io = io_add(data->handle, IO_READ|IO_WRITE,
			  (IOFunc) io_server_connected, data);
        io_server_connected(data);
	return TRUE;
}

static void dns_connect_callback(const void *_data, IPADDR *ip)
{
	ConnectData *data = (void *) _data;
	char host[100];

        net_ip2host(ip, host);
	DEBUG((NULL, "host = %s", host));

	if (ip == NULL) {
		DEBUG((NULL, "DNS lookup failed: host not found"));
		connect_destroy(data);
		return;
	}

        memcpy(&data->ip, ip, sizeof(IPADDR));
	server_connect(data);
}

static void cmd_connect_server(Ircd *ircd, const char *host, int port)
{
        ConnectData *data;

	data = p_new(pool, ConnectData, 1);
        data->ircd = ircd;
	data->host = p_strdup(pool, host);
	data->pass = p_strdup(pool, "pass"); /* FIXME */
	data->port = port;
        data->handle = -1;

	DEBUG((NULL, "Creating connection to %s port %d", host, port));

	if (net_host2ip(data->host, &data->ip) == 0) {
		/* no need to do DNS lookup */
                server_connect(data);
                return;
	}

	if (!dns_lookup_ip(data->host, dns_connect_callback, data)) {
		DEBUG((NULL, "DNS lookup failed"));
		connect_destroy(data);
		return;
	}
}

void ircd_command_parse(Ircd *ircd, char *line)
{
        IPADDR ip, *ipptr;
	const char *cmd, *args[COMMAND_MAX_ARGS+1];
	char *const *modules;
        int argc, port;

	cmd = irc_parse_command(line, &argc, args);
	if (cmd == NULL)
		return;

	switch (*cmd) {
	case IAUTH_CMD_DIE:
                iauth_quit();
		break;

	case IAUTH_CMD_MODULE_ADD:
		/* <module>,[...] */
		if (argc < 1)
			break;

		modules = t_strsplit(args[0], ",");
		while (*modules != NULL) {
                        module_register_name(*modules);
                        modules++;
		}
		break;
	case IAUTH_CMD_MODULE_REMOVE:
                /* <module>,[...] */
		if (argc < 1)
			break;

		modules = t_strsplit(args[0], ",");
		while (*modules != NULL) {
                        module_unregister(*modules);
                        modules++;
		}
		break;

	case IAUTH_CMD_LISTEN_ADD:
	case IAUTH_CMD_LISTEN_REMOVE:
                /* <port> [<ip>] */
		if (argc < 1)
			break;

		if (argc < 2)
			ipptr = NULL;
		else {
			net_host2ip(args[1], &ip);
                        ipptr = &ip;
		}
		port = atoi(args[0]);

                if (*cmd == IAUTH_CMD_LISTEN_ADD)
			iauth_add_listener(port, ipptr);
		else
			iauth_remove_listener(port, ipptr);
		break;

	case IAUTH_CMD_CONNECT_SERVER:
		/* <host> <port> */
		if (argc < 2)
                        break;

                cmd_connect_server(ircd, args[0], atoi(args[1]));
		break;

	case IAUTH_CMD_MYINFO:
		/* <server name> <version> <flags> <sid> :<info> */
		if (argc < 4)
			break;

		iauth_set_ircd_info(ircd, args[0], args[1], args[2],
				    args[3], args[4]);
                break;
	}
}

void ircd_commands_init(void)
{
        pool = pool_create("ircd commands", 512, TRUE);
}

void ircd_commands_deinit(void)
{
        pool_unref(pool);
}
