/*
   m_login.c : login parser module for iauth, for both client and
               server connections

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "numeric-strings.h"

#define CLIENT_IS_REGISTERED(client) \
	(((client)->nick_given && (client)->user_given) || \
	 ((client)->info.version[0] != '\0' && \
	  (client)->info.servername[0] != '\0'))

static void client_connected(Client *client, void **data);
static void client_rejected(Client *client, void *data);
static void client_timeout(Client *client, void *data);
static int client_read_line(Client *client, const char *cmd,
			    int argc, const char *args[], void *data);

IAuthModule m_login = {
	"login",

	NULL, /* init */
	NULL, /* deinit */

	NULL, /* connection_accepted */
	NULL, /* client_registered */

	client_connected,
	client_rejected,
        NULL, /* client_accepted */

        client_timeout,
	client_read_line
};

static void client_connected(Client *client, void **data)
{
        *data = (void *) 1;
	client_register_ref(client);
        client_input_ref(client);
}

static void client_rejected(Client *client, void *data)
{
	if (!client->registered) {
                client_input_unref(client);
		client_register_unref(client);
	}
}

#define my_strcpy(dest, src) \
        strncpy(dest, src, sizeof(dest)-1)

static int client_read_line(Client *client, const char *cmd,
			    int argc, const char *args[], void *data)
{
        IAuthClientInfo *info;
	char buf[256];
	int moreparams;

        info = &client->info;

	/* handle only PASS, NICK, USER, SERVER, QUIT and ERROR commands */
        moreparams = FALSE;
	if (strcmp(cmd, "PASS") == 0) {
                /* PASS <pass> [<version> <flags> [<options>]] */
		if (argc == 0 || (argc > 1 && argc < 3))
                        moreparams = TRUE;
		else {
			my_strcpy(info->pass, args[0]);
                        if (argc > 1) my_strcpy(info->version, args[1]);
                        if (argc > 2) my_strcpy(info->flags, args[2]);
			if (argc > 3) my_strcpy(info->options, args[3]);
		}
	} else if (strcmp(cmd, "NICK") == 0) {
                /* NICK <nick> */
		if (argc < 1)
                        moreparams = TRUE;
		else {
			my_strcpy(info->nick, args[0]);
                        client->nick_given = TRUE;
		}
	} else if (strcmp(cmd, "USER") == 0) {
		/* USER <user> <ignore> <ignore> [?] <real name> */
		if (argc < 4)
                        moreparams = TRUE;
		else {
                        my_strcpy(info->user, args[0]);
			my_strcpy(info->realname, args[argc-1]);
                        client->user_given = TRUE;
		}
	} else if (strcmp(cmd, "SERVER") == 0) {
		/* SERVER <name> <hopcount> [<sid/token> [?]] <info> */
		if (argc < 3)
                        moreparams = TRUE;
		else {
			my_strcpy(info->servername, args[0]);
			my_strcpy(info->sid, argc == 3 ? "1" : args[2]);
			my_strcpy(info->serverinfo, args[argc-1]);
                        info->hopcount = atoi(args[1]);
		}
	} else if (strcmp(cmd, "QUIT") == 0) {
		/* QUIT [<message>] */
		client_auth_abort(client, args[0]);
	} else if (strcmp(cmd, "ERROR") == 0) {
                /* ERROR [<message>] */
		client_auth_abort(client, args[0]);
	} else {
                return FALSE;
	}

	if (moreparams) {
                /* command needs more parameters */
		i_snprintf(buf, sizeof(buf), s_ERR_NEEDMOREPARAMS,
			   info->nick, cmd);
		client_send_data(client, buf);
	}

	if (CLIENT_IS_REGISTERED(client)) {
		/* we're done */
		DEBUG((client, "logged in"));
                client_input_unref(client);
		client_register_unref(client);
	}
        return TRUE;
}

static void client_timeout(Client *client, void *data)
{
	if (!CLIENT_IS_REGISTERED(client))
                client_auth_abort(client, "timeout");
}
