/*
 modules.c : iauth modules handling

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "modules.h"

#define CLIENT_MODULE_DATA(client, module) \
	((client)->module_data_size < (module)->refnum ? NULL : \
         (void *) &(client)->module_data[(module)->refnum])

typedef struct {
	const char *name;
        IAuthModule *module;
} BuiltinModule;

extern IAuthModule m_login, m_dns, m_ident, m_ping, m_proxy;
static BuiltinModule builtin_modules[] = {
	{ "login",	&m_login },
	{ "dns",	&m_dns },
	{ "ident",	&m_ident },
	{ "ping",	&m_ping },
	{ "proxy",	&m_proxy },
	{ NULL, NULL }
};

int modules_max_refnum;
static IAuthModule *modules;

static int get_module_refnum(void)
{
        IAuthModule *module;
	int refnum;

	for (refnum = 0; ; refnum++) {
		for (module = modules; module != NULL; module = module->next) {
			if (module->refnum == refnum)
                                break;
		}

		if (module == NULL)
                        return refnum;
	}
}

static void update_highest_refnum(void)
{
        IAuthModule *module;

	modules_max_refnum = -1;

	for (module = modules; module != NULL; module = module->next) {
		if (module->refnum > modules_max_refnum)
                        modules_max_refnum = module->refnum;
	}
}

static void module_ref(IAuthModule *module)
{
	module->refcount++;
}

static void module_unref(IAuthModule *module)
{
	IAuthModule **tmp;

	if (--module->refcount > 0)
		return;

	/* update linked list */
	for (tmp = &modules; *tmp != NULL; tmp = &(*tmp)->next) {
		if (*tmp == module) {
			*tmp = module->next;
			break;
		}
	}

        if (module->deinit != NULL)
		module->deinit();
        update_highest_refnum();
}

static int module_is_registered(IAuthModule *module)
{
	IAuthModule *tmp;

	for (tmp = modules; tmp != NULL; tmp = tmp->next) {
		if (tmp == module)
			return TRUE;
	}

        return FALSE;
}

void module_register(IAuthModule *module)
{
	if (module_is_registered(module))
		return;

        module->refnum = get_module_refnum();
	module->refcount = 1;
	module->unregistered = FALSE;

	if (module->init != NULL && !module->init()) {
		fprintf(stderr, "module_register(): init() of %s failed\n",
			module->name);
                return;
	}

	if (module->refnum > modules_max_refnum)
                modules_max_refnum = module->refnum;

	module->next = modules;
        modules = module;
}

void module_register_name(const char *name)
{
        BuiltinModule *module;

	for (module = builtin_modules; module->name != NULL; module++) {
		if (strcasecmp(module->name, name) == 0) {
                        module_register(module->module);
			break;
		}
	}
}

void module_unregister(const char *name)
{
	IAuthModule *module;

	for (module = modules; module != NULL; module = module->next) {
		if (strcasecmp(module->name, name) == 0) {
			/* deinit module */
			module->unregistered = TRUE;
			module_unref(module);
			break;
		}
	}
}

void module_connection_accepted(Client *client)
{
	IAuthModule *module;

	for (module = modules; module != NULL; module = module->next) {
		if (!module->unregistered &&
		    module->connection_accepted != NULL)
			module->connection_accepted(client);
	}
}

void module_client_connected(Client *client)
{
	IAuthModule *module;
	void **data;

	client_register_ref(client);
	client_input_ref(client);

	for (module = modules; module != NULL; module = module->next) {
		if (!module->unregistered &&
		    module->client_connected != NULL) {
			data = CLIENT_MODULE_DATA(client, module);
			if (data != NULL) {
				module->client_connected(client, data);
				if (*data != NULL) module_ref(module);
			}
		}
	}

	client_input_unref(client);
	client_register_unref(client);
}

void module_client_rejected(Client *client)
{
	IAuthModule *module;
	void **data;

	for (module = modules; module != NULL; module = module->next) {
		data = CLIENT_MODULE_DATA(client, module);
		if (data != NULL && *data != NULL) {
			if (module->client_rejected != NULL)
				module->client_rejected(client, *data);
			module_unref(module);
			*data = NULL;
		}
	}
}

void module_client_registered(Client *client)
{
	IAuthModule *module;

	for (module = modules; module != NULL; module = module->next) {
		if (!module->unregistered &&
		    module->client_registered != NULL)
			module->client_registered(client);
	}
}

void module_client_accepted(Client *client)
{
	IAuthModule *module;
	void **data;

	for (module = modules; module != NULL; module = module->next) {
		data = CLIENT_MODULE_DATA(client, module);
		if (data != NULL && *data != NULL) {
			if (module->client_accepted != NULL)
				module->client_accepted(client, *data);
			module_unref(module);
			*data = NULL;
		}
	}
}

void module_client_timeout(Client *client)
{
	IAuthModule *module;
	void **data;

        /* first notify with the timeout event */
	for (module = modules; module != NULL; module = module->next) {
		data = CLIENT_MODULE_DATA(client, module);
		if (data != NULL && *data != NULL) {
			if (module->client_timeout != NULL)
				module->client_timeout(client, *data);
		}
	}

	iauth_client_authenticated(client);
}

int module_client_read_line(Client *client, const char *cmd,
			    int argc, const char *args[])
{
	IAuthModule *module;
	void **data;
        int handled;

        handled = FALSE;
	for (module = modules; module != NULL; module = module->next) {
		if (module->client_read_line == NULL)
			continue;

		data = CLIENT_MODULE_DATA(client, module);
		if (data != NULL && *data != NULL &&
		    module->client_read_line(client, cmd, argc, args, *data))
			handled = TRUE;
	}

        return handled;
}

void module_session_stop(IAuthModule *module, Client *client)
{
	void **data;

	data = CLIENT_MODULE_DATA(client, module);
        *data = NULL;
}

void modules_init(void)
{
	modules = NULL;
	modules_max_refnum = -1;
}

void modules_deinit(void)
{
	while (modules != NULL)
                module_unregister(modules->name);
}
