/*
 m_proxy.c : open proxy detection module for iauth

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "module.h"

/* max. 30sec wait after client is let in to server */
#define MAX_WAIT_TIME (1000*30)

static int init(void);
static void deinit(void);

static void client_connected(Client *client, void **data);
static void client_rejected(Client *client, void *data);
static void client_accepted(Client *client, void *data);

IAuthModule m_proxy = {
	"proxy",

	init,
	deinit,

	NULL, /* connection_acccepted */
	NULL, /* client_registered */

	client_connected,
	client_rejected,
        client_accepted,

        NULL, /* client_timeout */
	NULL /* client_read_line */
};

Pool proxy_pool;

void proxy_close(ProxyModuleData *data, int open_proxy_found)
{
	if (data->closed)
		return;
        data->closed = TRUE;

        if (open_proxy_found)
		client_auth_abort(data->client, "open proxy");

        proxy_checks_close(data);

        DEBUG((data->client, "proxy done"));
	client_register_unref(data->client);
}

static void client_connected(Client *client, void **data)
{
	ProxyModuleData *proxy_data;

        /* skip proxy checks for outgoing connections */
	if (client->info.connect)
                return;

	proxy_data = p_new(proxy_pool, ProxyModuleData, 1);
        proxy_data->client = client;
	if (net_getsockname(client->handle, &proxy_data->my_ip,
			    &proxy_data->my_port) != 0) {
		/* failed, not good */
                fprintf(stderr, "net_getsockname() failed\n");
		p_free(proxy_pool, proxy_data);
		return;
	}

        *data = proxy_data;
	client_register_ref(client);

        proxy_scan_start(proxy_data);
}

static void proxy_data_destroy(ProxyModuleData *data)
{
        proxy_close(data, FALSE);
        p_free(proxy_pool, data);
}

static void client_rejected(Client *client, void *data)
{
        proxy_data_destroy(data);
}

static void abort_timeout(ProxyModuleData *data)
{
        /* timeouted, port is probably firewalled or network is very laggy */
	proxy_data_destroy(data);
}

static void client_accepted(Client *client, void *data)
{
	ProxyModuleData *proxy_data = data;

	if (proxy_data->closed)
		proxy_data_destroy(proxy_data);
	else {
		/* proxy checks are still going on - finish them and
		   if open proxies are found, kill the client later */
		proxy_data->to = timeout_add(MAX_WAIT_TIME,
					     (TimeoutFunc) abort_timeout, data);
	}
}

static int init(void)
{
        proxy_pool = pool_create("Proxy", 1024, TRUE);
        return TRUE;
}

static void deinit(void)
{
        pool_unref(proxy_pool);
}
