/*
 scan-checks.c : actual detection code for open proxies

    Copyright (C) 2002 Timo Sirainen

    Design of proxy scan modules, implementation specifics, and comments are
    mostly based on bopm (Blitzed Open Proxy Monitor,
    http://www.blitzed.org/bopm/) by

    Copyright (C) 2002 Erik Fears

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "module.h"

#define SOCKS_MAX_OUTBUF 22

int proxy_scan_r_squid(ProxyCheck *check)
{
        char *line;

	if (check->inbuf == NULL) {
		check->inbuf = tbuffer_create(check->handle, proxy_pool,
					      IO_PRIORITY_LOW, 256);
	}

	if (tbuffer_read(check->inbuf) < 0) {
                /* disconnected / buffer full */
		return check->internal_status;
	}

	line = (char *) tbuffer_next_line(check->inbuf);
	if (line == NULL)
                return -1;

	if (check->internal_status == 0) {
		if (strncasecmp(line, "HTTP/1.0 200", 12) != 0 &&
		    strncasecmp(line, "HTTP/1.1 200", 12) != 0)
			return 0;

		check->internal_status = 1;

		line = (char *) tbuffer_next_line(check->inbuf);
		if (line == NULL) {
			/* needs more data */
			return -1;
		}
	}

	/* Apache hack: (apache sends HTTP/1.x 200
	   sometimes when it shouldn't. This code checks
	   for the second header apache should send.*/
	if (strncasecmp(line, "Date:", 5) == 0)
		return 0;

        return 1;
}

int proxy_scan_w_squid(ProxyCheck *check)
{
	char *query, host[MAX_IP_LEN];
        int len;

	check->outbuf = tbuffer_create(check->handle, proxy_pool,
				       IO_PRIORITY_LOW, 256);
	query = (char *) tbuffer_get_space(check->outbuf, 256);

        net_ip2host(&check->module->my_ip, host);
	len = i_snprintf(query, 256, "CONNECT %s:%d HTTP/1.0\r\n\r\n",
			 host, check->module->my_port);

	if (tbuffer_send_buffer(check->outbuf, len) == -1)
		return 0;

        return -1;
}

static int proxy_read_data(ProxyCheck *check, int size, unsigned char **buffer)
{
	int buf_size;

	if (check->inbuf == NULL) {
		check->inbuf = tbuffer_create(check->handle, proxy_pool,
					      IO_PRIORITY_LOW, size);
	}

	if (tbuffer_read(check->inbuf) < 0) {
                /* disconnected - no proxy */
		return 0;
	}

	*buffer = tbuffer_get_data(check->inbuf, &buf_size);
	if (buf_size < size) {
                /* need more data */
		return -1;
	}

        return 1;
}

/*  REPLY byte order for socks4 
 *
 * 		+----+----+----+----+----+----+----+----+
 * 		| VN | CD | DSTPORT |      DSTIP        |
 * 		+----+----+----+----+----+----+----+----+
 * # of bytes:	   1    1      2              4
 * 
 * 
 * VN is the version of the reply code and should be 0. CD is the result
 * code with one of the following values:
 *
 * 	90: request granted
 *      91: request rejected or failed
 *      92: request rejected becasue SOCKS server cannot connect to
 *          identd on the client
 *      93: request rejected because the client program and identd
 *          report different user-ids
 *
 * 				    		     						 
 *
 */

int proxy_scan_r_socks4(ProxyCheck *check)
{
        unsigned char *buffer;
	int ret;

	ret = proxy_read_data(check, 8, &buffer);
	if (ret <= 0)
		return ret;

	return buffer[0] == 0 && buffer[1] == 90;
}

/*  CONNECT request byte order for socks4
 *
 *  		+----+----+----+----+----+----+----+----+----+----+....+----+
 *  		| VN | CD | DSTPORT |      DSTIP        | USERID       |NULL|
 *  		+----+----+----+----+----+----+----+----+----+----+....+----+
 *   # of bytes:  1    1      2              4           variable       1
 *
 *  VN = Version, CD = Command Code (1 is connect request)
 *
 */
int proxy_scan_w_socks4(ProxyCheck *check)
{
	unsigned char *query;
        int size;

        size = 9;
	check->outbuf = tbuffer_create(check->handle, proxy_pool,
				       IO_PRIORITY_LOW, size);
	query = tbuffer_get_space(check->outbuf, size);

        query[0] = 4; query[1] = 1;

        /* port */
	query[2] = (check->module->my_port & 0xff00) >> 8;
	query[3] = check->module->my_port & 0x00ff;

        /* IP */
	memcpy(query+4, &check->module->my_ip.ip, 4);

        /* user = empty */
        query[8] = 0;

	if (tbuffer_send_buffer(check->outbuf, 9) == -1)
		return 0;

        return -1;
}

int proxy_scan_w_socks5b(ProxyCheck *check)
{
	unsigned char *query;
	int len;

	query = tbuffer_get_space(check->outbuf, SOCKS_MAX_OUTBUF);

	query[0] = 5; query[1] = 1; query[2] = 0;
        len = 3;

#ifdef INET6
	if (IPADDR_IS_V6(&check->module->my_ip)) {
                query[3] = 4;
                memcpy(query+4, &check->module->my_ip.ip.s6_addr, 16);
		len += 17;
	} else
#endif
	{
                query[3] = 1;
                memcpy(query+4, &check->module->my_ip.ip, 4);
		len += 5;
	}

        query[len++] = (check->module->my_port & 0xff00) >> 8;
        query[len++] = check->module->my_port & 0x00ff;

	if (tbuffer_send_buffer(check->outbuf, len) == -1)
		return 0;

        return -1;
}

/*  Authentication selection reply message.
 *       +----+--------+
 *       |VER | METHOD |
 *       +----+--------+
 *       | 1  |   1    |
 *       +----+--------+
 *
 *   Version should always be 5, method should
 *   return back method selected by server for
 *   authentication. Method 0 indicates no authentication
 *   which indicates an open proxy.
 */
int proxy_scan_r_socks5(ProxyCheck *check)
{
        unsigned char *buffer;
	int ret;

	ret = proxy_read_data(check, 8, &buffer);
	if (ret <= 0)
		return ret;

	if (buffer[0] != 5 || buffer[1] != 0) {
		/* no socks */
		return 0;
	}

	if (check->internal_status != 0) {
                /* open socks */
		return 1;
	} else {
		/* found, but it still may be configured not to
		   allow open connections. Check #2 */
		check->internal_status = 1;
		proxy_scan_w_socks5b(check);
                return -1;
	}
}

/*  Send version authentication selection message to socks5
 *
 *      +----+----------+----------+
 *      |VER | NMETHODS | METHODS  |
 *      +----+----------+----------+
 *      | 1  |    1     | 1 to 255 |
 *      +----+----------+----------+
 *
 *  VER always contains 5, for socks version 5
 *  Method 0 is 'No authentication required'
 *
 */
int proxy_scan_w_socks5(ProxyCheck *check)
{
        unsigned char *query;

	check->outbuf = tbuffer_create(check->handle, proxy_pool,
				       IO_PRIORITY_LOW, SOCKS_MAX_OUTBUF);
	query = tbuffer_get_space(check->outbuf, 3);

	query[0] = 5; query[1] = 1; query[2] = 0;

	if (tbuffer_send_buffer(check->outbuf, 3) == -1)
		return 0;

        return -1;
}

int proxy_scan_r_wingate(ProxyCheck *check)
{
        char *line;

	if (check->inbuf == NULL) {
		check->inbuf = tbuffer_create(check->handle, proxy_pool,
					      IO_PRIORITY_LOW, 256);
	}

	if (tbuffer_read(check->inbuf) < 0) {
                /* disconnected / buffer full - no proxy */
		return 0;
	}

	line = (char *) tbuffer_next_line(check->inbuf);
	if (line == NULL)
                return -1;

	if (strncasecmp(line, "WinGate>", 8) == 0)
                return 1;

	if (strncasecmp(line,
			"Too many connected users - try again later", 42) == 0)
		return 1;

        return 0;
}

/*  Open wingates require no authentication, they
 *  will send a prompt when connect. No need to
 *  send any data.
 */
int proxy_scan_w_wingate(ProxyCheck *check)
{
        return -1;
}
