/*
 scan.c : manages the proxy scan queues

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "module.h"

typedef struct {
	const char *name;
	int port;

	ProxyScanFunc read_func;
        ProxyScanFunc write_func;
} Proxy;

/* NOTE: remember to update PROXY_SCAN_COUNT in module.h if you add/remove
   entries in this structure */
static Proxy scan_proxies[] = {
	{ "HTTP 8080",	8080,	proxy_scan_r_squid,	proxy_scan_w_squid },
	{ "HTTP 3128",	3128,	proxy_scan_r_squid,	proxy_scan_w_squid },
	{ "HTTP 80",	80,	proxy_scan_r_squid,	proxy_scan_w_squid },
	{ "Socks4",	1080,	proxy_scan_r_socks4,	proxy_scan_w_socks4 },
	{ "Socks5",	1080,	proxy_scan_r_socks5,	proxy_scan_w_socks5 },
	{ "Wingate",	23,	proxy_scan_r_wingate,	proxy_scan_w_wingate }
};

static void proxy_scan_check(ProxyModuleData *data);

/* close the proxy check */
static void proxy_check_close(ProxyCheck *check)
{
	if (check->status == PROXY_STATUS_CHECKED)
		return;
	check->status = PROXY_STATUS_CHECKED;

	DEBUG((check->module->client,
	       "proxy check %s done", scan_proxies[check->num].name));

	if (check->handle != -1)
		net_disconnect(check->handle);
	if (check->io != NULL)
		io_remove(check->io);

	tbuffer_destroy(check->inbuf);
	tbuffer_destroy(check->outbuf);
}

/* set all proxy checks for given port as CHECKED. */
static void proxy_port_refused(ProxyModuleData *data, int port)
{
	int i;

	for (i = 0; i < PROXY_SCAN_COUNT; i++) {
		if (scan_proxies[i].port == port)
                        proxy_check_close(&data->checks[i]);
	}
}

void proxy_checks_close(ProxyModuleData *data)
{
	int i;

	for (i = 0; i < PROXY_SCAN_COUNT; i++)
                proxy_check_close(&data->checks[i]);
}

static void io_read(ProxyCheck *check)
{
	switch (scan_proxies[check->num].read_func(check)) {
	case 0:
                /* no proxy */
		proxy_check_close(check);
		proxy_scan_check(check->module);
                break;
	case 1:
		/* open proxy */
                proxy_close(check->module, TRUE);
		break;
	}
}

static void io_connected(ProxyCheck *check)
{
	if (net_geterror(check->handle) != 0) {
                /* connect() failed */
		proxy_port_refused(check->module,
				   scan_proxies[check->num].port);
		proxy_check_close(check);
                proxy_scan_check(check->module);
		return;
	}

        /* switch to read-only IO waiting */
	io_remove(check->io);
	check->io = io_add(check->handle, IO_READ, (IOFunc) io_read, check);

        /* send the query */
	switch (scan_proxies[check->num].write_func(check)) {
	case 0:
                /* no proxy */
		proxy_check_close(check);
		proxy_scan_check(check->module);
                break;
	case 1:
                /* open proxy */
                proxy_close(check->module, TRUE);
		break;
	}
}

static void proxy_scan(ProxyCheck *check)
{
	int port;

	DEBUG((check->module->client,
	       "checking for proxy %s", scan_proxies[check->num].name));

	check->status = PROXY_STATUS_CHECKING;

	port = scan_proxies[check->num].port;
	check->handle = net_connect_ip(&check->module->client->ip, port,
				       &check->module->my_ip);
	if (check->handle == -1) {
                proxy_port_refused(check->module, port);
		proxy_scan_check(check->module);
	} else {
		check->io = io_add(check->handle, IO_READ|IO_WRITE,
				   (IOFunc) io_connected, check);
	}
}

static void proxy_scan_check(ProxyModuleData *data)
{
	int i, j, all_checked;

        all_checked = TRUE;
	for (i = 0; i < PROXY_SCAN_COUNT; i++) {
		if (data->checks[i].status != PROXY_STATUS_UNCHECKED)
			continue;

		/* don't connect more than once at a time to same port */
		for (j = 0; j < PROXY_SCAN_COUNT; j++) {
			if (i != j &&
			    scan_proxies[i].port == scan_proxies[j].port &&
			    data->checks[j].status == PROXY_STATUS_CHECKING) {
                                break;
			}
		}

		if (j != PROXY_SCAN_COUNT) {
			all_checked = FALSE;
			continue;
		}

		/* ok, scan this */
                proxy_scan(&data->checks[i]);
		all_checked = FALSE;
	}

	if (all_checked)
                proxy_close(data, FALSE);
}

void proxy_scan_start(ProxyModuleData *data)
{
	int i;

	for (i = 0; i < PROXY_SCAN_COUNT; i++) {
                data->checks[i].module = data;
		data->checks[i].num = i;
		data->checks[i].status = PROXY_STATUS_UNCHECKED;
                data->checks[i].handle = -1;
	}

	proxy_scan_check(data);
}

