#include "lib.h"
#include "tree.h"
#include "irc-limits.h"
#include "commands.h"

struct _CommandList {
	Tree *commands;
        Pool pool;
};

void command_register(CommandList *cmdlist, const char *name,
		      int min_params, void (*func)())
{
	Command *cmd;

	return_if_fail(tree_lookup(cmdlist->commands, name) == NULL);

	cmd = p_new(cmdlist->pool, Command, 1);
	cmd->name = p_strdup(cmdlist->pool, name);
        cmd->min_params = min_params;
	cmd->func = func;
	str_ucase(cmd->name);

        tree_insert(cmdlist->commands, cmd->name, cmd);
}

void command_unregister(CommandList *cmdlist, const char *name, void (*func)())
{
	Command *cmd;

	cmd = tree_lookup(cmdlist->commands, name);
	return_if_fail(cmd != NULL);
	return_if_fail(cmd->func == func);

	tree_remove(cmdlist->commands, cmd->name);
        p_free(cmdlist->pool, cmd->name);
	p_free(cmdlist->pool, cmd);
}

Command *command_find(CommandList *cmdlist, const char *name)
{
	return tree_lookup(cmdlist->commands, name);
}

const char *get_current_line(const char *orig_line, int known_args,
			     int use_last, const char *cmd,
			     int argc, const char *args[])
{
	int lengths[COMMAND_MAX_ARGS];
        char *line, *p;
	int i, len, cmdlen;

	if (argc <= known_args)
		return orig_line;

	if (use_last)
                argc = known_args;

	/* first calculate the size */
        cmdlen = strlen(cmd);
	len = cmdlen+1;
	for (i = 0; i < known_args-1; i++) {
                lengths[i] = strlen(args[i]);
		len += lengths[i]+1;
	}

        argc--;
        lengths[argc] = strlen(args[argc]);
	len += lengths[argc]+1;

        /* then copy the params */
	p = line = (char *) t_malloc(len);
	memcpy(p, cmd, cmdlen); p[cmdlen] = ' '; p += cmdlen+1;

	for (i = 0; i < known_args-1; i++) {
		memcpy(p, args[i], lengths[i]);
		p += lengths[i];
                *p++ = ' ';
	}

	memcpy(p, args[argc], lengths[argc]);
	p[lengths[i]] = '\0';

        return line;
}

CommandList *command_list_create(Pool pool)
{
	CommandList *list;

	list = p_new(pool, CommandList, 1);
	list->pool = pool;

	list->commands = tree_create(pool, (TreeCompareFunc) strcmp);
        return list;
}

void command_list_destroy(CommandList *list)
{
	Pool pool;

	pool = list->pool;
	tree_destroy(list->commands);
	p_free(pool, list);
}
