/*
 irc.c : prefix and name/args parsing for IRC commands

    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "irc-limits.h"

char *irc_parse_prefix(char *line, char **prefix, int *timestamp,
		       const char **id)
{
	*id = *prefix = NULL;
	*timestamp = 0;

	if (*line != ':') {
		/* no prefix given */
                return line;
	}

	*prefix = ++line;
	while (*line != ':' && *line != ' ' && *line != '\0') {
		line++;
	}

	if (*line == ':') {
		/* timestamp given as hex. parse it directly
		   for better performance. */
		*line++ = '\0';

		*timestamp = 0;
		while (*line != ':' && *line != ' ' && *line != '\0') {
			if (*line >= 'a' && *line <= 'f')
				*timestamp = (*timestamp << 4) |
					(10 + *line - 'a');
			else if (*line >= 'A' && *line <= 'F')
				*timestamp = (*timestamp << 4) |
					(10 + *line - 'A');
			else if (*line >= '0' && *line <= '9')
				*timestamp = (*timestamp << 4) |
					(*line - '0');
			else {
				/* invalid timestamp */
				*timestamp = 0;
				while (*line != ':' && *line != ' ' &&
				       *line != '\0') line++;
				break;
			}
			line++;
		}
	}

	if (*line == ':') {
		/* ID given for the command. */
		*line++ = '\0';

		*id = line;
		while (*line != ':' && *line != ' ' && *line != '\0') {
			line++;
		}
	}

	if (*line == ':') {
		/* reserved for future, skip the rest */
		*line++ = '\0';
		while (*line != ' ' && *line != '\0') {
			line++;
		}
	}

	if (*line == ' ')
		*line++ = '\0';

	/* ignore empty ID/prefix */
	if (*prefix != NULL && **prefix == '\0') *prefix = NULL;
	if (*id != NULL && **id == '\0') *id = NULL;

        return line;
}

const char *irc_parse_command(char *line, int *argc, const char *args[])
{
	char *cmd;
	int pos;

	/* skip leading whitespace */
	while (*line == ' ') line++;
	if (*line == '\0') {
		*argc = 0;
		return NULL; /* No command given */
	}

        /* first parameter is the command - convert it to uppercase */
	cmd = line;
	while (*line != '\0' && *line != ' ')
		line++;
	if (*line != '\0') *line++ = '\0';
        str_ucase(cmd);

	/* split the parameters into array */
	memset(args, 0, sizeof(args[0]) * (COMMAND_MAX_ARGS+1));
	pos = 0;
	while (*line != '\0' && pos < COMMAND_MAX_ARGS) {
		while (*line == ' ') line++;
		if (*line == '\0')
			break;

		if (*line == ':') {
			line++;
			args[pos++] = (const char *) line;
                        break;
		}

		args[pos++] = (const char *) line;
		while (*line != '\0' && *line != ' ') line++;

		if (*line == ' ')
			*line++ = '\0';
	}

        *argc = pos;
        return cmd;
}
