#include "common.h"
#include "channel.h"
#include "channel-modes.h"

struct _ModeBuffer {
	Channel *channel;
        SendOptions *sopts;

	int max_mode_args, max_args_length;
        int modes_pos, args_pos;
	int args_count;

	int modify_state; /* 1 = '+', 0 = '-' or -1 = unknown */
	char *modes;
        char *args;
};

ModeBuffer *mode_buffer_create(Channel *channel, SendOptions *sopts,
			       int max_mode_args, int max_line_length)
{
	ModeBuffer *buf;

	buf = t_malloc(sizeof(ModeBuffer));
        memset(buf, 0, sizeof(ModeBuffer));

	buf->channel = channel;
        buf->sopts = sopts;
	buf->max_mode_args = max_mode_args;
	buf->max_args_length = max_line_length;
        buf->modify_state = -1;
	buf->modes = t_malloc((sizeof(CHANNEL_MODES_LIST) +
			       max_mode_args)*2 + 1);
	buf->args = t_malloc(buf->max_args_length+1);

        return buf;
}

int mode_buffer_flush(ModeBuffer *buf)
{
	const char *data;

	if (buf->modes_pos == 0)
		return FALSE;

        buf->modes[buf->modes_pos] = '\0';
	buf->args[buf->args_pos] = '\0';

        buf->modes_pos = buf->args_pos = buf->args_count = 0;

	data = t_strconcat("MODE ", buf->channel->name, " ",
			   buf->modes, " ", buf->args, NULL);

	if (buf->sopts != NULL && buf->sopts->target_server != NULL)
		server_send(buf->sopts->target_server, buf->sopts, data);
	else
		channel_send(buf->channel, buf->sopts, data);
        return TRUE;
}

void mode_buffer_append(ModeBuffer *buf, int set, char mode, const char *arg)
{
	int arglen;

	if (arg == NULL || *arg == '\0')
		arglen = 0;
	else {
		arglen = strlen(arg);
		if (buf->args_count == buf->max_mode_args ||
		    buf->args_pos + arglen + 1 > buf->max_args_length) {
			mode_buffer_flush(buf);
		}

                buf->args_count++;
	}

	if (buf->args_pos + arglen > buf->max_args_length)
		i_panic("mode argument larger than line size: %s", arg);

	if (buf->args_pos > 0)
                buf->args[buf->args_pos++] = ' ';
	if (arglen > 0) {
		memcpy(buf->args + buf->args_pos, arg, arglen);
		buf->args_pos += arglen;
	}

	if (buf->modify_state != set) {
                buf->modify_state = set;
		buf->modes[buf->modes_pos++] = set ? '+' : '-';
	}
	buf->modes[buf->modes_pos++] = mode;
}

void mode_buffer_append_status(ModeBuffer *buf, int set,
			       unsigned int status, Nick *nick)
{
	if (status & (CSTATUS_OP|CSTATUS_OWNER))
		mode_buffer_append(buf, set, 'o', nick_get_outputid(nick));
	if (status & CSTATUS_VOICE)
		mode_buffer_append(buf, set, 'v', nick_get_outputid(nick));
}

char *channel_modes_add(char *str, int mode)
{
        if (mode & CMODE_PRIVATE) *str++ = 'p';
        if (mode & CMODE_SECRET) *str++ = 's';
        if (mode & CMODE_INVITEONLY) *str++ = 'i';
        if (mode & CMODE_OPTOPIC) *str++ = 't';
        if (mode & CMODE_NO_OUTSIDE_MSGS) *str++ = 'n';
        if (mode & CMODE_MODERATED) *str++ = 'm';
        if (mode & CMODE_QUIET) *str++ = 'q';
        if (mode & CMODE_ANONYMOUS) *str++ = 'a';
        if (mode & CMODE_REOP) *str++ = 'r';
	*str = '\0';

        return str;
}

/* Sends all masks in list matching given mask.
   Returns TRUE if anything was found. */
static int masklist_send(Channel *channel, Client *client,
			 char mode, const char *mask)
{
	PArray list;
	const char *item, *data, *numeric, *id;
        int i, exists;

	switch (mode) {
	case 'b':
		list = channel->bans;
                numeric = s_RPL_BANLIST;
                break;
	case 'e':
		list = channel->banexs;
                numeric = s_RPL_EXCEPTLIST;
                break;
	case 'I':
		list = channel->invites;
                numeric = s_RPL_INVITELIST;
		break;
	default:
                return FALSE;
	}

	if (list == NULL)
                return FALSE;

	exists = FALSE;
        parray_freeze(list);
	for (i = 0; i < list->last; i++) {
		item = parray_index(list, i);

		if (item != NULL && (mask == NULL || mask_match(mask, item) ||
				     mask_match(item, mask))) {
			exists = TRUE;
			if (client != NULL) {
                                id = nick_get_outputid(&client->nick);
				data = t_strdup_printf(numeric, id,
						       channel->name, item);
				client_send(client, NULL, data);
			}
		}
	}
        parray_thaw(list);

        return exists;
}

/* Returns TRUE if an exact match is found from list */
static int masklist_find(Channel *channel, char mode, const char *mask)
{
	PArray *list;
	const char *item;
        int i;

	list = chanmode_get_list(channel, mode);
	if (list == NULL || *list == NULL)
                return FALSE;

	for (i = 0; i < (*list)->last; i++) {
		item = parray_index(*list, i);

		if (irc_mask_cmp(item, mask) == 0)
                        return TRUE;
	}

        return FALSE;
}

/* Add/remove mask in list - don't check for duplicates. */
static void masklist_change(Channel *channel, int set,
			    char mode, const char *mask)
{
	PArray *list;
	char *item;
        int i;

	list = chanmode_get_list(channel, mode);

	if (set) {
                /* just add it */
		if (*list == NULL)
                        *list = parray_create(channel->pool, 10);
		parray_alloc(*list, p_strdup(channel->pool, mask));
		return;
	}

	if (*list == NULL)
                return;

        /* find and remove */
	for (i = 0; i < (*list)->last; i++) {
		item = parray_index(*list, i);

		if (irc_mask_cmp(item, mask) == 0) {
                        p_free(channel->pool, item);
			parray_free(*list, i);
			break;
		}
	}
}

int channel_modes_parse(Channel *channel, Client *client,
			SendOptions *sopts, SendOptions *reply_sopts,
			int reverse, int noprivileges,
			const char *mode, const char *mode_args[])
{
#define SKIP_FLAG_LIMIT      	0x01
#define SKIP_FLAG_BANLIST	0x02
#define SKIP_FLAG_EXCEPTLIST	0x04
#define SKIP_FLAG_INVITELIST 	0x08
        IrcNetwork *net;
	Nick *nick;
        ChannelNick *channick;
	ModeBuffer *mode_buf, *reply_buf;
	const char *mode_arg, *key;
	char limitstr[MAX_INT_STRLEN], modes[sizeof(CHANNEL_MODES_LIST)+1], *p;
        unsigned int status;
	int i, set, skip_flags, redundant;
        int noprivileges_complain, check_bans;
	int modeflag, modes_added, modes_removed, mode_args_used;

        net = channel->network;
	mode_buf = mode_buffer_create(channel, sopts, client_max_mode_args,
				      CLIENT_MAX_LINE_LENGTH);
        reply_buf = NULL;
	set = TRUE; modes_added = modes_removed = 0;

        noprivileges_complain = FALSE; check_bans = FALSE;
	mode_arg = NULL; mode_args_used = 0; skip_flags = 0;
	for (; *mode != '\0'; mode++) {
		/* we'll get the parameter here - if it's used
		   it should be set to NULL */
		if (mode_arg == NULL && *mode_args != NULL) {
			if (client == NULL ||
			    mode_args_used < client_max_mode_args) {
				mode_arg = *mode_args;
				mode_args++;
			}
		}

		switch (*mode) {
		case '+':
			set = TRUE;
                        break;
		case '-':
                        set = FALSE;
                        break;

		case 'k':
			/* channel key */
			if (mode_arg == NULL) {
				/* argument is required, even if it's not
				   used with -k */
				break;
			}

			key = mode_arg;
			mode_arg = NULL;

			if (noprivileges) {
                                /* not a chanop */
				noprivileges_complain = TRUE;
				break;
			}

			if (!set) {
                                /* -k */
				if ((channel->mode & CMODE_KEY) == 0) {
					/* already didn't have any key.. */
					break;
				}

				/* always use the channel's key instead of
				   the given parameter - IMHO that argument
				   shouldn't be needed like with -l, but
				   changing it now would break clients */
				key = channel->key;
			} else if (set && channel->key != NULL) {
				if (strcmp(key, channel->key) == 0) {
					/* identical mode is already set */
					break;
				}

				/* can't set key without removing old,
				   not even the servers. this would create
				   desyncs if servers had different keys in
				   channel, but they'd probably just swap the
				   keys with each other if this was allowed */
				if (client != NULL) {
					client_send_numeric(client, ERR_KEYSET,
							    channel->name);
				}
				break;
			} else if (strlen(key) > channel_max_key) {
				/* +k too_long_key, limit it */
                                key = t_strndup(key, channel_max_key);
			}

			if (!reverse) {
				mode_buffer_append(mode_buf, set, 'k', key);
			} else if (channel->mode & CMODE_KEY) {
				/* reverse - set back the old key */
				mode_buffer_append(mode_buf, '+', 'k',
						   channel->key);
			} else {
                                /* reverse - remove the key */
				mode_buffer_append(mode_buf, '-', 'k', key);
			}

			if (!reverse && set) {
				channel->mode |= CMODE_KEY;
				p_strdup_replace(channel->pool, &channel->key,
						 key);
			} else if (!reverse) {
				channel->mode &= ~CMODE_KEY;
				p_free(channel->pool, channel->key);
                                channel->key = NULL;
			}
			break;

		case 'l':
			if (noprivileges) {
                                /* not a chanop */
				noprivileges_complain = TRUE;
				if (set) mode_arg = NULL;
				break;
			}

			if (!set) {
                                /* -l doesn't have argument */
				if ((channel->mode & CMODE_LIMIT) == 0)
					break;

				if (reverse) {
					/* reverse the limit back */
					itoa(limitstr, channel->limit);
					mode_buffer_append(mode_buf, '+', 'l',
							   limitstr);
                                        break;
				}

                                /* remove limit */
				channel->mode &= ~CMODE_LIMIT;
                                channel->limit = 0;
				mode_buffer_append(mode_buf, '-', 'l', NULL);
				break;
			}

			/* +l <limit> */
			if (mode_arg == NULL || !is_numeric(mode_arg, '\0') ||
			    atoi(mode_arg) == 0)
				break;

                        /* allow only one limit change per mode */
			if (skip_flags & SKIP_FLAG_LIMIT)
                                break;
                        skip_flags |= SKIP_FLAG_LIMIT;

			if (!reverse) {
				mode_buffer_append(mode_buf, '+', 'l',
						   mode_arg);
			} else if (channel->mode & CMODE_LIMIT) {
                                /* reverse - set back the old limit */
				itoa(limitstr, channel->limit);
				mode_buffer_append(mode_buf, '+', 'l',
						   limitstr);
			} else {
                                /* reverse - remove the limit */
				mode_buffer_append(mode_buf, '-', 'l', NULL);
			}

			if (!reverse) {
				channel->mode |= CMODE_LIMIT;
				channel->limit = atoi(mode_arg);
			}
			mode_arg = NULL;
			break;

		case 'o':
		case 'v':
			if (noprivileges) {
                                /* not a chanop */
				noprivileges_complain = TRUE;
                                mode_arg = NULL;
				break;
			}

			nick = client != NULL ?
				ircnet_find_nick_name(net, mode_arg) :
                                ircnet_find_nick(net, mode_arg);
			if (nick == NULL) {
				/* unknown nick - nick changes might have
				   caused desyncs here, but we'll leave it up
				   to UIDs to fix this problem. */
				if (client != NULL) {
					client_send_numeric(client,
							    ERR_NOSUCHNICK,
							    mode_arg);
				}
				mode_arg = NULL;
                                break;
			}

                        channick = channel_get_nick(channel, nick);
			if (channick == NULL) {
				if (client != NULL) {
					client_send_numeric(client,
						ERR_USERNOTINCHANNEL,
						mode_arg, channel->name);
				}
                                mode_arg = NULL;
				break;
			}
			mode_arg = NULL;

			status = *mode == 'O' ? CSTATUS_OWNER :
				*mode == 'o' ? CSTATUS_OP :
				*mode == 'v' ? CSTATUS_VOICE : 0;

			redundant = (set &&
				     (channick->status & status) == status) ||
				(!set && (channick->status & status) == 0);

			if (reverse) {
				if (!redundant) {
					/* reverse */
					mode_buffer_append_status(
						mode_buf, !set, status, nick);
				}
                                break;
			}

			if (redundant && client == NULL) {
				/* server originated redundant mode
				   change, ignore */
				break;
			}

			/* always forward user-originated o/v modes
			   so they can be used to fix desyncs */
			if (set)
				channick->status |= status;
			else {
				if (status & CSTATUS_OP) {
                                        /* -o removes owner status as well */
					status |= CSTATUS_OWNER;
				}
				channick->status &= ~status;
			}
			mode_buffer_append_status(mode_buf, set, status, nick);

			if (!nick->server->local || reply_sopts != NULL)
				break;

			if (*mode == 'o') {
                                /* ops override mute-on-bans */
				check_bans = TRUE;
			}

			/* received mode change for local nick. send back the
			   mode change to avoid desyncs (yes, while
			   generating mostly useless network traffic) */
			if (reply_buf == NULL) {
				reply_buf = mode_buffer_create(channel,
					reply_sopts, server_max_mode_args,
					server_max_line_length);
			}
                        mode_buffer_append_status(reply_buf, set, status, nick);
			break;

		case 'b':
			/* allow only one ban list query per mode */
			if (mode_arg == NULL &&
			    (skip_flags & SKIP_FLAG_BANLIST)) {
				break;
			}
		case 'e':
			/* allow only one except list query per mode */
			if (mode_arg == NULL &&
			    (skip_flags & SKIP_FLAG_EXCEPTLIST)) {
				break;
			}
		case 'I':
			if (mode_arg == NULL) {
				/* querying for the list */
				if (client == NULL) {
                                        /* only clients need to query it */
					break;
				}

				/* allow only one invite list query per mode */
                                if (*mode == 'I' && mode_arg == NULL &&
				    (skip_flags & SKIP_FLAG_INVITELIST)) {
					break;
				}

				masklist_send(channel, client, *mode, NULL);
				switch (*mode) {
				case 'b':
					client_send_numeric(client,
							    RPL_ENDOFBANLIST,
                                                            channel->name);
					skip_flags |= SKIP_FLAG_BANLIST;
					break;
				case 'e':
					client_send_numeric(client,
							    RPL_ENDOFEXCEPTLIST,
                                                            channel->name);
					skip_flags |= SKIP_FLAG_EXCEPTLIST;
					break;
				case 'I':
					client_send_numeric(client,
							    RPL_ENDOFINVITELIST,
							    channel->name);
					skip_flags |= SKIP_FLAG_INVITELIST;
                                        break;
				}
				break;
			}

			if (noprivileges) {
                                /* not a chanop */
				noprivileges_complain = TRUE;
                                mode_arg = NULL;
				break;
			}

                        mode_arg = nick_mask_fix(mode_arg);
			if (set && masklist_send(channel, client,
						 *mode, mode_arg)) {
                                /* conflicts with some other mask */
                                mode_arg = NULL;
				break;
			}

			if (!reverse) {
				mode_buffer_append(mode_buf, set,
						   *mode, mode_arg);
				masklist_change(channel, set, *mode, mode_arg);
				if (*mode == 'b' || *mode == 'e')
                                        check_bans = TRUE;
			} else if (set ||
				   masklist_find(channel, *mode, mode_arg)) {
                                /* reverse */
				mode_buffer_append(mode_buf, !set,
						   *mode, mode_arg);
			}

                        mode_arg = NULL;
			break;

		default:
			/* check for the simple channel modes */
			switch (*mode) {
			case 'p':
				modeflag = CMODE_PRIVATE;
				break;
			case 's':
				modeflag = CMODE_SECRET;
				break;
			case 'i':
				modeflag = CMODE_INVITEONLY;
				break;
			case 't':
				modeflag = CMODE_OPTOPIC;
				break;
			case 'n':
				modeflag = CMODE_NO_OUTSIDE_MSGS;
				break;
			case 'm':
				modeflag = CMODE_MODERATED;
				break;
			case 'q':
				modeflag = CMODE_QUIET;
				break;
			case 'a':
				modeflag = CMODE_ANONYMOUS;
				break;
			case 'r':
				modeflag = CMODE_REOP;
				break;
			default:
				modeflag = 0;
				break;
			}

			if (modeflag == 0) {
				/* unknown mode change */
				if (client != NULL) {
					client_send_numeric(client,
							    ERR_UNKNOWNMODE,
							    *mode,
							    channel->name);
				}
			} else if (set) {
				if (modes_removed & modeflag)
					modes_removed &= ~modeflag;
                                else
					modes_added |= modeflag;
			} else {
				if (modes_added & modeflag)
					modes_added &= ~modeflag;
                                else
					modes_removed |= modeflag;
			}
                        break;
		}
	}

	modes_added &= ~channel->mode;
	if (modes_added != 0) {
		/* add modes */
		if (noprivileges) {
			/* not a chanop */
			noprivileges_complain = TRUE;
		} else {
			p = channel_modes_add(modes, modes_added);

			for (p = modes; *p != '\0'; p++) {
				mode_buffer_append(mode_buf, !reverse,
						   *p, NULL);
			}

			if (!reverse)
				channel->mode |= modes_added;
		}
	}

	modes_removed &= channel->mode;
	if (modes_removed != 0) {
		/* remove modes */
		if (noprivileges) {
			/* not a chanop */
			noprivileges_complain = TRUE;
		} else {
			channel_modes_add(modes, modes_removed);

			for (p = modes; *p != '\0'; p++)
				mode_buffer_append(mode_buf, reverse, *p, NULL);

			if (!reverse)
				channel->mode &= ~modes_removed;
		}
	}

	if (noprivileges_complain && client != NULL) {
		client_send_numeric(client, ERR_CHANOPRIVSNEEDED,
				    channel->name);
	}

	if (check_bans && channel->clients != NULL) {
		/* recheck banned clients in this channel */
		for (i = 0; i < channel->clients->last; i++) {
			Client *client = parray_index(channel->clients, i);

                        channel_check_client_bans(channel, client);
		}
	}

	return mode_buffer_flush(mode_buf);
}
