#include "common.h"
#include "ioloop.h"
#include "channel.h"
#include "channel-modes.h"

#define CHANNEL_POOL_INITIAL_SIZE	1024
#define CHANNEL_INITIAL_CLIENTS_COUNT	16
#define NICK_INITIAL_CHANNELS           10

static Pool channel_pool;

Channel *channel_create(IrcNetwork *net, const char *name)
{
	Channel *channel;

        return_val_if_fail(net != NULL, NULL);
        return_val_if_fail(name != NULL, NULL);

	channel = p_new(channel_pool, Channel, 1);
	channel->pool = pool_create(t_strdup_printf("Channel %s", name),
				    CHANNEL_POOL_INITIAL_SIZE, TRUE);
        channel->network = net;
        channel->name = p_strdup(channel->pool, name);
	channel->nicks = tree_create(channel->pool, NULL);
        channel->created = ioloop_time;

	net->channels_count++;
        net->server->channels_count++;
	hash_insert(net->channels, channel->name, channel);
        return channel;
}

#ifdef POOL_CHECK_LEAKS
static int channel_nicks_clean(Nick *nick, ChannelNick *channick,
			       Channel *channel)
{
	p_free(channel->pool, channick);
        return TRUE;
}

static void mask_free(char *mask, Channel *channel)
{
        p_free(channel->pool, mask);
}
#endif

void channel_destroy(Channel *channel)
{
	channel->network->channels_count--;
	channel->network->server->channels_count--;

        hash_remove(channel->network->channels, channel->name);

#ifdef POOL_CHECK_LEAKS
	tree_traverse(channel->nicks, (TreeTraverseFunc) channel_nicks_clean,
		      channel);

	tree_destroy_clean(channel->nicks);
        parray_destroy_clean(channel->clients);
	tree_destroy_clean(channel->servers);

        parray_foreach(channel->bans, (PForeachFunc) mask_free, channel);
        parray_foreach(channel->banexs, (PForeachFunc) mask_free, channel);
        parray_foreach(channel->invites, (PForeachFunc) mask_free, channel);
        parray_destroy_clean(channel->bans);
        parray_destroy_clean(channel->banexs);
        parray_destroy_clean(channel->invites);

	p_free_clean(channel->pool, channel->name);
	p_free_clean(channel->pool, channel->topic);
	p_free_clean(channel->pool, channel->topic_nick);
	p_free_clean(channel->pool, channel->key);
#endif

	pool_unref(channel->pool);
        p_free(channel_pool, channel);
}

static int server_refcount_update(Channel *channel, Server *server, int change)
{
	int refcount;

	server = server_find_first_link(server);

	refcount = POINTER_TO_INT(tree_lookup(channel->servers, server));
	refcount += change;

	if (refcount != 0) {
		tree_insert(channel->servers, server,
			    INT_TO_POINTER(refcount));
	} else {
		tree_remove(channel->servers, server);
	}

        return refcount;
}

ChannelNick *channel_add_nick(Channel *channel, Nick *nick)
{
	ChannelNick *channick;

	if (nick->server->local) {
		/* add to local clients list */
		if (channel->clients == NULL) {
			channel->clients =
				parray_create(channel->pool,
					      CHANNEL_INITIAL_CLIENTS_COUNT);
		}
                parray_alloc(channel->clients, client_find(nick->nick));
	} else {
		/* update server refcount in servers list */
		if (channel->servers == NULL) {
			channel->servers =
				tree_create(channel->pool, server_name_cmp);
		}

                server_refcount_update(channel, nick->server, 1);
	}

        /* update list of channels for nick */
	if (nick->channels == NULL) {
		nick->channels = parray_create(nick->server->pool,
					       NICK_INITIAL_CHANNELS);
	}
        parray_alloc(nick->channels, channel);

	channick = p_new(channel->pool, ChannelNick, 1);
	channick->nick = nick;

	channel->nicks_count++;
	tree_insert(channel->nicks, nick, channick);

        return channick;
}

void channel_remove_nick(Channel *channel, Nick *nick)
{
        ChannelNick *channick;
	int index;

	channick = tree_lookup(channel->nicks, nick);
	return_if_fail(channick != NULL);
        p_free(channel->pool, channick);

	if (nick->server->local) {
		/* remove from local clients list */
		index = parray_find(channel->clients, client_find(nick->nick));
		if (index >= 0)
                        parray_free(channel->clients, index);
	} else {
		/* update server refcount in servers list */
		if (server_refcount_update(channel, nick->server, -1) == 0) {
			/* FIXME: no references to server - if it's a leaf
			   server we should drop the channel.. */
		}
	}

        /* update list of channels for nick */
	index = parray_find(nick->channels, channel);
	if (index >= 0)
                parray_free(nick->channels, index);

        channel->nicks_count--;
	tree_remove(channel->nicks, nick);

	if (channel->nicks_count == 0)
                channel_destroy(channel);
}

void channels_remove_nick(Nick *nick)
{
	Channel *channel;
        PArray channels;
        int i;

	if (nick->channels == NULL)
		return;

	channels = nick->channels;
	nick->channels = NULL;

	for (i = 0; i < channels->last; i++) {
		channel = parray_index(channels, i);
		channel_remove_nick(channel, nick);
	}

        parray_destroy(channels);
}

ChannelNick *channel_get_nick(Channel *channel, Nick *nick)
{
        return tree_lookup(channel->nicks, nick);
}

void channel_add_client_invite(Channel *channel, Client *client, Nick *inviter)
{
	ClientInvite *rec, *next;
        int count;

	rec = p_new(client->pool, ClientInvite, 1);
        rec->nick = p_strdup(client->pool, inviter->nick);
        rec->channel = p_strdup(client->pool, channel->name);

        rec->next = client->invites;
	client->invites = rec;

        /* remove oldest invites if there's too many of them */
        count = 0;
	for (rec = client->invites; rec != NULL; rec = next) {
		next = rec->next;

                if (count < client_max_invites)
			count++;
		else {
			rec->next = NULL;
                        client_invite_destroy(client, rec);
		}
	}
}

void channel_remove_client_invite(Channel *channel, Client *client)
{
        ClientInvite **tmp, *invite;

        /* remove from client's invite list */
	for (tmp = &client->invites; *tmp != NULL; tmp = &(*tmp)->next) {
		invite = *tmp;

		if (irc_channel_cmp(invite->channel, channel->name) == 0) {
			*tmp = invite->next;
			client_invite_destroy(client, invite);
			break;
		}
	}
}

void channel_set_mode(Channel *channel, unsigned int mode)
{
        channel->mode = mode | (channel->mode & (CMODE_LIMIT|CMODE_KEY));
}

const char *channel_get_mode(Channel *channel, const char **args)
{
	char modes[sizeof(CHANNEL_MODES_LIST)+2], *p;
        char limit[MAX_INT_STRLEN];

	p = modes; *p++ = '+';
        p = channel_modes_add(p, channel->mode);

	if (channel->mode & CMODE_LIMIT) {
                itoa(limit, channel->limit);
		*p++ = 'l';
	}
        if (channel->mode & CMODE_KEY)
		*p++ = 'k';

	if (args != NULL) {
		if ((channel->mode & (CMODE_KEY|CMODE_LIMIT)) ==
		    (CMODE_KEY|CMODE_LIMIT))
			*args = t_strconcat(limit, " ", channel->key, NULL);
		else if (channel->mode & CMODE_LIMIT)
			*args = t_strdup(limit);
		else if (channel->mode & CMODE_KEY)
			*args = t_strdup(channel->key);
		else
                        *args = "";
	}

	*p = '\0';
	return t_strdup(modes);
}

int channel_set_topic(Channel *channel, const char *topic, const char *nick,
		      time_t timestamp, int check_limit)
{
	p_free(channel->pool, channel->topic);
        p_free(channel->pool, channel->topic_nick);

	channel->topic = p_strndup(channel->pool, topic, channel_max_topic);
	channel->topic_nick = p_strdup(channel->pool, nick);
        channel->topic_time = timestamp;
        return TRUE;
}

const char *channel_match_ban(Channel *channel, Client *client)
{
	const char *mask, *match;
        int i;

        if (channel->bans == NULL)
		return NULL;

	/* +b matches? */
        match = NULL;
	for (i = 0; i < channel->bans->last; i++) {
		mask = parray_index(channel->bans, i);

		if (mask_match(mask, client->nickmask) ||
		    mask_match(mask, client->ipmask)) {
			match = mask;
                        break;
		}
	}

	if (match == NULL)
		return NULL;

	if (channel->banexs != NULL) {
		/* +e matches? */
		for (i = 0; i < channel->banexs->last; i++) {
			mask = parray_index(channel->banexs, i);

			if (mask_match(mask, client->nickmask) ||
			    mask_match(mask, client->ipmask))
				return NULL;
		}
	}

        return match;
}

void channel_check_client_bans(Channel *channel, Client *client)
{
	ChannelNick *channick;

	channick = channel_get_nick(channel, &client->nick);
	if (channick == NULL)
		return;

	if ((channick->status & CSTATUS_OP) == 0 &&
	    channel_match_ban(channel, client)) {
                /* banned */
		channick->status |= CSTATUS_BANNED;
	} else {
                /* not banned */
		channick->status &= ~CSTATUS_BANNED;
	}
}

void channel_send(Channel *channel, SendOptions *sopts, const char *data)
{
        if (sopts == NULL || !sopts->channel_no_servers)
		channel_send_servers(channel, sopts, data);
        if (sopts == NULL || !sopts->channel_no_clients)
		channel_send_clients(channel, sopts, data);
}

typedef struct {
	SendOptions *sopts;
	const char *data;
        int skip_nonleafs;
} ChannelServerSendData;

static int channel_send_server_data(Server *server, void *value,
				    ChannelServerSendData *data)
{
	if ((!data->skip_nonleafs || server->leaf) &&
	    (data->sopts == NULL || data->sopts->skip_server != server))
		server_send(server, data->sopts, data->data);

        return TRUE;
}

void channel_send_servers(Channel *channel, SendOptions *sopts,
			  const char *data)
{
        ChannelServerSendData senddata;
	Server *server;
        PArray links;
        int i, sending_talk;

        sending_talk = sopts != NULL && sopts->channel_send_talk;
	if (channel->servers != NULL) {
		senddata.sopts = sopts;
		senddata.data = data;
		senddata.skip_nonleafs = !sending_talk;

		/* when sending PRIVMSG/NOTICEs to channel ("talk"),
		   we send it here to all servers who are in channel->servers
		   list.

		   When sending non-talk, we send here only to leaf servers,
		   and later send it to all non-leafs. */
		tree_traverse(channel->servers,
			      (TreeTraverseFunc) channel_send_server_data,
			      &senddata);
	}

	if (sending_talk)
                return;

	/* send non-talk commands to all non-leaf servers */
        links = ircnet_links(channel->network);
	if (links == NULL)
                return;

        parray_freeze(links);
	for (i = 0; i < links->last; i++) {
		server = parray_index(links, i);

		if (server != NULL && !server->leaf &&
		    (sopts == NULL || server != sopts->skip_server))
			server_send(server, sopts, data);
	}
        parray_thaw(links);
}

void channel_send_clients(Channel *channel, SendOptions *sopts,
			  const char *data)
{
        Client *client;
	int i;

	if (channel->clients == NULL)
                return;

        parray_freeze(channel->clients);
	for (i = 0; i < channel->clients->last; i++) {
		client = parray_index(channel->clients, i);

		if (client != NULL &&
		    (sopts == NULL || client != sopts->skip_client))
			client_send(client, sopts, data);
	}
        parray_thaw(channel->clients);
}

void channels_init(void)
{
	channel_pool = pool_create("Channels", 1024, TRUE);
}

void channels_deinit(void)
{
        pool_unref(channel_pool);
}
