#ifndef __CHANNEL_H
#define __CHANNEL_H

#define CHANNEL_IS_MODELESS(channel) ((channel)->name[0] == '+')
#define CHANNEL_IS_LOCAL(channel) ((channel)->name[0] == '&')

#define NICK_CAN_TALK(channel, channick) \
	((channel->mode & CMODE_QUIET) == 0 && /* +q */ \
	 (((channel)->mode & CMODE_NO_OUTSIDE_MSGS) == 0 || /* +n */ \
		(channick) != NULL) && \
	 (((channel)->mode & CMODE_MODERATED) == 0 || /* +m */ \
		((channick) != NULL && \
		((channick)->status & (CSTATUS_OP|CSTATUS_VOICE)) != 0)) && \
	 ((channick) == NULL || ((channick)->status & CSTATUS_BANNED) == 0))


#define CHANNEL_MODES_LIST	"abeiIklmnoOpqrstv"

#define CMODE_PRIVATE		0x001 /* +p */
#define CMODE_SECRET		0x002 /* +s */
#define CMODE_INVITEONLY	0x004 /* +i */
#define CMODE_OPTOPIC		0x008 /* +t */
#define CMODE_NO_OUTSIDE_MSGS	0x010 /* +n */
#define CMODE_MODERATED		0x020 /* +m */
#define CMODE_LIMIT		0x040 /* +l */
#define CMODE_KEY		0x080 /* +k */
#define CMODE_QUIET		0x100 /* +q (invisible to users) */
#define CMODE_ANONYMOUS		0x200 /* +a */
#define CMODE_REOP		0x400 /* +r */

struct _Channel {
        IrcNetwork *network;
	char *name;

        time_t created;
	unsigned int nicks_count;

	char *topic;
        char *topic_nick;
        time_t topic_time;

	unsigned int mode;
	unsigned int limit;
        char *key;

	PArray bans, banexs, invites; /* +b, +e, +I - may be NULL */

        Tree *nicks; /* Nick -> ChannelNicks */
	PArray clients; /* Local clients we have in this channel.
		May be NULL. */
	Tree *servers; /* Server -> refcount, servers we're connected to that
		have clients, or it's downlinks have clients in this channel.
		May be NULL. */

	/* All data in channel structure is allocated from this memory pool,
	   except the above. There's no hardcoded limits for pretty much
	   anything, but this pool limits the total amount of memory to
	   allocate for a channel. Smaller channels should have a small pool,
	   and bigger channels a bigger pool. Only commands from users should
	   check the pool size, never the commands from other servers. */
        Pool pool;
};

#define CSTATUS_OWNER		0x01
#define CSTATUS_OP		0x02
#define CSTATUS_VOICE		0x04
#define CSTATUS_BANNED		0x10 /* only for local clients */

typedef struct {
	Nick *nick;

        unsigned int status;
} ChannelNick;

/* Create a new channel. */
Channel *channel_create(IrcNetwork *net, const char *name);
/* Destroy a channel. */
void channel_destroy(Channel *channel);

/* Add new nick to channel. */
ChannelNick *channel_add_nick(Channel *channel, Nick *nick);
/* Remove a nick from channel. */
void channel_remove_nick(Channel *channel, Nick *nick);
/* Remove a nick from all channels. */
void channels_remove_nick(Nick *nick);

/* Returns nick channel record for specified nick in channel. */
ChannelNick *channel_get_nick(Channel *channel, Nick *nick);

/* Add new invitation to channel. */
void channel_add_client_invite(Channel *channel, Client *client, Nick *by);
/* Remove the client from channel's invite list. */
void channel_remove_client_invite(Channel *channel, Client *client);

/* Set the channel mode. You can't change the +k or +l modes with this. */
void channel_set_mode(Channel *channel, unsigned int status);
/* Returns the channel mode as string, always prefixed with '+'.
   Returns +k and +l arguments in args. */
const char *channel_get_mode(Channel *channel, const char **args);

/* Each of the functions below returns TRUE if successful, FALSE if
   channel's pool size is full and check_limit was TRUE. */
int channel_set_topic(Channel *channel, const char *topic, const char *nick,
		      time_t timestamp, int check_limit);
/*FIXME: int channel_set_key(Channel *channel, const char *key);
int channel_set_limit(Channel *channel, unsigned int limit);

int channel_set_ban(Channel *channel, const char *ban, int add);
int channel_set_banex(Channel *channel, const char *banex, int add);
int channel_set_invite(Channel *channel, const char *invite, int add);*/

/* Returns ban mask for client, or NULL if client isn't banned */
const char *channel_match_ban(Channel *channel, Client *client);
/* Update CSTATUS_BANNED flag in given channel. */
void channel_check_client_bans(Channel *channel, Client *client);

/* Send data to all clients in channel and all servers that have
   clients in this channel. */
void channel_send(Channel *channel, SendOptions *sopts, const char *data);
void channel_send_servers(Channel *channel, SendOptions *sopts,
			  const char *data);
void channel_send_clients(Channel *channel, SendOptions *sopts,
			  const char *data);

void channels_init(void);
void channels_deinit(void);

#endif
