#include "common.h"
#include "client-commands.h"
#include "command-replies.h"
#include "channel.h"
#include "channel-modes.h"
#include "messages.h"

/* Find the given channel, send ERR_NOSUCHCHANNEL if it's not found. */
static Channel *client_find_channel_noisy(Client *client, const char *name)
{
	Channel *channel;

	channel = ircnet_find_channel(client->server->network, name);
	if (channel == NULL)
		client_send_numeric(client, ERR_NOSUCHCHANNEL, name);

        return channel;
}

/* Find the channel nick for given client/channel,
   send ERR_NOTONCHANNEL if not found. */
static ChannelNick *client_find_channick_noisy(Client *client, Channel *channel)
{
	ChannelNick *channick;

        channick = channel_get_nick(channel, &client->nick);
	if (channick == NULL)
		client_send_numeric(client, ERR_NOTONCHANNEL, channel->name);

        return channick;
}

/* returns 0 = not invited, 1 = invited, 2 = +I matches */
static ClientInvite *client_find_invite(Client *client, Channel *channel)
{
        ClientInvite *invite;

	for (invite = client->invites; invite != NULL; invite = invite->next) {
		if (irc_channel_cmp(invite->channel, channel->name) == 0)
			return invite;
	}

	return NULL;
}

static const char *client_match_invitelist(Client *client, Channel *channel)
{
	const char *mask;
        int i;

        if (channel->invites == NULL)
		return NULL;

	/* +I matches? */
	for (i = 0; i < channel->invites->last; i++) {
		mask = parray_index(channel->invites, i);

		if (mask_match(mask, client->nickmask) ||
		    mask_match(mask, client->ipmask))
                        return mask;
	}

        return NULL;
}

static int client_can_join(Client *client, Channel *channel, const char *key)
{
	ClientInvite *invite;
        const char *banmask, *data;

	/* +k */
	if ((channel->mode & CMODE_KEY) &&
	    (key == NULL || irc_case_cmp(channel->key, key) != 0)) {
                client_send_numeric(client, ERR_BADCHANNELKEY, channel->name);
		return FALSE;
	}

        /* inviting overrides everything */
	invite = client_find_invite(client, channel);

	/* +b */
        banmask = channel_match_ban(channel, client);
	if (banmask != NULL && invite == NULL) {
		client_send_numeric(client, ERR_BANNEDFROMCHAN,
				    channel->name);
		return FALSE;
	}

	/* +i */
	if ((channel->mode & CMODE_INVITEONLY) && invite == NULL &&
	    client_match_invitelist(client, channel) == NULL) {
                client_send_numeric(client, ERR_INVITEONLYCHAN, channel->name);
		return FALSE;
	}

        /* +l */
	if (invite == NULL && (channel->mode & CMODE_LIMIT) &&
	    channel->limit >= channel->nicks_count) {
                client_send_numeric(client, ERR_CHANNELISFULL, channel->name);
		return FALSE;
	}

	if (banmask != NULL && invite != NULL) {
		/* the user is banned, so notify users about the invite */
		data = t_strdup_printf(INVITED_BANNED_USER_MSG, channel->name,
				       client->nick.nick, invite->nick,
				       banmask);
		channel_send(channel, NULL, data);
	}

        if (invite != NULL)
		channel_remove_client_invite(channel, client);
	return TRUE;
}

/* JOIN <channel>[,...] [<key>[,...]] */
CLIENT_COMMAND(cmd_join)
{
        SendOptions sopts;
        IrcNetwork *network;
	Channel *channel;
	ChannelNick *channick;
        const char *data;
	char *channame, *key, *mode, *channels, *keys;
        int channels_count;

	if (strcmp(args[0], "0") == 0) {
		/* parting all channels */
		/*FIXME: cmd_join_0(client, args[1]); */
                return;
	}

        network = client->server->network;

	channels = (char *) args[0]; keys = (char *) args[1];
	channame = i_strtoken(&channels, ',');
	key = i_strtoken(&keys, ',');

	channels_count = client->nick.channels == NULL ? 0 :
		parray_size(client->nick.channels);
	for (; channame != NULL; channame = i_strtoken(&channels, ','),
	     key = i_strtoken(&keys, ',')) {
                /* always ignore the mode */
		mode = strchr(channame, '\007');
		if (mode != NULL)
			*mode++ = '\0';

		if (!is_chanprefix_char(*channame)) {
			/* unknown channel prefix, return a bit
			   non-descriptive error message to keep
			   compatibility */
			client_send_numeric(client, ERR_NOSUCHCHANNEL,
					    channame);
                        continue;
		}

		channel = ircnet_find_channel(network, channame);
		if (channel != NULL &&
		    channel_get_nick(channel, &client->nick) != NULL) {
                        /* already joined - ignore */
			continue;
		}

		if (channels_count == client_max_channels) {
                        /* joined in too many channels */
			client_send_numeric(client, ERR_TOOMANYCHANNELS,
					    channame);
                        continue;
		}

		/* FIXME: add support for channel masks here */

		if (channel == NULL) {
			/* channel doesn't exist yet - create it */
			if (strlen(channame) > channel_max_name) {
                                /* channel name too long, limit it */
                                ((char *)channame)[channel_max_key] = '\0';
			}
			channel = channel_create(network, channame);
		} else if (!client_can_join(client, channel, key))
                        return;

		/* if client was invited through ban, we still need to
		   update it's banned-flag */
		channel_check_client_bans(channel, client);

		channels_count++;

		/* add nick to channel */
		channick = channel_add_nick(channel, &client->nick);
		if (channel->nicks_count == 1 &&
		    !CHANNEL_IS_MODELESS(channel)) {
			/* created empty channel - set client the channel
                           operator */
                        channick->status = CSTATUS_OP;
                        mode = "\007o";
		} else {
			mode = "";
		}

                /* notify all clients, including the one who requested JOIN */
		channel_send_clients(channel,
				     send_origin_nick(&sopts, &client->nick),
				     t_strconcat("JOIN ", channel->name, NULL));

                /* notify servers */
		if (!CHANNEL_IS_LOCAL(channel)) {
                        send_origin_nick(&sopts, &client->nick);
			data = t_strconcat("JOIN ", channel->name, mode, NULL);
			channel_send_servers(channel, &sopts, data);
		}

                /* send the /NAMES list to client */
		command_reply_names(client, channel);
	}
}

/* PART <channel>[,...] [?] <reason> */
CLIENT_COMMAND(cmd_part)
{
        SendOptions sopts;
	Channel *channel;
        ChannelNick *channick;
        char *channels, *channame, *reason;
	const char *data;

	if (argc == 1) {
		/* default part reason is the nick */
		reason = client->nick.nick;
	} else {
		reason = (char *) args[argc-1];
		if (strlen(reason) > channel_max_partreason) {
                        /* limit the part reason length */
			reason[channel_max_partreason] = '\0';
		}
	}

	send_origin_nick(&sopts, &client->nick);

	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		channel = client_find_channel_noisy(client, channame);
		if (channel == NULL)
			continue;

                channick = client_find_channick_noisy(client, channel);
		if (channick == NULL)
                        continue;

		if (channick->status & CSTATUS_BANNED) {
			/* banned, hide part reason */
			data = t_strdup_printf("PART %s :%s", channame,
					       client->nick.nick);
		} else {
			data = t_strdup_printf("PART %s :%s", channame, reason);
		}

		channel_send(channel, &sopts, data);
		channel_remove_nick(channel, &client->nick);
	}
}

static int kick_nicks(Channel *channel, Client *client, SendOptions *sopts,
		      char *nicks, const char *reason, int *count)
{
        ChannelNick *channick;
        Nick *nick;
	const char *data;
        char *nickname;

	channick = client_find_channick_noisy(client, channel);
	if (channick == NULL)
                return TRUE;

	if ((channick->status & CSTATUS_OP) == 0) {
		/* client isn't op in this channel */
		client_send_numeric(client,
				    ERR_CHANOPRIVSNEEDED,
				    channel->name);
                return TRUE;
	}

	for (nickname = i_strtoken(&nicks, ','); nickname != NULL;
	     nickname = i_strtoken(&nicks, ',')) {
		/* make sure the nick exists and is in channel */
		nick = client_find_nick_noisy(client, nickname);
		if (nick == NULL)
			return TRUE;

		channick = channel_get_nick(channel, nick);
		if (channick == NULL) {
			client_send_numeric(client, ERR_NOTONCHANNEL,
					    channel->name);
			return TRUE;
		}

		/* and kick it */
		data = t_strdup_printf("KICK %s %s :%s", channel->name,
				       nick->nick, reason);
		channel_send(channel, sopts, data);
		channel_remove_nick(channel, nick);

		if (--(*count) <= 0)
                        return FALSE;
	}

        return TRUE;
}

/* KICK <channel>[,...] <nick>[,...] [?] <reason> */
CLIENT_COMMAND(cmd_kick)
{
        SendOptions sopts;
	Channel *channel;
	char *channels, *channame, *nicks, *reason;
        int count;

	if (argc == 2)
		reason = client->nick.nick;
	else {
		reason = (char *) args[argc-1];
		if (strlen(reason) > channel_max_kickreason) {
                        /* limit the kick reason length */
			reason[channel_max_kickreason] = '\0';
		}
	}

	send_origin_nick(&sopts, &client->nick);
	channel = NULL;

	/* NOTE: Behaviour when giving multiple channels+nicks is not the
	   way RFC-2812 specifies it, but no-one else seems to respect it
	   either. Either way, it's pretty useless. */
	count = client_max_kick_targets;
	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		channel = client_find_channel_noisy(client, channame);
		if (channel == NULL)
			continue;

                nicks = (char *) t_strdup(args[1]);
		if (!kick_nicks(channel, client, &sopts, nicks,
				reason, &count))
                        break;
	}
}

/* INVITE <nick> <channel> */
CLIENT_COMMAND(cmd_invite)
{
        SendOptions sopts;
	Nick *nick;
        Channel *channel;
	ChannelNick *channick;
        char *p;

	nick = client_find_nick_noisy(client, args[0]);
	if (nick == NULL)
		return;

        /* don't allow any invalid chars in channel parameter */
	for (p = (char *) args[1]; *p != '\0'; p++) {
		if (*p == ' ' || *p == ',' || *p == 7) {
			*p = '\0';
                        break;
		}
	}

	channel = ircnet_find_channel(client->server->network, args[1]);
	if (channel != NULL) {
                /* check that the user isn't already in the channel */
		channick = channel_get_nick(channel, nick);
		if (channick != NULL) {
			client_send_numeric(client, ERR_USERONCHANNEL,
					    nick->nick, channel->name);
                        return;
		}
	}

	if (channel != NULL && (channel->mode & CMODE_INVITEONLY)) {
		/* invite only channel, are we ops? */
		channick = channel_get_nick(channel, &client->nick);
		if (channick == NULL || (channick->status & CSTATUS_OP) == 0) {
			client_send_numeric(client, ERR_CHANOPRIVSNEEDED,
					    channel->name);
                        return;
		}

		/* if it's local client, update the invite lists */
		if (nick->server->local) {
			channel_add_client_invite(channel,
						  client_find(nick->nick),
						  &client->nick);
		}
	}

        /* send INVITE + inviting-numeric */
	nick_send_data(nick, send_origin_nick(&sopts, &client->nick),
		       t_strdup_printf("INVITE %s :%s",
				       nick_get_outputid(nick), args[1]));
	client_send_numeric(client, RPL_INVITING, nick->nick, args[1]);
}

/* MODE <channel> [<modes> [<params> [...]] -
   called by cmd_mode in client-commands.c */
CLIENT_COMMAND_EXT(ccmd_cmode)
{
        SendOptions sopts;
	Channel *channel;
	ChannelNick *channick;
	const char *mode, *mode_args;
        int noprivileges;

	channel = client_find_channel_noisy(client, args[0]);
	if (channel == NULL)
		return;

	if (CHANNEL_IS_MODELESS(channel)) {
                /* modeless channel */
		client_send_numeric(client, ERR_NOCHANMODES, channel->name);
                return;
	}

	channick = channel_get_nick(channel, &client->nick);

	if (argc == 1) {
		/* just querying the mode */
                mode = channel_get_mode(channel, &mode_args);
		if (channick == NULL) {
			/* client is not joined to channel, allow asking for
			   mode but don't show the key or limit */
			mode_args = "";
		}
		client_send_numeric(client, RPL_CHANNELMODEIS,
				    channel->name, mode, mode_args);
                return;
	}

        /* not in channel / not opped / restricted */
	noprivileges = channick == NULL ||
		(channick->status & CSTATUS_OP) == 0 ||
		(client->nick.mode & UMODE_RESTRICTED) != 0;

	channel_modes_parse(channel, client,
			    send_origin_nick(&sopts, &client->nick), NULL,
			    FALSE, noprivileges, args[1], args+2);
}

/* TOPIC <channel>[,...] [?] [<topic>] */
CLIENT_COMMAND(cmd_topic)
{
        SendOptions sopts;
	Channel *channel;
        ChannelNick *channick;
	char *channels, *channame;
	const char *data;

	send_origin_nick(&sopts, &client->nick);

	if (argc > 1 && strlen(args[argc-1]) > channel_max_topic) {
		/* limit the topic size */
                ((char *) args[argc-1])[channel_max_topic] = '\0';
	}

	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		channel = ircnet_find_channel(client->server->network,
					      channame);
		channick = channel == NULL ? NULL :
                        client_find_channick_noisy(client, channel);
		if (channick == NULL)
                        continue;

		if (CHANNEL_IS_MODELESS(channel)) {
                        /* no topic in modeless channels */
			client_send_numeric(client, ERR_NOCHANMODES,
					    channel->name);
			continue;
		}

		if (argc == 1) {
			/* querying the topic */
			if (channel->topic == NULL) {
				client_send_numeric(client, RPL_NOTOPIC,
						    channel->name);
			} else {
				client_send_numeric(client, RPL_TOPIC,
						    channel->name,
						    channel->topic);
			}
			continue;
		}

                /* changing topic - check that we have needed privileges */
		if ((channel->mode & CMODE_OPTOPIC) &&
		    (channick->status & CSTATUS_OP) == 0) {
                        continue;
		}

		p_strdup_replace(channel->pool, &channel->topic, args[argc-1]);

		data = t_strdup_printf("TOPIC %s :%s", channel->name,
				       args[argc-1]);
		channel_send(channel, &sopts, data);
	}
}

static void who_send(Client *client, Nick *nick,
		     Channel *channel, ChannelNick *channick)
{
	char status[10];
        int pos;

	/* status = <G|H> [*] [@|+] */

        /* G = gone, H = here */
	status[0] = (nick->mode & UMODE_AWAY) ? 'G' : 'H';
	pos = 1;

        /* * = IRC operator (+o / +O) */
	if (nick->mode & UMODE_IRCOP) {
		status[pos++] = '*';
	}

        /* @ = ops, + = voices */
	if (channick != NULL) {
		if (channick->status & CSTATUS_OP)
			status[pos++] = '@';
		else if (channick->status & CSTATUS_VOICE)
			status[pos++] = '+';
	}
	status[pos] = '\0';

	client_send_numeric(client, RPL_WHOREPLY, channel->name, nick->user,
			    nick->host, nick->server->name, nick->nick,
			    status, nick->hopcount, nick->real_name);
}

typedef struct {
        Client *client;
	Channel *channel;
        int samechannel;
} WhoChannelData;

static int who_send_channel_nick(Nick *nick, ChannelNick *channick,
				 WhoChannelData *data)
{
	if (data->samechannel || (channick->nick->mode & UMODE_INVISIBLE) == 0)
		who_send(data->client, nick, data->channel, channick);

        return TRUE;
}

static void who_send_channel(Client *client, Channel *channel)
{
        WhoChannelData data;

	data.client = client;
        data.channel = channel;
	data.samechannel = channel_get_nick(channel, &client->nick) != NULL;

	tree_traverse(channel->nicks,
		      (TreeTraverseFunc) who_send_channel_nick, &data);
}

/* WHO <mask>[,...] [o] */
CLIENT_COMMAND(cmd_who)
{
        Channel *channel;
	char *masks, *mask;

	if (args[0] == NULL || args[0][0] == '\0')
                args[0] = "*";

	masks = (char *) t_strdup(args[0]);
	for (mask = i_strtoken(&masks, ','); mask != NULL;
	     mask = i_strtoken(&masks, ',')) {
		channel = !is_chanprefix_char(mask[0]) ? NULL :
			ircnet_find_channel(client->server->network, mask);
		if (channel != NULL) {
			who_send_channel(client, channel);
                        continue;
		}

                /* FIXME */
	}

        client_send_numeric(client, RPL_ENDOFWHO, args[0]);
}

/* NAMES <channel>[,...] */
CLIENT_COMMAND(cmd_names)
{
	Channel *channel;

	channel = ircnet_find_channel(client->server->network, args[0]);
	if (channel == NULL) {
		client_send_numeric(client, ERR_NOSUCHCHANNEL, args[0]);
		return;
	}

        command_reply_names(client, channel);
}

void client_channel_commands_init(void)
{
        client_command_register("JOIN", 1, cmd_join);
        client_command_register("PART", 1, cmd_part);
        client_command_register("KICK", 2, cmd_kick);
        client_command_register("INVITE", 2, cmd_invite);
        client_command_register("TOPIC", 1, cmd_topic);
        client_command_register("WHO", 1, cmd_who);
        client_command_register("NAMES", 1, cmd_names);
}

void client_channel_commands_deinit(void)
{
        client_command_unregister("JOIN", cmd_join);
        client_command_unregister("PART", cmd_part);
        client_command_unregister("KICK", cmd_kick);
        client_command_unregister("INVITE", cmd_invite);
        client_command_unregister("TOPIC", cmd_topic);
        client_command_unregister("WHO", cmd_who);
        client_command_unregister("NAMES", cmd_names);
}
