#include "common.h"
#include "ioloop.h"
#include "server-connection.h"
#include "client-commands.h"
#include "command-replies.h"
#include "channel.h"
#include "iauth-connection.h"

void client_channel_commands_init(void);
void client_channel_commands_deinit(void);

static Pool pool;
static CommandList *client_commands;

void client_command_register(const char *name, int min_params,
			     ClientCommandFunc func)
{
        command_register(client_commands, name, min_params, func);
}

void client_command_unregister(const char *name, ClientCommandFunc func)
{
        command_unregister(client_commands, name, func);
}

void client_command_parse(Client *client, const char *name,
			  int argc, const char *args[])
{
        Command *cmd;
	ClientCommandFunc func;

        cmd = command_find(client_commands, name);
	if (cmd == NULL) {
		client_send_numeric(client, ERR_UNKNOWNCOMMAND, name);
                return;
	}

	if (cmd->min_params > 0 && args[cmd->min_params-1] == NULL) {
		client_send_numeric(client, ERR_NEEDMOREPARAMS, name);
                return;
	}
	func = cmd->func;

	func(client, name, argc, args);
}

Nick *client_find_nick_noisy(Client *client, const char *name)
{
	Nick *nick;

	nick = ircnet_find_nick(client->server->network, name);
	if (nick == NULL)
		client_send_numeric(client, ERR_NOSUCHNICK, name);

        return nick;
}

/* QUIT [?] [<reason>] */
CLIENT_COMMAND(cmd_quit)
{
	const char *reason;

	/* if client is banned in any channel, don't show the quit.
	   user quit reasons are inside quotes */
	reason = argc > 0 && client_get_banned_channel(client) == NULL ?
                t_strconcat("\"", args[argc-1], "\"", NULL) : client->nick.nick;

	client_disconnect(client, reason, TRUE, TRUE);
}

/* KILL <nick> [?] <reason> */
CLIENT_COMMAND(cmd_kill)
{
        SendOptions sopts;
	Nick *nick;

	if ((client->nick.mode & UMODE_LOCAL_IRCOP) == 0) {
		client_send_numeric_noargs(client, ERR_NOPRIVILEGES);
		return;
	}

        nick = client_find_nick_noisy(client, args[0]);
	if (nick == NULL)
		return;

	if (nick->server->local) {
                /* local kill */
		client_kill(client_find(nick->nick), args[1]);
                return;
	}

        /* remote kill */
	if (!nick->server->local && (client->nick.mode & UMODE_IRCOP) == 0) {
                /* requires +o */
		client_send_numeric_noargs(client, ERR_NOPRIVILEGES);
		return;
	}

	/* notify other servers */
	server_send_nick_update(nick, send_origin_nick(&sopts, &client->nick),
		t_strdup_printf("KILL %s :%s",
				nick_get_outputid(nick), args[1]));

        /* notify our clients of the kill with a QUIT message */
	client_send_nick_update(nick, send_origin_nick(&sopts, nick),
        	t_strconcat("QUIT :Killed ", args[1], NULL));

        server_remove_nick(nick);
}

/* NICK <new_nick> */
CLIENT_COMMAND(cmd_nick)
{
	SendOptions sopts;
        Channel *ban_channel;
	Nick *nick;

	if (!nick_is_valid(args[0])) {
		client_send_numeric(client, ERR_ERRONEOUSNICKNAME, args[0]);
		return;
	}

	if (strlen(args[0]) > client_nick_max_length)
                ((char *) args[0])[client_nick_max_length] = '\0';

	nick = ircnet_find_nick_name(client->server->network, args[0]);
	if (nick != NULL && nick != &client->nick) {
		client_send_numeric(client, ERR_NICKNAMEINUSE, args[0]);
                return;
	}

        ban_channel = client_get_banned_channel(client);
	if (ban_channel != NULL) {
		/* we can't change nick while banned. Use the same
		   numeric as bahamut */
		client_send_numeric(client, ERR_BANNEDNICKCHANGE,
				    ban_channel->name);
                return;
	}

	if (nick == NULL || strcmp(nick->nick, args[0]) != 0) {
		send_nick_update(&client->nick,
                                 send_origin_nick(&sopts, &client->nick),
				 t_strconcat("NICK ", args[0], NULL));

		nick_change(&client->nick, args[0], ioloop_time);
	}
}

static void channel_send_data(Client *client, SendOptions *sopts,
			      const char *channame, const char *data)
{
        Channel *channel;
	ChannelNick *channick;

	channel = ircnet_find_channel(client->server->network, channame);
	if (channel == NULL) {
		if (!client->server->leaf) {
			client_send_numeric(client, ERR_NOSUCHNICK, channame);
			return;
		}

		/* we're a leaf server and don't know if this channel exists
		   or not, so just forward the message to our uplink */
		server_send(client->server->uplink, sopts, data);
		return;
	}

        channick = channel_get_nick(channel, &client->nick);
	if (!NICK_CAN_TALK(channel, channick)) {
		client_send_numeric(client, ERR_CANNOTSENDTOCHAN, channame);
                return;
	}

	channel_send(channel, sopts, data);
}

/* PRIVMSG <target>[,...] [?] <msg>
   NOTICE <target>[,...] [?] <msg> */
CLIENT_COMMAND(cmd_talk)
{
	SendOptions sopts;
	Nick *nick;
	const char *data, *reason;
	char *targets, *target;
        int privmsg, count;

        privmsg = strcmp(cmd, "PRIVMSG") == 0;
	if (privmsg) {
                /* privmsg - reset idle time */
		client->last_privmsg = ioloop_time;
                count = client_max_msg_targets;
	} else {
                /* notice */
                count = client_max_notice_targets;
	}

	send_origin_nick(&sopts, &client->nick);
	sopts.skip_client = client;
	sopts.channel_send_talk = TRUE;

	/* FIXME: check the TODO from server-commands.c, we should probably
	   combine this code with it at least partly.. */
	targets = (char *) args[0];
	for (target = i_strtoken(&targets, ','); target != NULL && count > 0;
	     target = i_strtoken(&targets, ','), count--) {
		if (is_chanprefix_char(target[0])) {
			data = t_strconcat(cmd, " ", target,
					   " :", args[1], NULL);
			channel_send_data(client, &sopts, target, data);
		} else {
			nick = client_find_nick_noisy(client, target);
			if (nick == NULL)
				continue;

			/* send to nick */
			data = t_strdup_printf("%s %s :%s", cmd,
					       nick_get_outputid(nick),
					       args[1]);
			nick_send_data(nick, &sopts, data);

			if (privmsg && (nick->mode & UMODE_AWAY)) {
				/* user is away, notify */
				reason = nick->away_message != NULL ?
					nick->away_message :
					client_default_away_reason;
				client_send_numeric(client, RPL_AWAY,
						    nick->nick, reason);
			}
		}
	}
}

CLIENT_COMMAND_EXT(ccmd_cmode); /* in client-channel-commands.c */

/* MODE <nick> [<modes> [<params>]] */
CLIENT_COMMAND(cmd_umode)
{
	Nick *nick;

        nick = client_find_nick_noisy(client, args[0]);
	if (nick == NULL)
                return;

	if (nick != &client->nick) {
                /* nick argument and client's nick must match */
		client_send_numeric_noargs(client, ERR_USERSDONTMATCH);
                return;
	}

	if (argc == 1) {
		/* just querying the mode */
		client_send_numeric(client, RPL_UMODEIS,
				    nick_get_umode_str(nick));
                return;
	}

	/* FIXME: add checks.. */
        /*nick_set_umode(nick, args[1]);*/
}

/* MODE <nick/channel>[,...] [<modes> [<params>]] */
CLIENT_COMMAND(cmd_mode)
{
	char *names, *name;

	names = (char *) args[0];
	for (name = i_strtoken(&names, ','); name != NULL;
	     name = i_strtoken(&names, ',')) {
		if (is_chanprefix_char(name[0])) {
                        /* channel mode change */
			ccmd_cmode(client, cmd, argc, args);
		} else {
                        /* user mode change */
			cmd_umode(client, cmd, argc, args);
		}
	}
}

/* OPER <nick> <password> */
CLIENT_COMMAND(cmd_oper)
{
	const char *data;

	/* FIXME: do proper oper checking */
	if (strstr(client->ipmask, "!127.0.0.1") != 0) {
		client_send_numeric_noargs(client, ERR_NOOPERHOST);
                return;
	}

        client->nick.mode |= UMODE_IRCOP;

	/* notify client */
	data = t_strdup_printf("MODE %s +o", client->nick.nick);
	client_send(client, NULL, data);
	client_send_numeric_noargs(client, RPL_YOUREOPER);

        /* notify servers */
	data = t_strdup_printf("MODE %s +o", nick_get_outputid(&client->nick));
        server_send_all(client->server->network, NULL, data);
}


/* AWAY [<reason>] */
CLIENT_COMMAND(cmd_away)
{
	const char *data;

	if (is_empty_str(args[0])) {
		/* unaway */
		if ((client->nick.mode & UMODE_AWAY) == 0)
                        return;

		client->nick.mode &= ~UMODE_AWAY;
		p_free_and_null(client->server->pool,
				client->nick.away_message);

		/* notify client */
                client_send_numeric_noargs(client, RPL_UNAWAY);

                /* notify servers */
		data = t_strdup_printf("MODE %s -a",
				       nick_get_outputid(&client->nick));
		server_send_all(client->server->network, NULL, data);
	} else {
                /* away */
		if (client->nick.mode & UMODE_AWAY)
                        return;

		client->nick.mode |= UMODE_AWAY;
                p_free(client->server->pool, client->nick.away_message);
		client->nick.away_message =
			p_strndup(client->server->pool, args[0],
				  client_max_away_msg);

		/* notify client */
                client_send_numeric_noargs(client, RPL_NOWAWAY);

                /* notify servers */
		data = t_strdup_printf("MODE %s +a",
				       nick_get_outputid(&client->nick));
		server_send_all(client->server->network, NULL, data);
	}
}

/* PING <origin> [<destination>] */
CLIENT_COMMAND(cmd_ping)
{
        SendOptions sopts;
	Server *server;

	/* always ignore origin */
	if (args[1] == NULL || *args[1] == '\0' ||
	    server_match(client->server->network->server, args[1])) {
		/* answer immediately */
		client_send(client, NULL, t_strdup_printf("PONG %s :%s",
							  client->server->name,
							  client->nick.nick));
	} else {
		/* forward to given server */
		server = ircnet_find_server_name(client->server->network,
						 args[1]);
		if (server == NULL) {
			client_send_numeric(client, ERR_NOSUCHSERVER, args[1]);
			return;
		}

		server_send(server, send_origin_nick(&sopts, &client->nick),
			    t_strdup_printf("PING %s :%s", client->nick.nick,
					    server->name));
	}
}

/* PONG <origin> [<destination>] */
CLIENT_COMMAND(cmd_pong)
{
        SendOptions sopts;
	Server *server;

	/* always ignore origin */
        client->ping_sent = FALSE;
	if (argc < 2)
		server = client->server;
        else {
		/* allow destination to be only server, never a nick
		   (other ircds do allow it, accidentally I believe) */
		server = ircnet_find_server_name(client->server->network,
						 args[1]);
		if (server == NULL) {
			client_send_numeric(client, ERR_NOSUCHSERVER, args[1]);
			return;
		}
	}

	if (!server->local) {
		server_send(server, send_origin_nick(&sopts, &client->nick),
			    t_strdup_printf("PONG %s :%s", client->nick.nick,
					    server->name));
	} else {
		/* server replied to our PONG */
                client->ping_sent = FALSE;
	}
}

/* Returns FALSE if the specified server name was our own server */
static int forward_to_server(Client *client, const char *name,
			     const char *cmd, int argc, const char *args[],
			     int wanted_args, int use_last)
{
        SendOptions sopts;
	Server *server;
        Nick *nick;

        /* forward to remote server */
	server = ircnet_find_server_name(client->server->network, name);

	if (server == NULL && is_mask(name)) {
                /* allow wildcards */
		server = ircnet_find_server_wildcard(client->server->network,
						     name);
	}

	if (server == NULL) {
		/* allow also using nick as the server - it's then forwarded
		   to the server where the nick is connected.

		   ircnet ircd allows wildcarded nicks as well, but that just
		   slows things down and it's useless IMHO. */
		nick = ircnet_find_nick_name(client->server->network, name);
		if (nick != NULL)
                        server = nick->server;
	}

	if (server == NULL) {
		client_send_numeric(client, ERR_NOSUCHSERVER, name);
		return TRUE;
	}

	if (server->local)
                return FALSE;

	server_send(server, send_origin_nick(&sopts, &client->nick),
		    CLIENT_CURRENT_LINE(wanted_args, use_last));
        return TRUE;
}

/* CONNECT <target server> <port> [<remote server>] */
CLIENT_COMMAND(cmd_connect)
{
        SendOptions sopts;
	Server *server;
        int port;

	if ((client->nick.mode & UMODE_LOCAL_IRCOP) == 0) {
		/* requires +o / +O
		   FIXME: what about connecting remote servers, needs +o? */
		client_send_numeric_noargs(client, ERR_NOPRIVILEGES);
		return;
	}

	if (argc >= 3) {
		/* forward request to another server */
		server = ircnet_find_server_name(client->server->network,
						 args[2]);
		if (server == NULL) {
			client_send_numeric(client, ERR_NOSUCHSERVER, args[2]);
			return;
		}

		server_send(server, send_origin_nick(&sopts, &client->nick),
                            CLIENT_CURRENT_LINE(3, FALSE));
		return;
	}

	/* connect to specified server ourself */
        port = atoi(args[1]);
	if (!is_numeric(args[1], '\0') || port <= 0 || port > 65535) {
		client_send(client, NULL, "Connect: illegal port number");
                return;
	}

	iauth_connect_server(args[0], port);
}

/* WHOIS [<target server>] <nick>[,...] */
CLIENT_COMMAND(cmd_whois)
{
        SendOptions sopts;

	/* I don't think we need to support masks in nicks, who would really
	   want to do it? and since it would work for only non-invisible
	   users there's not much point in it besides eating CPU. */
	if (argc < 2 ||
	    !forward_to_server(client, args[0], cmd, argc, args, 2, FALSE)) {
                command_reply_whois(send_to_client(&sopts, client),
				    client->server, args[0], FALSE);
	}
}

/* WHOWAS <nick>,... [<count> [<target>]] */
CLIENT_COMMAND(cmd_whowas)
{
        SendOptions sopts;

	if (argc < 3 ||
	    !forward_to_server(client, args[2], cmd, argc, args, 3, FALSE)) {
                command_reply_whowas(send_to_client(&sopts, client),
				     client->server, args[0], args[1]);
	}
}

/* VERSION [<target server>] */
CLIENT_COMMAND(cmd_version)
{
        SendOptions sopts;

	if (argc < 1 ||
	    !forward_to_server(client, args[0], cmd, argc, args, 1, FALSE)) {
		command_reply_version(send_to_client(&sopts, client),
				      client->server);
	}
}

/* MOTD [<target server>] */
CLIENT_COMMAND(cmd_motd)
{
        SendOptions sopts;

	if (argc < 1 ||
	    !forward_to_server(client, args[0], cmd, argc, args, 1, FALSE)) {
		command_reply_motd(send_to_client(&sopts, client),
				   client->server);
	}
}

/* LUSERS [<mask> [<target server>]] */
CLIENT_COMMAND(cmd_lusers)
{
        SendOptions sopts;

	if (argc < 2 ||
	    !forward_to_server(client, args[1], cmd, argc, args, 2, FALSE)) {
		command_reply_lusers(send_to_client(&sopts, client),
				     client->server, args[0]);
	}
}

/* LINKS [[<target server>] <mask>] */
CLIENT_COMMAND(cmd_links)
{
        SendOptions sopts;

	if (argc < 2 ||
	    !forward_to_server(client, args[0], cmd, argc, args, 2, FALSE)) {
		/* local reply */
		command_reply_links(send_to_client(&sopts, client),
				    client->server, args[0]);
	}
}

/* ISON <nicks> */
CLIENT_COMMAND(cmd_ison)
{
        Nick *nick;
        char *nicks, *name, buffer[CLIENT_MAX_LINE_LENGTH+1];
	int len, size;

	size = sprintf(buffer, s_RPL_ISON, client->nick.nick);

        nicks = (char *) args[0];
	for (name = i_strtoken(&nicks, ' '); name != NULL;
	     name = i_strtoken(&nicks, ' ')) {
		nick = ircnet_find_nick_name(client->server->network, name);
		if (nick != NULL) {
                        len = strlen(nick->nick);
			if (size+len >= sizeof(buffer))
				break;

			memcpy(buffer+size, nick->nick, len);
			buffer[size+len] = ' ';

			size += len+1;

		}
	}
        buffer[size] = '\0';

        client_send(client, NULL, buffer);
}

void client_commands_init(void)
{
	pool = pool_create("Client commands", 1024, TRUE);
        client_commands = command_list_create(pool);

        /* commands sending data to other servers */
        client_command_register("QUIT", 0, cmd_quit);
        client_command_register("KILL", 2, cmd_kill);
        client_command_register("NICK", 1, cmd_nick);
        client_command_register("PRIVMSG", 2, cmd_talk);
        client_command_register("NOTICE", 2, cmd_talk);
        client_command_register("MODE", 1, cmd_mode);
        client_command_register("OPER", 2, cmd_oper);
        client_command_register("AWAY", 0, cmd_away);

        /* commands containing server target parameter */
	client_command_register("PING", 0, cmd_ping);
	client_command_register("PONG", 0, cmd_pong);
        client_command_register("CONNECT", 2, cmd_connect);
        client_command_register("WHOIS", 1, cmd_whois);
        client_command_register("WHOWAS", 1, cmd_whowas);
	client_command_register("VERSION", 0, cmd_version);
	client_command_register("MOTD", 0, cmd_motd);
	client_command_register("LUSERS", 0, cmd_lusers);
	client_command_register("LINKS", 0, cmd_links);

        /* commands creating only replies back to client */
	client_command_register("ISON", 1, cmd_ison);

	client_channel_commands_init();
}

void client_commands_deinit(void)
{
	client_channel_commands_deinit();

	client_command_unregister("QUIT", cmd_quit);
        client_command_unregister("KILL", cmd_kill);
        client_command_unregister("NICK", cmd_nick);
	client_command_unregister("PRIVMSG", cmd_talk);
        client_command_unregister("NOTICE", cmd_talk);
        client_command_unregister("MODE", cmd_mode);
        client_command_unregister("OPER", cmd_oper);
        client_command_unregister("AWAY", cmd_away);

	client_command_unregister("PING", cmd_ping);
	client_command_unregister("PONG", cmd_pong);
        client_command_unregister("CONNECT", cmd_connect);
        client_command_unregister("WHOIS", cmd_whois);
	client_command_unregister("WHOWAS", cmd_whowas);
	client_command_unregister("VERSION", cmd_version);
	client_command_unregister("MOTD", cmd_motd);
	client_command_unregister("LUSERS", cmd_lusers);
	client_command_unregister("LINKS", cmd_links);

	client_command_unregister("ISON", cmd_ison);

        command_list_destroy(client_commands);
        pool_unref(pool);
}
