#include "common.h"
#include "client-class.h"
#include "command-replies.h"
#include "client-login.h"
#include "channel.h"
#include "version.h"

/* Make sure the pass/nick/user is correct and login the user. */
static void client_check_login(Client *client, char *nick)
{
	ClientClass *cclass;
        SendOptions sopts;
	const char *check_nick;

	check_nick = nick != NULL ? nick : t_strdup(client->nick.nick);
	if (!nick_is_valid(check_nick)) {
                /* invalid nick */
		if (nick == NULL) {
			if (client->nick.nick[0] != '\0')
				strcpy(client->nick.nick, "*");
			else {
				p_strdup_replace(client->server->pool,
						 &client->nick.nick, "*");
			}
		}
		client_send_numeric(client, ERR_ERRONEOUSNICKNAME, check_nick);
                return;
	}

	if (strlen(check_nick) > client_nick_max_length) {
		if (nick != NULL)
			nick[client_nick_max_length] = '\0';
                else
			client->nick.nick[client_nick_max_length] = '\0';
	}

	if (hash_lookup(client->server->network->nicks, check_nick) != NULL) {
		/* nick already exists */
		client_send_numeric(client, ERR_NICKNAMEINUSE, check_nick);
                return;
	}

	if (nick != NULL) {
		p_strdup_replace(client->server->pool,
				 &client->nick.nick, nick);
	}

        /* set the client limits from connection class */
	cclass = client_get_class(client);

        client->ping_frequency = cclass->ping_frequency;
	client->inbuf_size = cclass->inbuf_size;
	client->max_outbuf_size = cclass->max_outbuf_size;
	tbuffer_destroy(client->inbuf); client->inbuf = NULL;
        tbuffer_destroy(client->outbuf); client->outbuf = NULL;

        client_login_deinit(client);

        /* update nickmask/ipmask */
        client_set_nick(client, client->nick.nick);

	/* send the login events */
	client_send_numeric(client, RPL_WELCOME, client->nickmask);
	client_send_numeric(client, RPL_YOURHOST, client->server->name,
			    IRCD_VERSION);
	client_send_numeric(client, RPL_CREATED, IRCD_VERSION_DATE_STRING);
	client_send_numeric(client, RPL_MYINFO, client->server->name,
			    IRCD_VERSION, USER_MODES_LIST, CHANNEL_MODES_LIST);

	/* set ourself logged in */
        client_set_logged_in(client);

	/* /LUSERS, /MOTD */
        send_to_nick(&sopts, &client->nick);
	command_reply_lusers(&sopts, client->server, NULL);
	command_reply_motd(&sopts, client->server);

        /* set the default mode */
	if (client_default_mode != NULL && *client_default_mode != '\0') {
		client_send(client, send_origin_nick(&sopts, &client->nick),
			    t_strdup_printf("MODE %s :%s", client->nick.nick,
					    client_default_mode));
	}
}

void client_login_init(Client *client)
{
	client->server->network->logging_count++;

	client_check_login(client, NULL);
}

void client_login_deinit(Client *client)
{
        client->server->network->logging_count--;
}

void client_parse_login(Client *client, const char *cmd,
			int argc, const char *args[])
{
	/* we need to parse only NICK/QUIT here - iauth has taken care of
	   everything else */
	if (strcmp(cmd, "NICK") == 0) {
		/* NICK <nick> */
		if (args[0] == NULL)
			client_send_numeric(client, ERR_NEEDMOREPARAMS, cmd);
		else
			client_check_login(client, (char *) args[0]);
	} else if (strcmp(cmd, "QUIT") == 0) {
		/* QUIT [<message>] */
                client_disconnect(client, args[0], TRUE, FALSE);
	} else {
		client_send_numeric_noargs(client, ERR_NOTREGISTERED);
	}
}

void client_logins_init(void)
{
}

void client_logins_deinit(void)
{
}
