#include "common.h"
#include "ioloop.h"

#include "irc.h"
#include "nick.h"
#include "client-login.h"
#include "client-commands.h"
#include "iauth-connection.h"
#include "channel.h"

static HashTable *logged_clients; /* hash from nick name */

static Pool pool;
static Client *latest_logged, *latest_logging;
static int client_notify_counter;
static int nick_id_counter;
static Timeout cmd_exec_to;

#define CLIENT_FLOOD_CONTROL(client, t_now) \
        ((client)->parse_next < (t_now))

static void client_parse_input(Client *client);

Client *client_create(Server *server, int handle, IPADDR *ip,
		      const char *nick, const char *user, const char *host,
		      const char *realname)
{
	Client *rec;

        return_val_if_fail(server != NULL, NULL);
        return_val_if_fail(ip != NULL, NULL);

	rec = p_new(pool, Client, 1);
	rec->pool = pool_create(t_strconcat("Client ", nick, NULL),
				INITIAL_CLIENT_POOL_SIZE, TRUE);
	rec->refcount = 1;
	rec->server = server;
	rec->handle = handle;
	rec->last_data = rec->last_privmsg = rec->login_time = ioloop_time;
	rec->nick.refcount = 2;
	rec->nick.server = server;
        rec->nick.created = ioloop_time;
	memcpy(&rec->nick.ip, ip, sizeof(IPADDR));

        rec->nick.nick = p_strdup(server->pool, nick);
        rec->nick.user = p_strdup(server->pool, user);
        rec->nick.host = p_strdup(server->pool, host);
        rec->nick.real_name = p_strdup(server->pool, realname);

	rec->io = io_add(handle, IO_READ, (IOFunc) client_parse_input, rec);

	if (latest_logging != NULL)
                latest_logging->prev = rec;

        rec->next = latest_logging;
	latest_logging = rec;

	client_login_init(rec);
	return rec;
}

static void client_remove_from_lists(Client *client)
{
	if (client->next != NULL)
		client->next->prev = client->prev;

	if (client->prev != NULL)
		client->prev->next = client->next;
	else {
                if (client->logged_in)
			latest_logged = client->next;
                else
			latest_logging = client->next;
	}

	if (client->logged_in)
		hash_remove(logged_clients, client->nick.nick);
}

static void remove_channel_invites(Client *client)
{
        ClientInvite *invite;

	for (invite = client->invites; invite != NULL; invite = invite->next)
		client_invite_destroy(client, invite);

	client->invites = NULL;
}

void client_ref(Client *client)
{
        client->refcount++;
}

int client_unref(Client *client)
{
	if (--client->refcount > 0)
		return TRUE;

	if (!client->disconnected)
		client_disconnect(client, NULL, TRUE, TRUE);

	if (!client->logged_in)
		client_login_deinit(client);
	else
		server_remove_nick(&client->nick);

	if (client->iauth != NULL) {
		hash_remove(client->iauth->clients,
			    INT_TO_POINTER(client->iauth_uid));
	}

	client_remove_from_lists(client);

	if (client->handle != -1)
		net_disconnect(client->handle);

        if (client->io != NULL)
		io_remove(client->io);

	tbuffer_destroy(client->inbuf);
        tbuffer_destroy(client->outbuf);

	remove_channel_invites(client);

#ifdef POOL_CHECK_LEAKS
	nick_destroy(&client->nick);
#endif

        client->prev = client->next = NULL;

        p_free_clean(client->pool, client->nickmask);
        p_free_clean(client->pool, client->ipmask);
	pool_unref(client->pool);
	p_free(pool, client);
        return FALSE;
}

Client *client_find(const char *nick)
{
	return_val_if_fail(nick != NULL, NULL);

        return hash_lookup(logged_clients, nick);
}

void client_set_logged_in(Client *client)
{
	if (client->logged_in)
		return;

	/* FIXME: make sure the ID isn't already used */
	nick_id_counter++;
	if (nick_id_counter > NICK_MAX_ID)
                nick_id_counter = 0;
	client->nick.output_id =
		p_strdup_printf(client->server->pool, "\r%s%x",
				server_get_id(client->server), nick_id_counter);

        client_remove_from_lists(client);

	if (latest_logged != NULL)
                latest_logged->prev = client;
        client->prev = NULL;
	client->next = latest_logged;
	latest_logged = client;
        hash_insert(logged_clients, client->nick.nick, client);

        server_add_nick(client->server, &client->nick);
        server_send_nickjoins(&client->nick);
	client->logged_in = TRUE;
}

#define CLIENT_MAX_NOWAIT (client_cmd_wait * client_cmd_threshold)

static void client_parse_next_line(Client *client)
{
        const char *cmd, *args[COMMAND_MAX_ARGS+1], *id;
        char *line, *prefix;
        int timestamp, argc;

        /* flood checking */
	if (!CLIENT_FLOOD_CONTROL(client, ioloop_time)) {
		client->cmd_pending = TRUE;
		return;
	}

	if (client->parse_next < ioloop_time - CLIENT_MAX_NOWAIT)
		client->parse_next = ioloop_time - CLIENT_MAX_NOWAIT;

        /* read the line */
	line = (char *) tbuffer_next_line(client->inbuf);
	if (line == NULL || *line == '\0') {
                client->cmd_pending = FALSE;
		return;
	}
	client->cmd_pending = !tbuffer_is_empty(client->inbuf);

        /* add the flood penalty for the command */
	client->parse_next += client_cmd_wait;

        /* parse the line, prefix and timestamp are ignored */
	line = irc_parse_prefix(line, &prefix, &timestamp, &id);
	if (*line == '\0')
                return;

	cmd = irc_parse_command((char *) t_strdup(line), &argc, args);
	if (cmd == NULL)
                return;

	client->current_cmd_id = id;
	client->current_cmd_line = line;

	if (!client->logged_in) {
                /* not logged in yet, use the simplified parser */
		client_parse_login(client, cmd, argc, args);
	} else {
                client_command_parse(client, cmd, argc, args);
	}

	client->current_cmd_id = NULL;
	client->current_cmd_line = NULL;
}

static void client_parse_input(Client *client)
{
	if (client->disconnected) {
		client_unref(client);
                return;
	}

	client->last_data = ioloop_time;

	if (client->inbuf == NULL) {
		client->inbuf = tbuffer_create(client->handle, client->pool,
					       IO_PRIORITY_DEFAULT,
					       client->inbuf_size);
	}

	switch (tbuffer_read(client->inbuf)) {
	case -1:
                /* disconnected */
                client_disconnect_errno(client);
		break;
	case -2:
                /* input buffer is full */
		client_disconnect(client, "Excess flood", TRUE, TRUE);
		break;
	default:
		client_parse_next_line(client);
	}

        if (client->disconnected)
		client_unref(client);
}

void client_send(Client *client, SendOptions *sopts, const char *data)
{
        Nick *nick;
	unsigned char *buffer, *p;
	const char *origin_str;
        char uid[UID_MAX_LENGTH+1];
        int len, space, uidpos;

        return_if_fail(client != NULL);

	if (client->disconnected)
                return;

	if (client->outbuf == NULL) {
		client->outbuf = tbuffer_create(client->handle, client->pool,
						IO_PRIORITY_DEFAULT,
						client->max_outbuf_size);
	}

        origin_str = NULL;
	if (sopts != NULL) {
		if (sopts->origin_nick != NULL)
			origin_str = nick_get_mask(sopts->origin_nick);
		else if (sopts->origin_server != NULL)
			origin_str = sopts->origin_server->name;
		else if (sopts->origin != NULL)
			origin_str = sopts->origin;
	}

	if (origin_str == NULL) {
                /* always show origin - default to our own server name */
		origin_str = client->server->name;
	}

	/* save the line directly to transmit buffer, so if it needs to be
	   buffered there won't be any memcpy()ing. */
	buffer = p = tbuffer_get_space(client->outbuf,
				       CLIENT_MAX_LINE_LENGTH+2);
	if (buffer == NULL) {
		client_disconnect(client, "Output buffer full", TRUE, TRUE);
                return;
	}

	space = CLIENT_MAX_LINE_LENGTH-1;
	*p++ = ':';

        /* add origin */
        len = strlen(origin_str);
	if (space < len)
		return;
	memcpy(p, origin_str, len); p += len;
        space -= len;

	if (client->current_cmd_id != NULL) {
		/* command ID */
		len = strlen(client->current_cmd_id);
		if (space < len+2)
                        return;

		*p++ = ':'; *p++ = ':';
                memcpy(p, client->current_cmd_id, len);
		space -= len+2;
	}

        /* data */
	if (space < 10)
		return;
	*p++ = ' '; space--;

	/* add the actual data - replace all \rUIDs */
	while (space > 0 && *data != '\0') {
		if (*data != '\r') {
			*p++ = *data++;
			space--;
                        continue;
		}

		/* replace UID with nick */
                data++;
		uidpos = 0;
		while (*data != '\0' && *data != ' ' && *data != ',' &&
		       uidpos < UID_MAX_LENGTH) {
                        uid[uidpos++] = *data++;
		}
                uid[uidpos] = '\0';

		nick = ircnet_find_nick_id(client->server->network, uid);
		if (nick == NULL) {
			server_log(client->server,
				   "client_send(): Unknown UID '%s'", uid);
                        return;
		}

                len = strlen(nick->nick);
		if (len > space) {
			server_log(client->server, "client_send(): Nick too "
				   "long '%s'", nick->nick);
                        return;
		}

		memcpy(p, nick->nick, len);
                p += len; space -= len;
	}


	/* CR + LF */
	*p++ = '\r';
        *p++ = '\n';

	if (tbuffer_send_buffer(client->outbuf, (int) (p-buffer)) < 0)
		client_disconnect_errno(client);
}

void client_send_nick_update(Nick *nick, SendOptions *sopts, const char *data)
{
	Channel *channel;
        Client *client;
	int i, j;

	client_notify_counter++;

	/* if the nick isn't in any channels, and it's our local client,
	   we need to notify only itself. if it's in some channels it
	   gets notified with others.. */
	if (parray_is_empty(nick->channels) && nick->server->local) {
		client = client_find(nick->nick);
		client->notify_counter = client_notify_counter;
		if (sopts->skip_client != client)
			client_send(client, sopts, data);
	}

	if (nick->channels == NULL)
                return;

        /* notify clients in all channels the nick has joined */
        parray_freeze(nick->channels);
	for (i = 0; i < nick->channels->last; i++) {
		channel = parray_index(nick->channels, i);

		if (channel == NULL || parray_is_empty(channel->clients))
                        continue;

		/* notify clients in channel */
                parray_freeze(channel->clients);
		for (j = 0; j < channel->clients->last; j++) {
			client = parray_index(channel->clients, j);

			if (client == NULL)
                                continue;

			/* notify the clients only if it hasn't been sent
			   this notify yet */
			if (client->notify_counter != client_notify_counter &&
			    sopts->skip_client != client) {
				client->notify_counter = client_notify_counter;
                                client_send(client, sopts, data);
			}
		}
                parray_thaw(channel->clients);
	}
        parray_thaw(nick->channels);
}

static void client_quit_server_nick(const char *nick_name, Nick *nick,
				    const char *data)
{
	SendOptions sopts;

	client_send_nick_update(nick, send_origin_nick(&sopts, nick), data);
}

static void client_quit_server(Server *server, Server *squit_server)
{
	Server *sub_server;
        const char *data;
        int i;

	data = t_strconcat("QUIT :", squit_server->name, " ",
			   squit_server->uplink->name, NULL);

        /* sent QUIT message of each nick */
	hash_foreach(server->nicks,
		     (HashForeachFunc) client_quit_server_nick,
		     (void *) data);

        /* quit all the servers under this */
        parray_freeze(server->downlinks);
	for (i = 0; i < server->downlinks->last; i++) {
		sub_server = parray_index(server->downlinks, i);
                if (sub_server != NULL)
			client_quit_server(sub_server, squit_server);
	}
        parray_thaw(server->downlinks);
}

void client_send_server_quit(Server *server)
{
        client_quit_server(server, server);
}

void client_kill(Client *client, const char *msg)
{
        SendOptions sopts;
	const char *uid;

	/* Send KILL to servers */
	uid = nick_get_outputid(&client->nick);
	server_send_nick_update(&client->nick, NULL,
				t_strdup_printf("KILL %s :%s", uid, msg));

	/* disconnect the client */
	client_disconnect(client, t_strconcat("Killed: ", msg, NULL),
			  TRUE, FALSE);

        /* notify our clients of the kill with a QUIT message */
	client_send_nick_update(&client->nick,
				send_origin_nick(&sopts, &client->nick),
				t_strconcat("QUIT :Killed: ", msg, NULL));
}

void client_disconnect(Client *client, const char *msg,
		       int notify_client, int notify_others)
{
	SendOptions sopts;

	if (client->disconnected || client->disconnecting)
		return;
        client->disconnecting = TRUE;

	if (msg == NULL)
                msg = client->nick.nick;

        t_push();
	if (notify_client)
		client_send(client, NULL, t_strdup_printf("ERROR :%s", msg));

	client->disconnected = TRUE;
	if (client->logged_in) {
		if (notify_others) {
			send_nick_update(&client->nick,
				send_origin_nick(&sopts, &client->nick),
				t_strconcat("QUIT :", msg, NULL));
		}
	}
        t_pop();
        client->disconnecting = FALSE;
}

void client_disconnect_errno(Client *client)
{
	client_disconnect(client, strerror(errno), FALSE, TRUE);
}

static void client_list_disconnect(Client *clientlist, Server *server,
				   const char *msg)
{
	Client *client, *next;

	for (client = clientlist; client != NULL; client = next) {
		next = client->next;

		if (client->server == server) {
			client_disconnect(client, msg, TRUE, FALSE);
                        client_unref(client);
		}
	}
}

void clients_disconnect(Server *server, const char *msg)
{
        client_list_disconnect(latest_logged, server, msg);
	client_list_disconnect(latest_logging, server, msg);
}

void client_set_nick(Client *client, const char *old_nick)
{
	char ip[MAX_IP_LEN];

	if (client->logged_in)
                hash_remove(logged_clients, old_nick);

        p_free(client->pool, client->nickmask);
	client->nickmask =
		p_strdup_printf(client->pool, "%s!%s@%s", client->nick.nick,
				client->nick.user, client->nick.host);

        net_ip2host(&client->nick.ip, ip);
        p_free(client->pool, client->ipmask);
	client->ipmask =
		p_strdup_printf(client->pool, "%s!%s@%s", client->nick.nick,
				client->nick.user, ip);

	if (client->logged_in)
                hash_insert(logged_clients, client->nick.nick, client);
}

Channel *client_get_banned_channel(Client *client)
{
        ChannelNick *channick;
	int i;

	if (client->nick.channels == NULL)
		return NULL;

	for (i = 0; i < client->nick.channels->last; i++) {
		Channel *channel = parray_index(client->nick.channels, i);

		channick = channel_get_nick(channel, &client->nick);
		if (channick->status & CSTATUS_BANNED)
                        return channel;
	}

        return NULL;
}

void client_invite_destroy(Client *client, ClientInvite *invite)
{
	p_free_clean(client->pool, invite->channel);
	p_free_clean(client->pool, invite->nick);
	p_free_clean(client->pool, invite);
}

static void client_ping(Client *client)
{
	const char *data;

        /* send PING */
	data = t_strdup_printf("PING %s %s", client->server->name,
			       client->nick.nick);
	client_send(client, NULL, data);

	/* reset the last_data so we can calculate
	   the ping time wait with it */
        client->ping_sent = TRUE;
	client->last_data = ioloop_time;
}

#define CLIENT_HAS_PING_TIMEOUTED(client) \
	((client)->ping_sent && \
	 (client)->last_data + client_max_pong_wait < ioloop_time)

#define CLIENT_NEED_PING(client) \
	(!(client)->ping_sent && \
	 (client)->last_data + (client)->ping_frequency < ioloop_time)

static void client_list_commands_exec(Client *clientlist)
{
	Client *client, *next;

	for (client = clientlist; client != NULL; client = next) {
		next = client->next;

		if (client->disconnected)
			client_unref(client);
		else if (client->cmd_pending && client->inbuf != NULL &&
			 CLIENT_FLOOD_CONTROL(client, ioloop_time))
			client_parse_next_line(client);
		else if (CLIENT_HAS_PING_TIMEOUTED(client)) {
			client_disconnect(client, "ping timeout", FALSE, TRUE);
                        client_unref(client);
		} else if (CLIENT_NEED_PING(client))
			client_ping(client);
	}
}

static int clients_commands_exec(void)
{
	/* go through all clients and see if there's commands to be
	   processed in input buffer */
        client_list_commands_exec(latest_logged);
	client_list_commands_exec(latest_logging);
        return 1;
}

static void client_list_deinit(Client *clientlist)
{
	Client *client, *next;
        int leaks;

        leaks = 0;
	for (client = clientlist; client != NULL; client = next) {
		next = client->next;

		if (!client->disconnected)
			leaks++;

		client_unref(client);
	}

	if (leaks > 0)
                i_error("leaked %d non-disconnected clients", leaks);
}

void clients_init(void)
{
	pool = pool_create("Clients", sizeof(Client)*64, TRUE);
        nick_id_counter = time(NULL) & 0x000fffff;
        client_notify_counter = 0;

	cmd_exec_to = timeout_add(500, (TimeoutFunc) clients_commands_exec,
				  NULL);
	logged_clients = hash_create(pool, 0, irc_nick_hash,
				     (HashCompareFunc) irc_nick_cmp);
	latest_logged = NULL;
        latest_logging = NULL;
}

void clients_deinit(void)
{
        client_list_deinit(latest_logged);
	client_list_deinit(latest_logging);

	timeout_remove(cmd_exec_to);

        hash_destroy_clean(logged_clients);
        pool_unref(pool);
}
