#ifndef __CLIENT_H
#define __CLIENT_H

#include "tbuffer.h"
#include "nick.h"
#include "send.h"

typedef struct _ClientInvite {
	struct _ClientInvite *next;

        char *channel;
        char *nick;
} ClientInvite;

struct _Client {
	struct _Client *prev, *next; /* not sorted */
        int refcount;

        IAuth *iauth; /* which iauth was used to accept this client */
        unsigned int iauth_uid;

	Server *server;
	int handle; /* socket handle */
        IO io;

        /* last_privmsg + last_data are same as login_time initially */
        time_t login_time;
	time_t last_privmsg; /* last PRIVMSG command (ie. idle time) */
	time_t last_data; /* last time we received data from client */
	time_t parse_next; /* timestamp when to parse the next user input
		command. may be less than the current time. */

	Nick nick;
        char *nickmask; /* nick!user@host */
        char *ipmask; /* nick!user@ip */
	ClientInvite *invites; /* only for +i channels */
	int notify_counter; /* used to determine if the client has already been
		notified of change in nick in some of client's channels */

	unsigned int kill_if_inbuf_full:1; /* kill clients whose input buffer gets full */
	unsigned int cmd_pending:1; /* commands are pending in input buffer */

	unsigned int disconnecting:1; /* client_disconnected() is being called */
	unsigned int disconnected:1; /* client disconnected */
	unsigned int logged_in:1; /* logged in */
	unsigned int service:1; /* this client is a service */

	unsigned int ping_sent:1; /* PING sent - waiting for PONG */

        int ping_frequency; /* How often to PING silent client */

	unsigned int inbuf_size;
	unsigned int max_outbuf_size; /* 0 = no limit */
        TBuffer *inbuf, *outbuf;

	const char *current_cmd_id; /* command reply ID for currently
		processed command */
	const char *current_cmd_line; /* command line without prefix */

	Pool pool;
};

/* Create a new client connection to specified server. */
Client *client_create(Server *server, int handle, IPADDR *ip,
		      const char *nick, const char *user, const char *host,
		      const char *realname);

void client_ref(Client *client);
/* Unreference the client. If you want to set the client disconnected,
   use client_disconnect(). */
int client_unref(Client *client);

/* Returns the client for the specified nick, if found. */
Client *client_find(const char *nick);
/* Set the client logged in */
void client_set_logged_in(Client *client);

/* Send data to client from specified origin. The data is prefixed with
   :origin and command ID if needed and CR+LF is appended. If origin is
   nick ID, it's replaced to full nick mask. */
void client_send(Client *client, SendOptions *sopts, const char *data);
/* Send data to all servers that need to know about changes in the nick
   (ie. the clients in common channels) */
void client_send_nick_update(Nick *nick, SendOptions *sopts, const char *data);

/* Send needed QUIT messages for clients of disconnected server (netsplit). */
void client_send_server_quit(Server *server);

/* Kill the client with given reason (local kill). */
void client_kill(Client *client, const char *msg);
/* Disconnect the client with specified reason. Client is notified with an
   ERROR message if notify_client is TRUE. Other clients and servers are
   notified with a QUIT message if notify_others is TRUE. */
void client_disconnect(Client *client, const char *msg,
		       int notify_client, int notify_others);
/* Client disconnected because of read/write error.
   Notifies others with a QUIT message. */
void client_disconnect_errno(Client *client);

/* Disconnect all clients in given server. */
void clients_disconnect(Server *server, const char *msg);

/* Updates client structure / lists for nick change. Note that the
   client->nick must be changed before calling this function. */
void client_set_nick(Client *client, const char *old_nick);
/* Returns the first channel where client is banned, or NULL if none */
Channel *client_get_banned_channel(Client *client);

void clients_init(void);
void clients_deinit(void);

/* INTERNAL: */
void client_invite_destroy(Client *client, ClientInvite *invite);

#endif
