#include "common.h"
#include "ioloop.h"
#include "channel.h"
#include "version.h"

void command_reply_version(SendOptions *sopts, Server *server)
{
	const char *target;

	target = SOPTS_GET_OUTPUT_ID(sopts);
	send_numeric_to(sopts, target, RPL_VERSION, IRCD_VERSION_TEXT, "",
			server->name, ircd_flags);
}

void command_reply_motd(SendOptions *sopts, Server *server)
{
	const char *target;

	target = SOPTS_GET_OUTPUT_ID(sopts);
	send_numeric_to_noargs(sopts, target, ERR_NOMOTD);

	/*FIXME:send_numeric_to_noargs(sopts, target, RPL_MOTDSTART, server->name);
	for () {
		send_numeric_to_noargs(sopts, target, RPL_MOTD, text);
	}
	send_numeric_to_noargs(sopts, target, RPL_MOTDEND);*/

}

/* Send list of channels the nick has joined to,
   that the target is allowed to see. */
static void whois_send_channels(SendOptions *sopts, Nick *nick)
{
	NetSendBuffer *send_buf;
	Channel *channel;
        ChannelNick *channick;
	const char *prefix;
        int i, len;

	if (nick->channels == NULL)
		return;

	prefix = t_strdup_printf(s_RPL_WHOISCHANNELS,
				 SOPTS_GET_OUTPUT_ID(sopts), nick->nick, "");
	send_buf = ircnet_sendbuf_create(nick->server->network, sopts,
                                         prefix, CLIENT_MAX_LINE_LENGTH);

        parray_freeze(nick->channels);
	for (i = 0; i < nick->channels->last; i++) {
		channel = parray_index(nick->channels, i);

		if (channel == NULL)
                        continue;

		/* show non-secret/private channels only if the nick who's
                   asking the whois is also in the channel */
		if ((channel->mode & (CMODE_SECRET|CMODE_PRIVATE)) != 0 &&
		    (sopts->target_nick == NULL ||
		     channel_get_nick(channel, sopts->target_nick) == NULL))
                        continue;

                channick = channel_get_nick(channel, nick);
		if (channick == NULL) {
                        /* shouldn't happen */
			i_error("whois_send_channels() : %s vs %s",
				channel->name, nick->nick);
                        continue;
		}

		len = strlen(channel->name);
                if (!ircnet_sendbuf_check(send_buf, len+2))
			ircnet_sendbuf_add(send_buf, " ", 1);

		if (channick->status & CSTATUS_OP)
			ircnet_sendbuf_add(send_buf, "@", 1);
		else if (channick->status & CSTATUS_VOICE)
			ircnet_sendbuf_add(send_buf, "+", 1);

		ircnet_sendbuf_add(send_buf, channel->name, len);
	}
        parray_thaw(nick->channels);

        ircnet_sendbuf_flush(send_buf);
}

void command_reply_whois(SendOptions *sopts, Server *server,
			 const char *nicks, int allow_uids)
{
        Nick *nick;
	const char *target, *nickname, *reason;
        char *tok_nicks;
        int count;

	target = SOPTS_GET_OUTPUT_ID(sopts);
	count = client_max_whois_nicks;

        tok_nicks = (char *) t_strdup(nicks);
	for (nickname = i_strtoken(&tok_nicks, ','); nickname != NULL;
	     nickname = i_strtoken(&tok_nicks, ',')) {
		nick = allow_uids ?
			ircnet_find_nick(server->network, nickname) :
			ircnet_find_nick_name(server->network, nickname);
		if (nick == NULL) {
			send_numeric_to(sopts, target, ERR_NOSUCHNICK,
					nickname);
			continue;
		}

		send_numeric_to(sopts, target, RPL_WHOISUSER,
				nick->nick, nick->user,
				nick->host, nick->real_name);
                whois_send_channels(sopts, nick);
		send_numeric_to(sopts, target, RPL_WHOISSERVER,
				nick->nick, nick->server->name,
				nick->server->info);

		if (nick->mode & UMODE_AWAY) {
			reason = nick->away_message != NULL ?
				nick->away_message :
				client_default_away_reason;
			send_numeric_to(sopts, target, RPL_AWAY,
					nick->nick, reason);
		}

		if (nick->mode & UMODE_IRCOP) {
                        /* +o / +O */
			send_numeric_to(sopts, target, RPL_WHOISOPERATOR,
					nick->nick);
		}

		if (nick->server->local) {
                        Client *client = client_find(nick->nick);
			send_numeric_to(sopts, target, RPL_WHOISIDLE,
					nick->nick,
					ioloop_time - client->last_privmsg);
		}

		if (--count == 0)
                        break;
	}

	send_numeric_to(sopts, target, RPL_ENDOFWHOIS, nicks);
}

void command_reply_whowas(SendOptions *sopts, Server *server,
			  const char *nicks, const char *countstr)
{
        const char *target;
	char *tok_nicks, *nickname;
        int count;

	count = countstr == NULL ? 0 : atoi(countstr);
	target = SOPTS_GET_OUTPUT_ID(sopts);

	tok_nicks = (char *) t_strdup(nicks);
	for (nickname = i_strtoken(&tok_nicks, ','); nickname != NULL;
	     nickname = i_strtoken(&tok_nicks, ',')) {
		/* FIXME: check the nick history */
		send_numeric_to(sopts, target, ERR_WASNOSUCHNICK, nickname);
	}
	send_numeric_to(sopts, target, RPL_ENDOFWHOWAS, nicks);
}

typedef struct {
	const char *mask;

	unsigned int users_count, opers_count;
	unsigned int services_count, servers_count;
} LusersCount;

static int luser_servers_count(const char *name, Server *server,
			       LusersCount *count)
{
	if (mask_match(count->mask, name)) {
                count->users_count += server->users_count;
                count->services_count += server->services_count;
                count->servers_count += server->servers_count;
                count->opers_count += server->operators_count;
	}

        return TRUE;
}

void command_reply_lusers(SendOptions *sopts, Server *server, const char *mask)
{
	IrcNetwork *net;
	LusersCount count;
        const char *target;

	net = server->network;

	if (mask == NULL || *mask == '\0') {
                count.users_count = net->users_count;
                count.services_count = net->services_count;
                count.servers_count = net->servers_count;
                count.opers_count = net->operators_count;
	} else {
                memset(&count, 0, sizeof(count));
		count.mask = mask;

		hash_foreach(net->servers,
			     (HashForeachFunc) luser_servers_count,
                             &count);
	}

	target = SOPTS_GET_OUTPUT_ID(sopts);
	send_numeric_to(sopts, target, RPL_LUSERCLIENT, count.users_count,
			count.services_count, count.servers_count);
	send_numeric_to(sopts, target, RPL_LUSEROP, count.opers_count);

	send_numeric_to(sopts, target, RPL_LUSERUNKNOWN,
			net->logging_count);
	send_numeric_to(sopts, target, RPL_LUSERCHANNELS,
			net->channels_count);
	send_numeric_to(sopts, target, RPL_LUSERME, server->users_count,
			server->services_count, server->services_count);
}

static void server_send_link(SendOptions *sopts, const char *target,
			     Server *server, const char *mask)
{
	if (mask != NULL && !mask_match(mask, server->name))
		return;

        t_push();
	send_numeric_to(sopts, target, RPL_LINKS, server->name,
			server->local ? server->name : server->uplink->name,
			server->hopcount, server->info);
        t_pop();
}

static void send_server_links(SendOptions *sopts, const char *target,
			      Server *server, const char *mask)
{
	PArray downlinks;
        int i;

	server_send_link(sopts, target, server, mask);

	downlinks = server->downlinks;
	if (downlinks == NULL)
		return;

        parray_freeze(downlinks);
	for (i = 0; i < downlinks->last; i++) {
		server = parray_index(downlinks, i);

                if (server != NULL)
			send_server_links(sopts, target, server, mask);
	}
        parray_thaw(downlinks);
}

void command_reply_links(SendOptions *sopts, Server *server, const char *mask)
{
	const char *target;

	target = SOPTS_GET_OUTPUT_ID(sopts);
	send_server_links(sopts, target, server, mask);

	send_numeric_to(sopts, target, RPL_ENDOFLINKS,
			mask == NULL ? "*" : mask);
}

static int client_send_names(Nick *nick, ChannelNick *channick,
			     NetSendBuffer *sendbuf)
{
	char mode;
        int nicklen;

	nicklen = strlen(nick->nick);

	mode = (channick->status & CSTATUS_OP) ? '@' :
		(channick->status & CSTATUS_VOICE) ? '+' : '\0';

	if (!ircnet_sendbuf_check(sendbuf, nicklen + mode != '\0' + 1))
		ircnet_sendbuf_add(sendbuf, " ", 1);
	if (mode != '\0') ircnet_sendbuf_add(sendbuf, &mode, 1);
	ircnet_sendbuf_add(sendbuf, nick->nick, nicklen);
        return TRUE;
}

void command_reply_names(Client *client, Channel *channel)
{
        SendOptions sopts;
        NetSendBuffer *sendbuf;
	const char *prefix;
        char chantype;

	if (channel->mode & CMODE_PRIVATE)
		chantype = '*';
	else if (channel->mode & CMODE_SECRET)
		chantype = '@';
        else
		chantype = '=';

	prefix = t_strdup_printf(s_RPL_NAMREPLY, client->nick.nick, chantype,
				channel->name, "");
	sendbuf = ircnet_sendbuf_create(client->server->network,
					send_to_client(&sopts, client),
                                        prefix, CLIENT_MAX_LINE_LENGTH);

	tree_traverse(channel->nicks, (TreeTraverseFunc) client_send_names,
		      sendbuf);
	ircnet_sendbuf_flush(sendbuf);

	client_send_numeric(client, RPL_ENDOFNAMES, channel->name);
}
