#include "common.h"
#include "server-connection.h"

#define INITIAL_POOL_SIZE 65536 /* NOTE: contains hash tables for all nicks/channels/servers so must be big */

static Pool pool;

IrcNetwork *irc_network_create(const char *name)
{
	IrcNetwork *net;

        net = p_new(pool, IrcNetwork, 1);
	memset(net, 0, sizeof(IrcNetwork));
        net->name = p_strdup(pool, name);

        /* first there's no servers -> no capabilities */
	net->no_caps = CAP_ALL;

	net->servers = hash_create(pool, 30, strcase_hash,
				   (HashCompareFunc) irc_case_cmp);
	net->server_ids = hash_create(pool, 30, irc_id_hash,
				      (HashCompareFunc) irc_id_cmp);
	net->nicks = hash_create(pool, 500, irc_nick_hash,
				 (HashCompareFunc) irc_nick_cmp);
	net->nick_ids = hash_create(pool, 500, irc_id_hash,
				    (HashCompareFunc) irc_id_cmp);
	net->channels = hash_create(pool, 100, irc_channel_hash,
				    (HashCompareFunc) irc_channel_cmp);
	return net;
}

void irc_network_destroy(IrcNetwork *net)
{
        server_destroy(net->server);

	hash_destroy(net->servers);
	hash_destroy(net->server_ids);
	hash_destroy(net->nicks);
	hash_destroy(net->nick_ids);
	hash_destroy(net->channels);

        p_free(pool, net->name);
        p_free(pool, net);
}

void irc_network_disconnect(IrcNetwork *net, const char *msg)
{
	PArray links;
        int i;

        /* disconnect clients */
        clients_disconnect(net->server, msg);

        /* disconnect servers */
	links = ircnet_links(net);
        parray_freeze(links);
	for (i = 0; i < links->last; i++) {
		Server *server = parray_index(links, i);

		if (server != NULL && server->sconn != NULL)
                        server_disconnect(server->sconn, msg);
	}
        parray_thaw(links);
}

Nick *ircnet_find_nick(IrcNetwork *net, const char *id)
{
	Nick *nick;

	nick = hash_lookup(net->nicks, id);
	if (nick != NULL)
		return nick;

        return ircnet_find_nick_id(net, id);
}

Nick *ircnet_find_nick_name(IrcNetwork *net, const char *nick)
{
        return hash_lookup(net->nicks, nick);
}

Nick *ircnet_find_nick_id(IrcNetwork *net, const char *id)
{
        return hash_lookup(net->nick_ids, id);
}

Server *ircnet_find_server(IrcNetwork *net, const char *server)
{
	Server *serverrec;

	serverrec = hash_lookup(net->servers, server);
	if (serverrec != NULL)
		return serverrec;

        return ircnet_find_server_id(net, server);
}

typedef struct {
        const char *mask;
        Server *server;
} ServerFindData;

static void server_find_mask(const char *name, Server *server,
			     ServerFindData *data)
{
	if (mask_match(data->mask, name)) {
		data->server = server;
                hash_foreach_stop();
	}
}

Server *ircnet_find_server_wildcard(IrcNetwork *net, const char *mask)
{
	ServerFindData data;

	data.mask = mask;
        data.server = NULL;
	hash_foreach(net->servers, (HashForeachFunc) server_find_mask,
		     &data);
        return data.server;
}

Server *ircnet_find_server_name(IrcNetwork *net, const char *name)
{
        return hash_lookup(net->servers, name);
}

Server *ircnet_find_server_id(IrcNetwork *net, const char *id)
{
        return hash_lookup(net->server_ids, id);
}

Channel *ircnet_find_channel(IrcNetwork *net, const char *name)
{
        return hash_lookup(net->channels, name);
}

void ircnet_update_caps(IrcNetwork *net)
{
	Server *server;
        PArray links;
        unsigned int caps;
	int i;

	links = ircnet_links(net);
	if (links == NULL)
                return;

	/* update all_caps */
	net->all_caps = CAP_ALL;
	net->no_caps = CAP_ALL;
	for (i = 0; i < links->last; i++) {
		server = parray_index(links, i);

		if (server != NULL) {
			caps = server_conn_find(server)->caps;
			net->all_caps &= caps;
			net->no_caps &= ~caps;
		}
	}
}

void irc_networks_init(void)
{
	pool = pool_create("IrcNetwork", INITIAL_POOL_SIZE, TRUE);
}

void irc_networks_deinit(void)
{
        pool_unref(pool);
}
