#include "common.h"
#include "ioloop.h"
#include "nick.h"
#include "server-connection.h"

Nick *nick_create(Server *server, const char *id, const char *nick_name,
		  const char *user, const char *host, const char *real_name)
{
	Nick *nick;

        return_val_if_fail(server != NULL, NULL);
        return_val_if_fail(id != NULL, NULL);
        return_val_if_fail(nick_name != NULL, NULL);

	nick = p_new(server->pool, Nick, 1);
        nick->refcount = 1;
        nick->server = server;
        nick->created = nick->last_nick_change = ioloop_time;

	nick->nick = p_strdup(server->pool, nick_name);
	nick->user = p_strdup(server->pool, user);
	nick->host = p_strdup(server->pool, host);
	nick->real_name = p_strdup(server->pool, real_name);

	if (nick->server->uids)
		nick->output_id = p_strconcat(server->pool, "\r", id, NULL);
	else {
		/* server doesn't support UIDs, use the nick name */
                nick->output_id = nick->nick;
	}

	server_add_nick(server, nick);
	return nick;
}

void nick_destroy(Nick *nick)
{
	Pool pool;

	pool = nick->server->pool;

        if (nick->output_id != nick->nick)
		p_free(pool, nick->output_id);
        p_free(pool, nick->nick);
        p_free(pool, nick->user);
        p_free(pool, nick->host);
	p_free(pool, nick->real_name);

	p_free(pool, nick->away_message);
        parray_destroy(nick->channels);
}

void nick_ref(Nick *nick)
{
        nick->refcount++;
}

void nick_unref(Nick *nick)
{
	if (--nick->refcount == 0) {
		/* reference counting should make sure we don't get here with
		   local clients. */
		return_if_fail(!nick->server->local);

		nick_destroy(nick);
		p_free(nick->server->pool, nick);
	}
}

void nick_change(Nick *nick, const char *new_nick, time_t timestamp)
{
	char *old_nick;

	if (timestamp == 0)
		timestamp = ioloop_time;
	nick->last_nick_change = timestamp;

        /* change the nick */
	old_nick = nick->nick;
	nick->nick = p_strdup(nick->server->pool, new_nick);
	server_rename_nick(nick, old_nick);

	if (nick->server->local) {
		/* local client */
		client_set_nick(client_find(old_nick), new_nick);
	}

	p_free(nick->server->pool, old_nick);

	if (!nick->server->uids) {
                /* server doesn't support UIDs, so nick == UID */
		nick->output_id = nick->nick;
	}
}

const char *nick_get_umode_str(Nick *nick)
{
	char *p, str[sizeof(USER_MODES_LIST)+1];

	p = str;
	if (nick->mode & UMODE_IRCOP)
		*p++ = 'O';
	else if (nick->mode & UMODE_LOCAL_IRCOP)
                *p++ = 'o';

	if (nick->mode & UMODE_AWAY) *p++ = 'a';
        if (nick->mode & UMODE_INVISIBLE) *p++ = 'i';
        if (nick->mode & UMODE_RESTRICTED) *p++ = 'r';
	if (nick->mode & UMODE_WALLOPS) *p++ = 'w';

        *p = '\0';
        return t_strdup(str);
}

void nick_set_umode(Nick *nick, const char *mode)
{
	int set, modenum;

        set = TRUE;
	while (*mode != '\0') {
		if (*mode == '+')
			set = TRUE;
		else if (*mode == '-')
			set = FALSE;
		else {
                        modenum = 0;
			switch (*mode) {
			case 'a':
				modenum = UMODE_AWAY;
                                break;
			case 'o':
				modenum = UMODE_LOCAL_IRCOP;
                                break;
			case 'O':
				modenum = UMODE_IRCOP;
                                break;
			case 'i':
				modenum = UMODE_INVISIBLE;
                                break;
			case 'r':
				modenum = UMODE_RESTRICTED;
                                break;
			case 'w':
				modenum = UMODE_WALLOPS;
                                break;
			}

                        if (set)
				nick->mode |= modenum;
			else
				nick->mode &= ~modenum;
		}
                mode++;
	}
}

const char *nick_get_mask(Nick *nick)
{
	return t_strconcat(nick->nick, "!", nick->user, "@", nick->host, NULL);
}

const char *nick_mask_fix(const char *mask)
{
	const char *p, *user, *host, *dot;

	user = host = dot = NULL;
	for (p = mask; *p != '\0'; p++) {
		if (*p == '!')
                        user = p;
		else if (*p == '@')
			host = p;
		else if (*p == '.')
                        dot = p;
	}

	if (user != NULL) {
		if (host != NULL) {
			/* full *!*@* mask */
			return mask;
		} else {
			/* nick!user -> nick!user@* */
                        return t_strconcat(mask, "@*", NULL);
		}
	}

	if (host != NULL) {
		/* user@host -> *!user@host */
                return t_strconcat("*!", mask, NULL);
	} else if (dot != NULL) {
		/* host.domain */
		return t_strconcat("*!*@", mask, NULL);
	} else {
		/* nick */
		return t_strconcat(mask, "!*@*", NULL);
	}
}

int nick_is_valid(const char *nick)
{
	if (*nick == '\0' || *nick == '-' || (*nick >= '0' && *nick <= '9'))
		return FALSE;

	while (*nick != '\0') {
		if (!is_nick_char(*nick))
                        return FALSE;
                nick++;
	}

	if (irc_nick_cmp(nick, "anonymous") == 0)
                return FALSE;
        return TRUE;
}

void nick_send_data(Nick *nick, SendOptions *sopts, const char *data)
{
	Client *client;

	if (!nick->server->local) {
                /* remote server */
		server_send(nick->server, sopts, data);
                return;
	}

	/* local delivery */
	client = client_find(nick->nick);
	return_if_fail(client != NULL);

        client_send(client, sopts, data);
}
