#include "common.h"
#include "send.h"
#include "server-commands.h"

struct _NetSendBuffer {
        IrcNetwork *network;
	SendOptions *sopts;

	int prefix_len;
	int pos, size;
        char data[1];
};

SendOptions *send_options_clear(SendOptions *opts)
{
	memset(opts, 0, sizeof(SendOptions));
	return opts;
}

SendOptions *send_origin(SendOptions *opts, const char *str)
{
	memset(opts, 0, sizeof(SendOptions));
	opts->origin = str;
	return opts;
}

SendOptions *send_origin_nick(SendOptions *opts, Nick *nick)
{
	memset(opts, 0, sizeof(SendOptions));
	opts->origin_nick = nick;
        opts->origin_server = nick->server;
	return opts;
}

SendOptions *send_origin_nick_skip(SendOptions *opts, Nick *nick)
{
	memset(opts, 0, sizeof(SendOptions));
	opts->origin_nick = nick;
        opts->origin_server = nick->server;
        opts->skip_server = server_find_first_link(nick->server);
	return opts;
}

SendOptions *send_origin_client_skip(SendOptions *opts, Client *client)
{
	memset(opts, 0, sizeof(SendOptions));
	opts->origin_nick = &client->nick;
        opts->origin_server = client->server;
        opts->skip_client = client;
	return opts;
}

SendOptions *send_origin_server(SendOptions *opts, Server *server)
{
	memset(opts, 0, sizeof(SendOptions));
        opts->origin_server = server;
	return opts;
}

SendOptions *send_origin_nickserver(SendOptions *opts, Nick *nick,
				    Server *server)
{
	memset(opts, 0, sizeof(SendOptions));
	opts->origin_nick = nick;
	opts->origin_server = server;
	return opts;
}

SendOptions *send_skip_server(SendOptions *opts, Server *skip_server)
{
	memset(opts, 0, sizeof(SendOptions));
        opts->skip_server = skip_server;
	return opts;
}

SendOptions *send_to_client(SendOptions *opts, Client *client)
{
	memset(opts, 0, sizeof(SendOptions));
	opts->target_client = client;
        opts->target_nick = &client->nick;
	return opts;
}

SendOptions *send_to_server(SendOptions *opts, Server *server)
{
	memset(opts, 0, sizeof(SendOptions));
        opts->target_server = server;
	return opts;
}

SendOptions *send_to_nick(SendOptions *opts, Nick *nick)
{
	memset(opts, 0, sizeof(SendOptions));
	if (nick->server->local)
		opts->target_client = client_find(nick->nick);
	else
                opts->target_server = nick->server;
        opts->target_nick = nick;
	return opts;
}

SendOptions *send_back(SendOptions *opts, CommandOrigin *origin)
{
	if (origin->nick != NULL)
		return send_to_nick(opts, origin->nick);
        return send_to_server(opts, origin->server);
}

void send_nick_update(Nick *nick, SendOptions *opts, const char *data)
{
        server_send_nick_update(nick, opts, data);
        client_send_nick_update(nick, opts, data);
}

void send_to(SendOptions *sopts, const char *data)
{
	if (sopts->target_client != NULL)
		client_send(sopts->target_client, sopts, data);

	if (sopts->target_server != NULL)
		server_send(sopts->target_server, sopts, data);
}

void send_command_reply(CommandOrigin *origin, const char *data)
{
	if (origin->nick != NULL)
                nick_send_data(origin->nick, NULL, data);
	else
                server_send(origin->server, NULL, data);
}

NetSendBuffer *ircnet_sendbuf_create(IrcNetwork *network, SendOptions *sopts,
				     const char *prefix, int max_line_size)
{
	NetSendBuffer *buf;
        int prefix_len;

        prefix_len = strlen(prefix);
        return_val_if_fail(prefix_len < max_line_size, NULL);

	buf = t_malloc(sizeof(NetSendBuffer) + max_line_size);
	buf->network = network;
	buf->sopts = sopts;

	buf->prefix_len = prefix_len;
        buf->pos = prefix_len;
	buf->size = max_line_size;
	memcpy(buf->data, prefix, prefix_len);

        return buf;
}

void ircnet_sendbuf_flush(NetSendBuffer *buf)
{
	if (buf->pos <= buf->prefix_len)
		return;

	buf->data[buf->pos] = '\0';

	if (buf->sopts->target_client != NULL)
		client_send(buf->sopts->target_client, buf->sopts, buf->data);
	else if (buf->sopts->target_server != NULL)
		server_send(buf->sopts->target_server, buf->sopts, buf->data);
	else
		server_send_all(buf->network, buf->sopts, buf->data);

        buf->pos = buf->prefix_len;
}

int ircnet_sendbuf_check(NetSendBuffer *buf, int need_size)
{
	if (buf->pos + need_size <= buf->size)
		return buf->pos == buf->prefix_len;

	ircnet_sendbuf_flush(buf);
	if (buf->pos + need_size > buf->size) {
                /* shouldn't happen */
		i_panic("ircnet_sendbuf_check_flush(%d) : "
			"buffer too small", need_size);
	}

        return TRUE;
}

void ircnet_sendbuf_add(NetSendBuffer *buf, const char *data, int size)
{
	ircnet_sendbuf_check(buf, size);

	memcpy(buf->data + buf->pos, data, size);
        buf->pos += size;
}

