#include "common.h"
#include "ioloop.h"
#include "server-connection.h"
#include "server-commands.h"
#include "channel.h"
#include "channel-modes.h"
#include "messages.h"

static const char *channel_status_get_string(unsigned int status)
{
	if ((status & CSTATUS_OWNER) == CSTATUS_OWNER) {
                if (status & CSTATUS_VOICE)
			return "Ov";
                else
			return "O";
	}

	if (status & CSTATUS_OP) {
		if (status & CSTATUS_VOICE)
			return "ov";
		else
                        return "o";
	}

	if (status & CSTATUS_VOICE)
		return "v";

        return "";
}

static int find_op(Nick *nick, ChannelNick *channick, int *ret)
{
	if ((channick->status & CSTATUS_OP) == 0)
		return TRUE;

        *ret = TRUE;
        return FALSE;
}

static int channel_have_ops(Channel *channel)
{
	int ret;

        ret = FALSE;
	tree_traverse(channel->nicks, (TreeTraverseFunc) find_op, &ret);
        return ret;
}

/* NJOIN <chan> [?] :[@@|@][+]<uid|nick>[,...] */
SERVER_COMMAND(cmd_njoin)
{
	IrcNetwork *network;
	SendOptions id_sopts, nick_sopts, old_sopts, sopts;
        SendOptions mode_sopts, mode_reverse_sopts;
	Channel *channel;
        Nick *nick;
	ChannelNick *channick;
	NetSendBuffer *id_buf, *nick_buf;
        ModeBuffer *mode_buf, *mode_reverse_buf;

	const char *data, *nickid;
	char *nicks, *p, statusbuf[10];
	unsigned int status;
	int len, nicklen, deny_ops, reverse_op;
        time_t timestamp;

	/* FIXME: handle leaf servers properly */

	timestamp = origin->sconn->current_cmd_timestamp;
        deny_ops = FALSE;

        network = origin->server->network;
	channel = ircnet_find_channel(network, args[0]);
	if (channel == NULL) {
                /* channel doesn't exist yet - create it */
		channel = channel_create(network, args[0]);

                if (timestamp > 0)
			channel->created = timestamp;
	} else if (timestamp-channel->created > channel_ts_max_diff) {
		/* their channel was created later than ours,
		   don't allow any of them to get ops,
		   unless we don't have ops.. */
                deny_ops = channel_have_ops(channel);
	}

	/* create ID and nick send buffers and
	   only if they're used for some server. */
        data = t_strdup_printf("NJOIN %s :", channel->name);

        send_skip_server(&id_sopts, origin->sconn->server);
        id_sopts.timestamp = ioloop_time;
	id_sopts.caps_set = CAP_UID | CAP_NJOIN;
	id_buf = (network->no_caps & CAP_UID) ? NULL :
		ircnet_sendbuf_create(network, &id_sopts, data,
				      network->min_line_length);

        send_skip_server(&nick_sopts, origin->sconn->server);
        nick_sopts.timestamp = ioloop_time;
	nick_sopts.caps_set = CAP_NJOIN;
	nick_sopts.caps_unset = CAP_UID;
	nick_buf = (network->all_caps & CAP_UID) ? NULL :
		ircnet_sendbuf_create(network, &nick_sopts, data,
				      network->min_line_length);

        send_skip_server(&old_sopts, origin->sconn->server);
	old_sopts.caps_unset = CAP_NJOIN;

        /* mode buffer is used to send nick mode changes in bursts to clients */
	send_options_clear(&mode_sopts);
        mode_sopts.channel_no_servers = TRUE;
	mode_buf = mode_buffer_create(channel, &mode_sopts,
				      client_max_mode_args,
				      CLIENT_MAX_LINE_LENGTH);

	/* mode buffer for reversing ops */
	send_to_server(&mode_reverse_sopts, origin->sconn->server);
	mode_reverse_buf = mode_buffer_create(channel, &mode_reverse_sopts,
					      origin->sconn->max_mode_args,
					      origin->sconn->max_line_length);

	/* parse the nicks list */
        nicks = (char *) args[argc-1];
	for (p = i_strtoken(&nicks, ','); p != NULL;
	     p = i_strtoken(&nicks, ',')) {
                len = 0;
		status = 0;

                reverse_op = FALSE;
		if (*p == '@') {
			p++;
			if (deny_ops)
                                reverse_op = TRUE;
			else {
				status |= CSTATUS_OP;
				statusbuf[len++] = '@';
			}
		}
		if (*p == '@') {
			p++;
			if (!deny_ops) {
				status |= CSTATUS_OWNER;
				statusbuf[len++] = '@';
			}
		}
		if (*p == '+') {
			status |= CSTATUS_OP;
			statusbuf[len++] = *p++;
		}
                statusbuf[len] = '\0';

                nick = ircnet_find_nick(network, p);
		if (nick == NULL) {
			/* unknown nick. */
			server_log(origin->server,
				   "Unknown nick %s joined to channel %s",
				   p, channel->name);
                        continue;
		}

		if (channel_get_nick(channel, nick) != NULL) {
                        /* duplicate nick. */
			server_log(origin->server,
				   "Nick %s already joined to channel %s",
				   nick->nick, channel->name);
                        continue;
		}

		if (reverse_op) {
			/* don't let it have ops */
			mode_buffer_append_status(mode_reverse_buf, FALSE,
						  CSTATUS_OP, nick);
		}

		/* add nick to channel */
		channick = channel_add_nick(channel, nick);
		channick->status = status;

		/* send JOIN to our clients */
		data = t_strdup_printf("JOIN %s", channel->name);
		channel_send_clients(channel, send_origin_nick(&sopts, nick),
				     data);

		/* update MODE buffer */
		if (status != 0)
			mode_buffer_append_status(mode_buf, TRUE, status, nick);

		/* update ID buffer */
		if (id_buf != NULL) {
                        /* append [,]<status><uid> */
			nickid = nick_get_id(nick);
			nicklen = strlen(nickid);
			if (!ircnet_sendbuf_check(id_buf, len + nicklen + 1))
				ircnet_sendbuf_add(id_buf, ",", 1);
                        ircnet_sendbuf_add(id_buf, statusbuf, len);
                        ircnet_sendbuf_add(id_buf, nickid, nicklen);
		}

		/* update nick buffer */
		if (nick_buf != NULL) {
                        /* append [,]<status><nick> */
			nicklen = strlen(nick->nick);
			if (!ircnet_sendbuf_check(nick_buf, len + nicklen + 1))
				ircnet_sendbuf_add(nick_buf, ",", 1);
                        ircnet_sendbuf_add(nick_buf, statusbuf, len);
                        ircnet_sendbuf_add(nick_buf, nick->nick, nicklen);
		}

		if ((network->all_caps & CAP_NJOIN) == 0) {
			/* some servers want the old JOIN command */
			old_sopts.origin_nick = nick;
                        data = t_strconcat("JOIN ", nick->nick, "\007",
					   channel_status_get_string(status),
					   NULL);
			server_send_all(network, &old_sopts, data);
		}
	}

	if (id_buf != NULL) ircnet_sendbuf_flush(id_buf);
	if (nick_buf != NULL) ircnet_sendbuf_flush(nick_buf);
        mode_buffer_flush(mode_buf);
        mode_buffer_flush(mode_reverse_buf);
}

/* JOIN 0 [<reason>] - PART all channels */
static void cmd_join_0(Nick *nick, const char *reason)
{
        SendOptions sopts;
	Channel *channel;
        const char *data;
	int i;

	if (parray_is_empty(nick->channels))
                return;

	if (reason == NULL)
		reason = nick->nick;

	/* notify our clients */
        send_origin_nick(&sopts, nick);

	parray_freeze(nick->channels);
	for (i = 0; i < nick->channels->last; i++) {
		channel = parray_index(nick->channels, i);

		if (channel != NULL) {
			data = t_strconcat("PART ", channel->name, " :",
					   reason, NULL);
			channel_send_clients(channel, &sopts, data);
		}
	}
	parray_thaw(nick->channels);

	channels_remove_nick(nick);
}

/* JOIN <channel>[^G<mode>][,...] */
SERVER_COMMAND(cmd_join)
{
        SendOptions sopts;
	IrcNetwork *network;
	Channel *channel;
	ChannelNick *channick;
	ModeBuffer *mode_buf;
	char *channame, *channels, *mode;
        unsigned int status;
        time_t timestamp;

	if (origin->nick == NULL) {
		server_log(origin->sconn->server, "No nick in origin");
		return;
	}

	/* forward first to other servers - we don't really need to
           validate it.. */
        network = origin->server->network;
	server_send_all(network, send_origin_nick_skip(&sopts, origin->nick),
			SERVER_CURRENT_LINE(1, FALSE));

	if (strcmp(args[0], "0") == 0) {
		/* parting all channels */
		cmd_join_0(origin->nick, args[1]);
                return;
	}

        timestamp = origin->sconn->current_cmd_timestamp;

	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		mode = strchr(channame, '\007');
		if (mode != NULL)
			*mode++ = '\0';

                /* parse the mode */
		status = 0;
		if (mode != NULL) {
			while (*mode != '\0') {
				if (*mode == 'O')
					status |= CSTATUS_OWNER|CSTATUS_OP;
				else if (*mode == 'o')
					 status |= CSTATUS_OP;
				else if (*mode == 'v')
                                        status |= CSTATUS_VOICE;
                                mode++;
			}
		}

                channel = ircnet_find_channel(network, channame);
		if (channel == NULL) {
			/* channel doesn't exist yet - create it */
			channel = channel_create(network, channame);
			if (timestamp > 0) channel->created = timestamp;
		}

		if (channel_get_nick(channel, origin->nick) != NULL) {
                        /* duplicate nick. */
			server_log(origin->server,
				   "Nick %s already joined to channel %s",
				   origin->nick->nick, channel->name);
                        continue;
		}

		/* add nick to channel */
		channick = channel_add_nick(channel, origin->nick);
		channick->status = status;

		/* send JOIN to our clients */
		channel_send_clients(channel,
				     send_origin_nick(&sopts, origin->nick),
				     t_strconcat("JOIN ", channel->name, NULL));

		if (status != 0) {
			/* send MODE to our clients */
			send_options_clear(&sopts);
                        sopts.channel_no_servers = TRUE;

			mode_buf = mode_buffer_create(channel, &sopts,
						      client_max_mode_args,
						      CLIENT_MAX_LINE_LENGTH);
			mode_buffer_append_status(mode_buf, TRUE,
						  status, origin->nick);
                        mode_buffer_flush(mode_buf);
		}
	}
}

/* PART <channel>[...] [?] [<reason>] */
SERVER_COMMAND(cmd_part)
{
        SendOptions sopts;
	Channel *channel;
        char *channels, *channame;

	if (origin->nick == NULL) {
		/* FIXME: leaf server wants to drop the channel */
                return;
	}

	send_origin_nick_skip(&sopts, origin->nick);

	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		channel = ircnet_find_channel(origin->server->network,
					      channame);
		if (channel == NULL) {
			send_numeric_reply(origin, ERR_NOSUCHCHANNEL, channame);
                        continue;
		}

		if (channel_get_nick(channel, origin->nick) == NULL) {
			send_numeric_reply(origin, ERR_NOTONCHANNEL,
					   channel->name);
			continue;
		}

		channel_send(channel, &sopts,
			     t_strconcat("PART ", channel->name, " :",
					 args[1], NULL));
		channel_remove_nick(channel, origin->nick);
	}
}

/* MODE <channel> [<modes> [<params> [...]] -
   called by cmd_mode in server-commands.c */
SERVER_COMMAND_EXT(cmd_cmode)
{
        SendOptions sopts, reply_sopts, *reply_sopts_ptr;
	Channel *channel;
	ChannelNick *channick;
	const char *mode, *mode_args, *data;
	int reverse;

	channel = ircnet_find_channel(origin->server->network, args[0]);
	if (channel == NULL) {
		send_numeric_reply(origin, ERR_NOSUCHCHANNEL, args[0]);
		return;
	}

	channick = origin->nick == NULL ? NULL :
		channel_get_nick(channel, origin->nick);

	if (argc == 1) {
		/* just querying the mode */
                mode = channel_get_mode(channel, &mode_args);
		if (channick == NULL) {
			/* this should never happen, really. but if it did,
			   just make sure we don't tell it the channel key */
                        mode_args = "";
		}
		send_numeric_reply(origin, RPL_CHANNELMODEIS,
				   channel->name, mode, mode_args);
                return;
	}

        reverse = FALSE;
	if (channick != NULL && (channick->status & CSTATUS_OP) == 0) {
		/* Here's the part where most of the channel desyncs happen -
		   a non-chanop made a mode change. Normally ircds don't allow
		   this and just drops the change. We'll tell the other side
                   to reverse the change. */
		server_log(origin->server, "Fake: %s %s", origin->nick->nick,
			   origin->sconn->current_cmd_line);
                reverse = TRUE;
	}

	/* parse the modes the server sent to us,
	   forward only the modes that have changed */
	if (reverse) {
		send_to_server(&sopts, origin->sconn->server);
                reply_sopts_ptr = NULL;
	} else {
		reply_sopts_ptr =
			send_to_server(&reply_sopts, origin->sconn->server);
		send_origin_nickserver(&sopts, origin->nick,
				       origin->sconn->server);
		sopts.skip_server = origin->sconn->server;
	}

	if (channel_modes_parse(channel, NULL, &sopts, reply_sopts_ptr,
				reverse, FALSE, args[1], args+2)) {
		if (reverse) {
			/* reversed some mode changes, notify users about it */
			data = t_strdup_printf(REVERSE_DESYNC_MODE_MSG,
					       channel->name,
					       origin->nick->nick);
			server_send(origin->sconn->server, NULL, data);
		}
	}
}

static void kick_nicks(Channel *channel, CommandOrigin *origin,
                       SendOptions *sopts, char *nicks, const char *reason)
{
        SendOptions sopts_nick;
	Nick *nick;
        ChannelNick *channick;
	const char *nickname, *status, *data;
        int desync_kick;

	desync_kick = FALSE;
	if (origin->nick != NULL) {
		channick = channel_get_nick(channel, origin->nick);
		if (channick == NULL || (channick->status & CSTATUS_OP) == 0) {
			/* the kicker is not in channel (shouldn't happen) or
			   isn't channel operator anymore. avoid desyncs by
			   either introducing the nick back, or forwarding
			   the kicks as PARTs. */
			desync_kick = TRUE;
			server_log(origin->server, "Fake: %s %s",
				   origin->nick->nick,
				   origin->sconn->current_cmd_line);
		}
	}

	for (nickname = i_strtoken(&nicks, ','); nickname != NULL;
	     nickname = i_strtoken(&nicks, ',')) {
		nick = ircnet_find_nick(channel->network, nickname);
		if (nick == NULL)
			continue;

		channick = channel_get_nick(channel, nick);
		if (channick == NULL)
			continue;

		if (!desync_kick) {
			/* normal KICK, forward the message and remove nick
                           from channel */
			data = t_strdup_printf("KICK %s %s :%s", channel->name,
					       nickname, reason);
			channel_send(channel, sopts, data);
			channel_remove_nick(channel, nick);
		} else if (nick->server->local ||
			   server_find_first_link(nick->server) !=
			   origin->sconn->server) {
			/* desync kick reverse - nick wasn't behind the
			   server where kick came from, so we'll just
			   JOIN the nick back */
                        status = channel_status_get_string(channick->status);
			data = t_strconcat("JOIN ", channel->name, "\007",
					   status, NULL);
                        send_origin_nick(&sopts_nick, nick);
			server_send(origin->sconn->server, &sopts_nick, data);

			data = t_strdup_printf(REVERSE_DESYNC_KICK_MSG,
					       channel->name, nick->nick);
			server_send(origin->sconn->server, NULL, data);
		} else {
			/* desynced kick - nick already knows it was kicked
			   from the channel, so we'll just forward it as
			   PART to rest of the network */
			data = t_strdup_printf("PART %s :Desync kick by %s: %s",
					       channel->name,
					       nick_get_mask(origin->nick),
                                               reason);

			send_origin_nick(&sopts_nick, nick);
			sopts_nick.skip_server = origin->sconn->server;

			channel_send(channel, &sopts_nick, data);
			channel_remove_nick(channel, nick);
		}
	}
}

/* KICK <channel>[,...] <uid|nick>[,...] [?] [<reason>] */
SERVER_COMMAND(cmd_kick)
{
        SendOptions sopts;
	Channel *channel;
	char *channels, *channame;
	const char *reason;

	send_origin_nickserver(&sopts, origin->nick, origin->server);
	sopts.skip_server = origin->sconn->server;

	reason = argc >= 3 ? args[argc-1] :
		origin->nick != NULL ? origin->nick->nick :
		origin->server->name;

	/* NOTE: Behaviour when giving multiple channels+nicks is not the
	   way RFC-2812 specifies it, but no-one else seems to respect it
	   either. Either way, it's pretty useless. */
	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		channel = ircnet_find_channel(origin->sconn->network, channame);
		if (channel == NULL) {
			send_numeric_reply(origin, ERR_NOSUCHCHANNEL, channame);
                        continue;
		}

		kick_nicks(channel, origin, &sopts,
			   (char *) t_strdup(args[1]), reason);
	}
}

/* INVITE <nick> <channel> */
SERVER_COMMAND(cmd_invite)
{
        SendOptions sopts;
	Nick *nick;
        Channel *channel;
	ChannelNick *channick;

	if (origin->nick == NULL)
                return;

	nick = ircnet_find_nick(origin->server->network, args[0]);
	if (nick == NULL)
		return;

        /* send data to nick */
	send_origin_nick(&sopts, origin->nick);
	nick_send_data(nick, &sopts, SERVER_CURRENT_LINE(2, FALSE));

	if (!nick->server->local)
                return;

	/* local nick, update invite lists if channel is +i */
	channel = ircnet_find_channel(origin->server->network, args[1]);
	if (channel != NULL && (channel->mode & CMODE_INVITEONLY)) {
		/* make sure the inviter is channel operator */
		channick = channel_get_nick(channel, origin->nick);
		if (channick == NULL || (channick->status & CSTATUS_OP) == 0)
                        return;

		channel_add_client_invite(channel, client_find(nick->nick),
                                          origin->nick);
	}
}

/* TOPIC <channel>[,...] [?] [<topic>] */
SERVER_COMMAND(cmd_topic)
{
        SendOptions sopts;
	Channel *channel;
	time_t timestamp;
	char *channels, *channame;
	const char *data, *topic_by;

        timestamp = origin->sconn->current_cmd_timestamp;
	if (timestamp <= 0)
		timestamp = ioloop_time;

	send_origin_nickserver(&sopts, origin->nick, origin->server);
	sopts.skip_server = origin->sconn->server;
        sopts.timestamp = origin->sconn->current_cmd_timestamp;

	topic_by = origin->nick == NULL ? NULL : origin->nick->nick;
	channels = (char *) args[0];
	for (channame = i_strtoken(&channels, ','); channame != NULL;
	     channame = i_strtoken(&channels, ',')) {
		channel = ircnet_find_channel(origin->sconn->network, channame);
		if (channel == NULL) {
			send_numeric_reply(origin, ERR_NOSUCHCHANNEL, channame);
                        continue;
		}

		if (channel->topic_time > timestamp) {
			/* we're getting an old topic from a burst, ignore */
                        continue;
		}

		channel_set_topic(channel, args[argc-1], topic_by,
				  timestamp, FALSE);
		data = t_strdup_printf("TOPIC %s :%s", channel->name,
				       args[argc-1]);
		channel_send(channel, &sopts, data);
	}
}

void server_channel_commands_init(void)
{
        server_command_register("NJOIN", 2, cmd_njoin);
        server_command_register("JOIN", 1, cmd_join);
        server_command_register("PART", 1, cmd_part);
        server_command_register("KICK", 2, cmd_kick);
        server_command_register("INVITE", 2, cmd_invite);
        server_command_register("TOPIC", 2, cmd_topic);
}

void server_channel_commands_deinit(void)
{
        server_command_unregister("NJOIN", cmd_njoin);
        server_command_unregister("JOIN", cmd_join);
        server_command_unregister("PART", cmd_part);
        server_command_unregister("KICK", cmd_kick);
        server_command_unregister("INVITE", cmd_invite);
        server_command_unregister("TOPIC", cmd_topic);
}
