#include "common.h"
#include "ioloop.h"
#include "server-commands.h"
#include "command-replies.h"
#include "server-connection.h"
#include "channel.h"

void server_channel_commands_init(void);
void server_channel_commands_deinit(void);

static Pool pool;
static CommandList *server_commands;

void server_command_register(const char *name, int min_params,
			     ServerCommandFunc func)
{
        command_register(server_commands, name, min_params, func);
}

void server_command_unregister(const char *name, ServerCommandFunc func)
{
        command_unregister(server_commands, name, func);
}

static void server_forward_numeric(CommandOrigin *origin, const char *num,
				   int argc, const char *args[])
{
        const char *data;
	SendOptions sopts;
	Nick *nick;

	if (args[0] == NULL)
		return;

	nick = ircnet_find_nick(origin->server->network, args[0]);
	if (nick == NULL)
		return;

	send_to_nick(&sopts, nick);
	sopts.origin_server = origin->server;

	/* last arg needs to be prefixed with ':' */
	args[argc-1] = t_strconcat(":", args[argc-1], NULL);

	data = t_strjoin_replace((char *const *) args, ' ', 0,
				 nick_get_outputid(nick));
	send_to(&sopts, t_strconcat(num, " ", data, NULL));
}

void server_command_parse(CommandOrigin *origin, const char *name,
			  int argc, const char *args[])
{
        Command *cmd;
	ServerCommandFunc func;

        cmd = command_find(server_commands, name);
	if (cmd == NULL) {
		if (is_numeric(name, '\0') && argc > 0)
                        server_forward_numeric(origin, name, argc, args);

                /* ignore unknown commands */
		return;
	}

	if (cmd->min_params > 0 && args[cmd->min_params-1] == NULL) {
		server_log(origin->sconn->server,
			   "Not enough parameters for command %s", name);
                return;
	}
	func = cmd->func;

	func(origin, name, argc, args);
}

int command_forward_server(CommandOrigin *origin, const char *cmd,
			   const char *args[], int serverarg)
{
        const char *server_name, *data;
	SendOptions sopts;
	Server *server;
        Nick *nick;

        server_name = args[serverarg];
	if (server_name == NULL)
                return TRUE;

	server = ircnet_find_server(origin->server->network, server_name);
	if (server == NULL && server_match(origin->server->network->server,
					   server_name)) {
                /* hostmask was given, that's us. */
		server = origin->server->network->server;
	}

	if (server == NULL && is_mask(server_name)) {
		/* servers shouldn't actually send us wildcarded
		   server names, but allow anyway.. */
		server = ircnet_find_server_wildcard(origin->server->network,
						     server_name);
	}

	if (server == NULL) {
		/* allow also using nick as the server - it's then forwarded
		   to the server where the nick is connected */
		nick = ircnet_find_nick(origin->server->network, server_name);
		if (nick != NULL)
                        server = nick->server;
	}

	if (server == NULL) {
		/* unknown server. */
		send_numeric_reply(origin, ERR_NOSUCHSERVER, server_name);
		return FALSE;
	}

	if (server->local)
                return TRUE;

	/* make sure the server didn't route the command wrong. */
	if (!server_check_noroute(server, origin->sconn->server))
		return FALSE;

	send_origin_nickserver(&sopts, origin->nick, origin->server);
	data = t_strjoin_replace((char *const *) args, ' ', serverarg,
				 server_get_outputid(server));
	server_send(server, &sopts, t_strconcat(cmd, " ", data, NULL));
        return FALSE;
}

/* SERVER <name> <hopcount> <sid|token> [<version> [?]] <info> */
SERVER_COMMAND(cmd_server)
{
        SendOptions sopts;
	Server *server, *server_link, *kill_link;
        const char *sid, *data;
	int hopcount;

	hopcount = atoi(args[1]);
	if (hopcount <= 0) {
		/* invalid hop count */
		server_log(origin->sconn->server,
			   "invalid hop count: %s", args[1]);
                hopcount = 1;
	}

	server = ircnet_find_server(origin->server->network, args[0]);
	server_link = server == NULL ? NULL : server_find_first_link(server);
	if (server_link == origin->sconn->server) {
		/* already exists, update the hopcount/token/info */
		server->hopcount = hopcount;
		p_strdup_replace(server->pool, &server->info, args[argc-1]);
	} else if (server_link != NULL) {
		/* the server was already connected elsewhere,
		   kill the older one. If the timestamps are the same,
		   both servers might get killed, but that was really it's
		   own fault for connecting too fast. Without timestamps,
		   the new server is always killed. */
		kill_link = origin->sconn->current_cmd_timestamp <
			server->created ? origin->sconn->server : server_link;

		server_log(origin->sconn->server,
			   "server %s already connected via %s - "
			   "killing the link via %s",
			   server->name, server_link->name, kill_link->name);

		server_send(kill_link, NULL,
			    t_strdup_printf("SQUIT %s :duplicate", args[0]));
		if (kill_link == origin->sconn->server)
			return;

		/* killed the server in our side of the network */
		server_destroy(server);
		server = NULL;
	}

	if (server == NULL) {
		/* create the new server, take the creation time from
		   command's timestamp if possible */
		sid = (origin->sconn->caps & CAP_UID) && args[2][0] != '$' ?
			args[2] : NULL;
		server = server_create(origin->server->network,
				       origin->server, sid,
				       args[0], args[argc-1], hopcount);
		server->created = origin->sconn->current_cmd_timestamp > 0 ?
			origin->sconn->current_cmd_timestamp : ioloop_time;
	} else if (server->uids) {
                /* update the unique server ID */
                p_free(server->pool, server->output_id);
		server->output_id = p_strconcat(server->pool, "\r",
						args[2], NULL);
	}

	if (!server->uids) {
		/* use server tokens */
		sid = args[2][0] == '$' ? args[2]+1 : args[2];
		server_conn_set_token(origin->sconn, server,
				      (unsigned int) atoi(sid));
	}

	send_origin_server(&sopts, server->uplink);
        sopts.timestamp = origin->sconn->current_cmd_timestamp;
	sopts.skip_server = origin->sconn->server;

	if (argc > 4) {
                /* with version */
		data = t_strdup_printf("SERVER %s %d %s %s :%s", server->name,
				       server->hopcount, server_get_id(server),
				       args[3], server->info);
	} else {
                /* without version */
		data = t_strdup_printf("SERVER %s %d %s :%s", server->name,
				       server->hopcount, server_get_id(server),
				       server->info);
	}
        server_send_all(server->network, &sopts, data);
}

/* SMASK <sid> <version> */
SERVER_COMMAND(cmd_smask)
{
        SendOptions sopts;
	Server *server, *mask_server;

	server = ircnet_find_server_id(origin->server->network, args[0]);
	if (server == NULL) {
		server_log(origin->server,
			   "Introducing duplicate SMASK ID '%s' (%s)",
			   args[0], server->name);
		return;
	}

        /* get the masking server */
	mask_server = origin->server;
	while (mask_server->masked)
		mask_server = mask_server->uplink;

	if (mask_server->local) {
		server_log(origin->server, "Sent SMASK of itself");
                return;
	}

        /* create the server */
	server = server_create(origin->server->network, origin->server, args[0],
			       mask_server->name, mask_server->info,
			       origin->server->hopcount+1);
        server->masked = TRUE;
	server->created = origin->sconn->current_cmd_timestamp > 0 ?
		origin->sconn->current_cmd_timestamp : ioloop_time;

        /* forward to UID capable servers */
	send_origin_server(&sopts, server->uplink);
        sopts.caps_set = CAP_UID;
	sopts.timestamp = origin->sconn->current_cmd_timestamp;
	sopts.skip_server = origin->sconn->server;

	server_send_all(server->network, &sopts,
			SERVER_CURRENT_LINE(2, FALSE));
}

/* SQUIT <server> [?] <comment> */
SERVER_COMMAND(cmd_squit)
{
        SendOptions sopts;
	Server *server, *server_link;
        const char *nickmask, *data;

	/* OK, let's break compatibility with other servers here for now.
	   Normally the servers are quit while the SQUIT is on the way
	   to the server that needs to be disconnected, however that server
	   might not want to close the connection for some reason. If we
	   process the SQUIT only if it originates from the link to the
	   server we allow this without horrible server desyncs. */

	/* FIXME: hostmasked servers aren't supported currently */
	server = ircnet_find_server(origin->sconn->network, args[0]);
	if (server == NULL) {
		server_log(origin->sconn->server,
			   "received SQUIT for unknown server %s", args[0]);
                return;
	}

	if (argc < 2) {
                /* nick is the default reason */
		args[1] = origin->nick != NULL ? origin->nick->nick :
			origin->server->name;
                argc = 2;
	}

	if (server->local) {
		/* they want to disconnect us from their side of the
		   network. */
		server = origin->sconn->server;
	}

	send_origin_nickserver(&sopts, origin->nick, origin->server);

	if (!server->linked && !server->local) {
		data = t_strdup_printf("SQUIT %s :%s", server->name,
				       args[argc-1]);
		server_link = server_find_first_link(server);
		if (server_link != origin->sconn->server) {
			/* non-authorative SQUIT received - forward it to
                           target server */
			server_send(server_link, &sopts, data);
			return;
		}

		/* authorative SQUIT received - send QUITs to clients,
		   and SQUITs servers */
		client_send_server_quit(server);
                sopts.skip_server = origin->sconn->server;
		server_send_all(origin->server->network, &sopts, data);
                server_destroy(server);
                return;
	}

	/* We were told to disconnect one of our server connections.
	   We could verify here if we want to allow it, but for now
	   just do it. */
	nickmask = origin->nick == NULL ? "<unknown>" :
		nick_get_mask(origin->nick);
	server_log(origin->server, "SQUIT %s sent by %s: %s",
		   server->name, nickmask, args[argc-1]);

        /* disconnect the server */
	server_disconnect(server_conn_find(server), args[argc-1]);
}

/* EOB <sid>,... */
SERVER_COMMAND(cmd_eob)
{
        Server *server;
	char *sids, *sid;

	sids = (char *) args[0];
	for (sid = i_strtoken(&sids, ','); sid != NULL;
	     sid = i_strtoken(&sids, ',')) {
		server = ircnet_find_server(origin->server->network, sid);
		if (server != NULL)
                        server->eob = TRUE;
	}
}

/* Send nick collision nick change messages. This is called when
   a) we noticed nick collision and want to SAVE our old nick,
   b) we recieved a SAVE message */
static void send_nick_collision_change(Server *origin_server,
				       Nick *nick, const char *path)
{
        IrcNetwork *network;
	SendOptions sopts;
        Server *skip_server;
        const char *data, *savedata;

	if (origin_server != NULL)
		path = t_strconcat(server_get_id(origin_server), " ", path, NULL);

        network = nick->server->network;
	skip_server = origin_server == NULL ? NULL :
		server_find_first_link(origin_server);
        data = t_strconcat("NICK ", nick_get_id(nick), NULL);

	if ((network->no_caps & CAP_UID) == 0) {
		/* forward SAVE to all servers capable of unique IDs,
		   except the one who sent it to us. */
		send_origin_server(&sopts, skip_server);
		savedata = t_strconcat("SAVE ", nick_get_id(nick), " :",
				       path, NULL);
		sopts.skip_server = skip_server;
		sopts.caps_set = CAP_UID;
		server_send_all(network, &sopts, savedata);
	}

	if ((network->all_caps & CAP_UID) == 0) {
		/* for servers who don't understand SAVE,
		   send it as NICK change. */
		send_origin_nick(&sopts, nick);
		sopts.skip_server = skip_server;
		sopts.caps_unset = CAP_UID;
		server_send_all(network, &sopts, data);
	}

        /* notify our clients with a nick change */
	client_send_nick_update(nick, send_origin_nick(&sopts, nick), data);
}

/* returns the new nick2 that should be used,
   or NULL if the nick should not be added. */
static const char *
nick_collision(Nick *nick, Server *server2, const char *id2,
	       const char *user2, const char *host2)
{
        ServerConn *sconn;
	SendOptions sopts;
        Client *client;
	const char *data, *new_nick;
	time_t timestamp;
        int sameuser;

	/* nick collision - the nick who's had it longer may keep it.

	   Without timestamping, the new nick may always keep it
	   (though the remote server might well change/kill the
	   other nick as well). */

        sconn = server_conn_link_find(server2);
	timestamp = sconn->current_cmd_timestamp;

	if (id2 != NULL && nick->server->uids) {
		/* unique IDs in use - change the other one's nick to
		   it's unique ID */
		if (nick->created < timestamp) {
                        /* change the new nick */
			if (sconn->caps & CAP_UID) {
				server_send(server2, NULL,
					    t_strdup_printf("SAVE %s", id2));
			} else {
				server_send(server2, send_origin(&sopts, id2),
					    t_strdup_printf("NICK :%s", id2));
			}
                        return id2;
		}

		/* change the old nick - notify everyone of the change */
		send_nick_collision_change(NULL, nick, NULL);

                new_nick = t_strdup(nick->nick);
		nick_change(nick, nick_get_id(nick), 0);
                return new_nick;
	}

	/* if the user/host matches the existing nick,
	   always kill the old user */
	sameuser = strcmp(user2, nick->user) == 0 &&
		strcmp(host2, nick->host) == 0;

	/* unique IDs not in use - we have to kill the other one. */
	if (!sameuser && nick->created < timestamp) {
		/* kill the new nick */
		data = t_strdup_printf("KILL %s :Nick collision with %s@%s(%s)",
				       nick_get_outputid(nick),
				       user2, host2, server2->name);
		server_send(server2, NULL, data);
                return NULL;
	}

        /* kill the old nick - send KILLs to servers */
	data = t_strdup_printf("KILL %s :Nick collision with %s@%s(%s)",
			       nick_get_outputid(nick), nick->user, nick->host,
			       nick->server->name);
	server_send_nick_update(nick, send_skip_server(&sopts, server2), data);

        /* QUIT message to other clients */
	client = client_find(nick->nick);
	data = t_strdup_printf("QUIT :Nick collision with %s@%s(%s)",
			       nick->user, nick->host, nick->server->name);
	send_origin_nick(&sopts, nick);
        sopts.skip_client = client;
	client_send_nick_update(nick, &sopts, data);

        /* nick collision numeric + ERROR to client who quit */
	if (!nick->server->local)
		server_remove_nick(nick);
	else {
		client_send_numeric(client, ERR_NICKCOLLISION,
				    nick->nick, user2, host2);
		client_disconnect(client, "Closing link - nick collision",
				  TRUE, FALSE);
	}

	return nick->nick;
}

/* Verify that the given unique ID isn't used. If it is, kill the new user. */
static int verify_unique_id(CommandOrigin *origin, const char *id)
{
	Nick *nick;

	nick = ircnet_find_nick_id(origin->server->network, id);
	if (nick == NULL)
		return TRUE;

	/* duplicate unique ID for nick - should never happen.
	   Kill the nick. */
	server_log(origin->sconn->server,
		   "duplicate unique ID %s (old %s!%s)",
		   id, nick->nick, nick->server->name);
	server_send(origin->sconn->server, NULL,
		    t_strdup_printf("KILL %s :duplicate ID", id));
	return FALSE;
}

/* Find server for given token. If the server isn't found, kill the nick. */
static Server *nick_find_token_server(ServerConn *sconn, const char *token,
				      const char *nick)
{
	Server *server;

	/* note that the server_conn_find_token() cannot return servers
	   that aren't either the server itself or one of it's downlinks, so
	   we don't have to check if routing is correct. */
	server = server_conn_find_token(sconn, atoi(token));
	if (server != NULL)
		return server;

	/* unknown server token, not allowed, kill the nick. */
	server_log(sconn->server, "unknown server token %s for %s",
		   token, nick);
	server_send(sconn->server, NULL,
		    t_strdup_printf("KILL %s :nick behind unknown server",
				    nick));
	return NULL;
}

/* :<source> UNICK <nick> <uid> <user> <host> <ip> +<umode> [?] <real name> */
SERVER_COMMAND(cmd_unick)
{
        Nick *nick;

	if (!verify_unique_id(origin, args[1]))
		return;

        /* check for nick collision */
        nick = ircnet_find_nick_name(origin->server->network, args[0]);
	if (nick != NULL) {
		args[0] = nick_collision(nick, origin->server,
					 args[1], args[2], args[3]);
                if (args[0] == NULL)
			return;
	}

	nick = nick_create(origin->server, args[1], args[0], args[2],
			   args[3], args[argc-1]);

	if (origin->sconn->current_cmd_timestamp > 0) {
		nick->created = nick->last_nick_change =
			origin->sconn->current_cmd_timestamp;
	}

	nick->hopcount = origin->server->hopcount;
        net_host2ip(args[4], &nick->ip);
	nick_set_umode(nick, args[5]);

	server_send_nickjoins(nick);
}

/* :<old_nick> NICK <new_nick> */
SERVER_COMMAND(cmd_nick_change)
{
        SendOptions sopts;
        Nick *nick;

	/* check for nick collision */
        nick = ircnet_find_nick(origin->server->network, args[0]);
	if (nick != NULL && nick != origin->nick) {
		args[0] = nick_collision(nick, origin->server,
					 nick_get_id(origin->nick),
					 origin->nick->user,
					 origin->nick->host);
                if (args[0] == NULL)
			return;
	}

	if (nick == NULL || strcmp(nick->nick, args[0]) != 0) {
		/* changing nick - notify everyone except the server where
		   the nick change came from */
		send_nick_update(origin->nick,
				 send_origin_nick(&sopts, origin->nick),
				 t_strconcat("NICK :", args[0], NULL));

		nick_change(origin->nick, args[0], ioloop_time);
	}
}

/* NICK <nick> <hopcount> <user> <host> <token> +<umode> <real name> */
SERVER_COMMAND(cmd_nick)
{
	Server *server;
        Nick *nick;

	if (args[0] != NULL && args[1] == NULL && origin->nick != NULL) {
                /* changing nick */
		cmd_nick_change(origin, cmd, argc, args);
                return;
	}

	if (args[6] == NULL) {
		server_log(origin->sconn->server,
			   "Not enough parameters for NICK registeration");
                return;
	}

	/* new user */
	server = nick_find_token_server(origin->sconn, args[4], args[0]);
	if (server == NULL)
		return;

        /* check for nick collision */
        nick = ircnet_find_nick(server->network, args[0]);
	if (nick != NULL) {
		args[0] = nick_collision(nick, server, NULL, args[2], args[3]);
                if (args[0] == NULL)
			return;
	}

	/* new user */
	nick = nick_create(server, args[0], args[0], args[2], args[3], args[6]);
	if (origin->sconn->current_cmd_timestamp > 0) {
		nick->created = nick->last_nick_change =
			origin->sconn->current_cmd_timestamp;
	}
	nick->hopcount = atoi(args[1]);
	nick_set_umode(nick, args[5]);
	server_send_nickjoins(nick);
}

/* Invalid QUIT - introduce the nick again and join it back to channels */
static void nick_abort_quit(Nick *nick, Server *server)
{
        /* send NICK */
	server_send_nickjoin(server, nick);

	/* FIXME: join back to channels */
}

/* QUIT [?] <message> */
SERVER_COMMAND(cmd_quit)
{
        SendOptions sopts;

	if (origin->nick == NULL) {
                /* server sent QUIT - I guess it wants to SQUIT itself. */
		args[1] = args[0];
                args[0] = origin->server->name;
		cmd_squit(origin, cmd, argc, args);
                return;
	}

	if (!server_check_route(origin->nick->server, origin->sconn->server)) {
		/* invalid route - shouldn't happen! Anyway, send back the
                   NICK so that there won't be any desyncs */
                nick_abort_quit(origin->nick, origin->sconn->server);
		return;
	}

        /* nick quits - notify everyone except the server who told it to us */
	send_nick_update(origin->nick,
			 send_origin_nick_skip(&sopts, origin->nick),
                         SERVER_CURRENT_LINE(1, TRUE));

        server_remove_nick(origin->nick);
}

/* KILL <uid|nick> :<path> (<comment>) */
SERVER_COMMAND(cmd_kill)
{
	SendOptions sopts;
        Nick *nick;

        nick = ircnet_find_nick(origin->server->network, args[0]);
	if (nick == NULL) {
		send_numeric_reply(origin, ERR_NOSUCHNICK, args[0]);
                return;
	}

	/* FIXME: update the path */

	/* forward KILL to all servers except the server who told it to us */
	send_origin_nickserver(&sopts, origin->nick, origin->server);
	sopts.skip_server = origin->server;
	server_send_nick_update(nick, &sopts,
		t_strdup_printf("KILL %s :%s",
				nick_get_outputid(nick), args[1]));

	if (nick->server->local) {
		/* disconnect the client */
		client_disconnect(client_find(nick->nick),
				  t_strconcat("Killed: ", args[1], NULL),
				  TRUE, FALSE);
	}

        /* notify our clients of the kill with a QUIT message */
	client_send_nick_update(nick, send_origin_nick(&sopts, nick),
        	t_strconcat("QUIT :Killed ", args[1], NULL));

        server_remove_nick(nick);
}

/* SAVE <uid> <path> */
SERVER_COMMAND(cmd_save)
{
	Nick *nick;

        nick = ircnet_find_nick_id(origin->server->network, args[0]);
	if (nick == NULL) {
		send_numeric_reply(origin, ERR_NOSUCHNICK, args[0]);
                return;
	}

	if (irc_id_cmp(nick->nick, nick_get_id(nick)) == 0)
                return; /* already saved */

	send_nick_collision_change(origin->server, nick, args[1]);

	nick_change(nick, nick_get_id(nick), ioloop_time);
}

static void channel_send_data(CommandOrigin *origin, SendOptions *sopts,
			      const char *channame, const char *data)
{
        Channel *channel;
	ChannelNick *channick;

	channel = ircnet_find_channel(origin->server->network, channame);
	if (channel == NULL) {
		send_numeric_reply(origin, ERR_NOSUCHNICK, channame);
		return;
	}

	channick = origin->nick == NULL ? NULL :
		channel_get_nick(channel, origin->nick);
	if (origin->nick != NULL &&
	    !NICK_CAN_TALK(channel, channick)) {
		send_numeric_reply(origin, ERR_CANNOTSENDTOCHAN, channame);
                return;
	}

	channel_send(channel, sopts, data);
}

/* PRIVMSG <target>[,<target2>,...] [?] <msg>
   NOTICE <target>[,<target2>,...] [?] <msg> */
SERVER_COMMAND(cmd_talk)
{
        SendOptions sopts;
	Nick *nick;
        const char *data;
        char *targets, *target;

	send_origin_nickserver(&sopts, origin->nick, origin->server);
	sopts.skip_server = origin->sconn->server;
	sopts.channel_send_talk = TRUE;

	/* FIXME: support for user[%host]@server, #hostmask, $servermask */
	targets = (char *) args[0];
	for (target = i_strtoken(&targets, ','); target != NULL;
	     target = i_strtoken(&targets, ',')) {
		if (is_chanprefix_char(target[0])) {
			data = t_strconcat(cmd, " ", target,
					   " :", args[argc-1], NULL);
			channel_send_data(origin, &sopts, target, data);
		} else {
			nick = ircnet_find_nick(origin->server->network,
						target);
			if (nick != NULL) {
				data = t_strconcat(cmd, " ",
						   nick_get_outputid(nick),
						   " :", args[argc-1], NULL);
				nick_send_data(nick, &sopts, data);
			}
		}
	}
}

SERVER_COMMAND_EXT(cmd_cmode); /* in server-channel-commands.c */

/* MODE <uid|nick> [<modes> [<params>]] */
SERVER_COMMAND(cmd_umode)
{
        SendOptions sopts;
	Nick *nick;

        nick = ircnet_find_nick(origin->server->network, args[0]);
	if (nick == NULL) {
		send_numeric_reply(origin, ERR_NOSUCHNICK, args[0]);
                return;
	}

	if (argc == 1) {
		/* just querying the mode */
		send_numeric_reply(origin, RPL_UMODEIS,
				   nick_get_umode_str(nick));
                return;
	}

	if (origin->nick != nick) {
                /* someone else than the user itself changed it's mode */
		server_log(origin->server, "MODE for user %s from %s",
			   nick->nick, origin->nick == NULL ?
			   origin->server->name : nick_get_mask(origin->nick));
	}

        /* forward the mode change to other servers */
	send_origin_nickserver(&sopts, origin->nick, origin->server);
	sopts.skip_server = origin->sconn->server;
	server_send_all(origin->sconn->network, &sopts,
			t_strjoin_replace((char *const *) args, ' ', 0,
					  nick_get_outputid(nick)));

        nick_set_umode(nick, args[1]);
}

/* MODE <uid|nick|channel>[,...] [<modes> [<params>]] */
SERVER_COMMAND(cmd_mode)
{
	char *names, *name;

	names = (char *) args[0];
	for (name = i_strtoken(&names, ','); name != NULL;
	     name = i_strtoken(&names, ',')) {
		if (is_chanprefix_char(name[0])) {
                        /* channel mode change */
			cmd_cmode(origin, cmd, argc, args);
		} else {
                        /* user mode change */
			cmd_umode(origin, cmd, argc, args);
		}
	}
}

/* PING <origin> [<destination>] */
SERVER_COMMAND(cmd_ping)
{
        const char *data;

	if (args[1] != NULL && args[1] == '\0')
		args[1] = NULL;

	if (!command_forward_server(origin, cmd, args, 1))
                return;

	/* that's us - reply with PONG. */
	if (args[1] == NULL)
		data = origin->server->network->server->name;
	else {
		/* destination parameter given, use it as origin
		   parameter in case it was a hostmask */
		data = args[1];
	}

	/* Use always the origin given to us in prefix,
	   and ignore the parameter. */
	data = t_strconcat("PONG ", data, " :",
			   ORIGIN_GET_OUTPUTID(origin), NULL);
	server_send(origin->server, NULL, data);
}

/* PONG <origin> [<destination>] */
SERVER_COMMAND(cmd_pong)
{
        SendOptions sopts;
	Nick *nick;
        const char *data;

        /* nick destination? */
	nick = args[1] == NULL ? NULL :
		ircnet_find_nick(origin->server->network, args[1]);
	if (nick != NULL && !nick->server->local)
		nick = NULL;

	if (nick == NULL && !command_forward_server(origin, cmd, args, 1))
                return;

	/* PONG was for us */
	if (nick != NULL) {
		/* forward to client */
		send_origin_server(&sopts, origin->server);
		data = t_strconcat("PONG ", ORIGIN_GET_NAME(origin), " ",
				   nick->nick, NULL);
		client_send(client_find(nick->nick), &sopts, data);
	}
}

/* WHOIS [<target server>] <uid|nick>[,...] */
SERVER_COMMAND(cmd_whois)
{
	SendOptions sopts;

	if (argc < 2 ||
	    command_forward_server(origin, cmd, args, 0)) {
		if (argc >= 2) args++;
		command_reply_whois(send_back(&sopts, origin),
				    origin->server->network->server,
				    args[0], TRUE);
	}
}

/* WHOWAS <nick>,... [<count> [<target>]] */
SERVER_COMMAND(cmd_whowas)
{
	SendOptions sopts;

	if (argc < 3 ||
	    command_forward_server(origin, cmd, args, 2)) {
		command_reply_whowas(send_back(&sopts, origin),
				     origin->server->network->server,
				     args[0], args[2]);
	}
}

/* VERSION [<target server>] */
SERVER_COMMAND(cmd_version)
{
	SendOptions sopts;

	if (argc < 1 ||
	    command_forward_server(origin, cmd, args, 0)) {
		command_reply_version(send_back(&sopts, origin),
				      origin->server->network->server);
	}
}

/* MOTD [<target server>] */
SERVER_COMMAND(cmd_motd)
{
	SendOptions sopts;

	if (argc < 1 ||
	    command_forward_server(origin, cmd, args, 0)) {
		command_reply_motd(send_back(&sopts, origin),
				   origin->server->network->server);
	}
}

/* LUSERS [<mask> [<target server>]] */
SERVER_COMMAND(cmd_lusers)
{
	SendOptions sopts;

	if (argc < 2 ||
	    command_forward_server(origin, cmd, args, 1)) {
		command_reply_lusers(send_back(&sopts, origin),
				     origin->server->network->server, args[0]);
	}
}

/* LINKS [[<target server>] <mask>] */
SERVER_COMMAND(cmd_links)
{
	SendOptions sopts;

	if (argc < 2 ||
	    command_forward_server(origin, cmd, args, 0)) {
                if (argc >= 2) args++;
		command_reply_links(send_back(&sopts, origin),
				    origin->server->network->server, args[1]);
	}
}

/* ERROR <message> */
SERVER_COMMAND(cmd_error)
{
        server_log(origin->server, "%s", args[0]);
}

void server_commands_init(void)
{
	pool = pool_create("Server commands", 2048, TRUE);
        server_commands = command_list_create(pool);

        server_command_register("SERVER", 4, cmd_server);
        server_command_register("SMASK", 2, cmd_smask);
        server_command_register("SQUIT", 1, cmd_squit);
        server_command_register("EOB", 1, cmd_eob);

	server_command_register("UNICK", 7, cmd_unick);
        server_command_register("NICK", 1, cmd_nick);
        server_command_register("QUIT", 0, cmd_quit);
        server_command_register("KILL", 1, cmd_kill);
        server_command_register("SAVE", 1, cmd_save);

	server_command_register("PRIVMSG", 2, cmd_talk);
        server_command_register("NOTICE", 2, cmd_talk);
        server_command_register("MODE", 1, cmd_mode);

        server_command_register("PING", 0, cmd_ping);
	server_command_register("PONG", 0, cmd_pong);

	server_command_register("WHOIS", 1, cmd_whois);
	server_command_register("WHOWAS", 1, cmd_whowas);
	server_command_register("VERSION", 0, cmd_version);
	server_command_register("MOTD", 0, cmd_motd);
	server_command_register("LUSERS", 0, cmd_lusers);
	server_command_register("LINKS", 0, cmd_links);

	server_command_register("ERROR", 0, cmd_error);

        server_channel_commands_init();
}

void server_commands_deinit(void)
{
	server_command_unregister("SERVER", cmd_server);
        server_command_unregister("SMASK", cmd_smask);
        server_command_unregister("SQUIT", cmd_squit);
        server_command_unregister("EOB", cmd_eob);

	server_command_unregister("UNICK", cmd_unick);
        server_command_unregister("NICK", cmd_nick);
        server_command_unregister("QUIT", cmd_quit);
        server_command_unregister("KILL", cmd_kill);
        server_command_unregister("SAVE", cmd_save);

	server_command_unregister("PRIVMSG", cmd_talk);
        server_command_unregister("NOTICE", cmd_talk);
        server_command_unregister("MODE", cmd_mode);

        server_command_unregister("PING", cmd_ping);
        server_command_unregister("PONG", cmd_pong);

	server_command_unregister("WHOIS", cmd_whois);
	server_command_unregister("WHOWAS", cmd_whowas);
	server_command_unregister("VERSION", cmd_version);
	server_command_unregister("MOTD", cmd_motd);
	server_command_unregister("LUSERS", cmd_lusers);
	server_command_unregister("LINKS", cmd_links);

	server_command_unregister("ERROR", cmd_error);

        server_channel_commands_deinit();

        command_list_destroy(server_commands);
        pool_unref(pool);
}
