#include "common.h"
#include "ioloop.h"
#include "server-connection.h"
#include "server-commands.h"
#include "irc.h"

ServerConn *server_connect(Server *own_server, int handle, const char *id,
			   const char *name, const char *info)
{
	ServerConn *sconn;
	Server *server;

        /* create a server for us */
	server = server_create(own_server->network, own_server,
			       id, name, info, 1);
        server->linked = TRUE;

        /* create the server connection */
	sconn = p_new(server->pool, ServerConn, 1);
	sconn->pool = server->pool;
        pool_ref(sconn->pool);

	sconn->refcount = 1;
	sconn->leaf_id = -1;
        sconn->network = server->network;
	sconn->server = server;
	sconn->handle = handle;

	/* reset the all_caps since we don't have any capabilities
	   in this server yet */
	server->network->all_caps = 0;

        sconn->max_mode_args = server_max_mode_args;
	sconn->max_line_length = server_max_line_length;
	if (sconn->network->min_line_length == 0 ||
	    sconn->network->min_line_length > sconn->max_line_length)
                sconn->network->min_line_length = sconn->max_line_length;

	sconn->inbuf = tbuffer_create(handle, sconn->server->pool,
				      IO_PRIORITY_HIGH, 0);
	sconn->outbuf = tbuffer_create(handle, sconn->server->pool,
				       IO_PRIORITY_HIGH, 0);
	sconn->outbuf->dump = TRUE; /* FIXME: debug! */
	sconn->tokens = hash_create(sconn->pool, 0, NULL, NULL);

	sconn->io = io_add_priority(handle, IO_PRIORITY_HIGH, IO_READ,
				    (IOFunc) server_parse_input, sconn);

        server->sconn = sconn;
	return sconn;
}

void server_conn_ref(ServerConn *sconn)
{
        sconn->refcount++;
}

void server_conn_unref(ServerConn *sconn)
{
	Pool pool;

	if (--sconn->refcount > 0)
                return;

	ircnet_update_caps(sconn->network);

	pool = sconn->pool;

	if (sconn->handle != -1)
		net_disconnect(sconn->handle);

        if (sconn->io != NULL)
		io_remove(sconn->io);

	if (sconn->server != NULL)
		server_destroy(sconn->server);

        tbuffer_destroy(sconn->inbuf);
	tbuffer_destroy(sconn->outbuf);

        hash_destroy_clean(sconn->tokens);
	hash_destroy_clean(sconn->unknown_nicks);

#ifdef POOL_CHECK_LEAKS
	while (sconn->first_unknown_nick != NULL) {
                NickAccessList *next = sconn->first_unknown_nick->next;

		p_free(sconn->pool, sconn->first_unknown_nick->nick);
		p_free(sconn->pool, sconn->first_unknown_nick);
                sconn->first_unknown_nick = next;
	}
#endif

	p_free(pool, sconn);
        pool_unref(pool);
}

ServerConn *server_conn_find(Server *server)
{
	return_val_if_fail(server != NULL, NULL);

	return server->sconn;
}

ServerConn *server_conn_link_find(Server *server)
{
	return_val_if_fail(server != NULL, NULL);

        return server_conn_find(server_find_first_link(server));
}

void server_conn_set_caps(ServerConn *sconn, unsigned int caps)
{
	sconn->caps |= caps;
        ircnet_update_caps(sconn->network);
}

void server_conn_set_token(ServerConn *sconn, Server *dest_server,
			     unsigned int token)
{
	hash_insert(sconn->tokens, INT_TO_POINTER(token), dest_server);
}

Server *server_conn_find_token(ServerConn *sconn, unsigned int token)
{
        return hash_lookup(sconn->tokens, INT_TO_POINTER(token));
}

NickAccessList *server_leaf_access_unknown_nick(Server *server,
						const char *nick)
{
	ServerConn *sconn;
	NickAccessList *list;

        return_val_if_fail(server->leaf && server->linked, NULL);

        sconn = server_conn_find(server);
	if (sconn->unknown_nicks == NULL) {
		sconn->unknown_nicks =
			hash_create(server->pool, 100, irc_nick_hash,
				    (HashCompareFunc) irc_nick_cmp);
	}

	list = hash_lookup(sconn->unknown_nicks, nick);
	if (list == NULL) {
                list = p_new(server->pool, NickAccessList, 1);
                list->nick = p_strdup(server->pool, nick);
		hash_insert(sconn->unknown_nicks, list->nick, list);
	} else {
                /* remove the item from the linked list */
		if (list->prev != NULL)
                        list->prev->next = list->next;
		if (list->next != NULL)
			list->next->prev = list->prev;
	}

        list->last_access = ioloop_time;

        /* add the item last in the list */
	list->prev = sconn->last_unknown_nick;
        list->next = NULL;
	if (sconn->last_unknown_nick != NULL)
		sconn->last_unknown_nick->next = list;
	if (sconn->first_unknown_nick == NULL)
                sconn->first_unknown_nick = list;
	sconn->last_unknown_nick = list;
        return list;
}

void server_leaf_remove_unknown_nick(Server *server, const char *nick)
{
	ServerConn *sconn;
	NickAccessList *list;

        return_if_fail(server->leaf && server->linked);

        sconn = server_conn_find(server);
	list = hash_lookup(sconn->unknown_nicks, nick);
	if (list != NULL) {
		if (list->prev != NULL)
                        list->prev->next = list->next;
		if (list->next != NULL)
			list->next->prev = list->prev;

		if (sconn->first_unknown_nick == list)
			sconn->first_unknown_nick = list->next;
		if (sconn->last_unknown_nick == list)
			sconn->last_unknown_nick = list->prev;

		hash_remove(sconn->unknown_nicks, nick);
                p_free(server->pool, list->nick);
		p_free(server->pool, list);
	}
}

int server_leaf_find_unknown_nick(Server *server, const char *nick)
{
	ServerConn *sconn;

        return_val_if_fail(server->leaf && server->linked, FALSE);

        sconn = server_conn_find(server);
	return hash_lookup(sconn->unknown_nicks, nick) != NULL;
}

static int server_fill_origin(CommandOrigin *origin, ServerConn *sconn,
			      char *prefix)
{
        char *id;

	memset(origin, 0, sizeof(CommandOrigin));
	origin->sconn = sconn;

	if (prefix == NULL) {
                /* no prefix, the command came from the linked server */
		origin->server = sconn->server;
                return TRUE;
	}

        /* prefix should only be a nick, nick ID or server */
	id = prefix;
	while (*prefix != '\0') {
		if (*prefix == '!' || *prefix == '@') {
                        *prefix = '\0';
			break;
		}
                prefix++;
	}

	if ((sconn->caps & CAP_UID) && id[0] >= '0' && id[0] <= '9') {
		/* most probably an ID */
		origin->nick = ircnet_find_nick_id(sconn->network, id);
		if (origin->nick == NULL) {
			origin->server =
				ircnet_find_server_id(sconn->network, id);
			if (origin->server == NULL) {
				server_log(sconn->server,
					   "command from unknown ID %s", id);
			}
		}
	} else if (strchr(id, '.') == NULL) {
		/* nick */
		origin->nick = ircnet_find_nick_name(sconn->network, id);
		if (origin->nick == NULL) {
			/* unknown nick */
			server_log(sconn->server,
				   "command from unknown nick %s", id);
		}
	} else {
		/* server */
		origin->server = ircnet_find_server_name(sconn->network, id);
		if (origin->server == NULL) {
			server_log(sconn->server,
				   "command from unknown server %s", id);
		}
	}

        if (origin->nick != NULL)
		origin->server = origin->nick->server;

	if (origin->server != NULL &&
	    server_find_first_link(origin->server) != sconn->server) {
		/* server shouldn't be sending this to us.. */
		server_log(sconn->server, "shouldn't route message from %s",
			   origin->server->name);
		origin->server = NULL;
	}

	return origin->server != NULL;
}

static int server_parse_next_line(ServerConn *sconn)
{
        CommandOrigin origin;
	const char *cmd, *args[COMMAND_MAX_ARGS+1], *id;
        char *line, *prefix;
        int timestamp, argc;

	line = (char *) tbuffer_next_line(sconn->inbuf);
	if (line == NULL || *line == '\0')
		return FALSE;

	line = irc_parse_prefix(line, &prefix, &timestamp, &id);
	if (*line == '\0')
                return TRUE;

	cmd = irc_parse_command((char *) t_strdup(line), &argc, args);
	if (cmd == NULL)
                return TRUE;

        if (!server_fill_origin(&origin, sconn, prefix))
		return TRUE;

	sconn->current_cmd_timestamp =
		timestamp == 0 ? 0 : ioloop_time - timestamp;
	sconn->current_cmd_id = id;
        sconn->current_cmd_line = line;

	server_command_parse(&origin, cmd, argc, args);

        sconn->current_cmd_timestamp = 0;
	sconn->current_cmd_id = NULL;
        sconn->current_cmd_line = NULL;

        return TRUE;
}

void server_parse_input(ServerConn *sconn)
{
        server_conn_ref(sconn);
	sconn->last_data = ioloop_time;

	switch (tbuffer_read(sconn->inbuf)) {
	case -1:
                /* disconnected */
                server_disconnect_errno(sconn);
		break;
	case -2:
		/* input buffer is full - shoudn't happen unless the
		   server don't send us any line breaks */
		server_disconnect(sconn, "Excess flood");
		break;
	default:
		while (server_parse_next_line(sconn) && !sconn->disconnected)
			;
	}

	server_conn_unref(sconn);
}

void server_disconnect(ServerConn *sconn, const char *msg)
{
        SendOptions sopts;

	if (sconn->disconnected)
		return;
	sconn->disconnected = TRUE;

	if (msg == NULL)
		msg = "";

	/* send QUITs to clients and SQUIT to all servers
	   (including the one we're disconnecting) */
	server_log(sconn->server, "Server disconnected: %s", msg);

	client_send_server_quit(sconn->server);

	/* FIXME: old servers need QUITs as well + SQUITs for all servers.. */
	server_send_all(sconn->server->network,
			send_origin_server(&sopts, sconn->server->uplink),
			t_strdup_printf("SQUIT %s :%s",
					sconn->server->name, msg));

	tbuffer_close(sconn->inbuf);
	tbuffer_close(sconn->outbuf);

	server_conn_unref(sconn);
}

void server_disconnect_errno(ServerConn *sconn)
{
	server_disconnect(sconn, t_strdup_printf("%s (%d)",
						 strerror(errno), errno));
}
