#include "common.h"
#include "version.h"
#include "server-connection.h"
#include "server-login.h"
#include "channel.h"
#include "channel-modes.h"

static const char *server_pass = "pass"; /* FIXME */

static void login_disconnect(int handle, const char **ret, const char *msg)
{
	char buffer[CLIENT_MAX_LINE_LENGTH+3];
        int len;

	*ret = t_strdup(msg);

	len = i_snprintf(buffer, sizeof(buffer), "ERROR :%s\r\n", msg);
        net_transmit(handle, buffer, len);
        net_disconnect(handle);
}

ServerConn *server_login(int handle, Server *uplink, const char *name,
			 const char *sid, const char *info, const char *version,
			 const char *flags, const char *options, int remote,
			 const char **error_msg)
{
	ServerConn *sconn;
	Server *server, *server_link;
        PArray links;
	char myoptions[10], *ptr, *optptr;
	unsigned int caps;
        int i, token;

	if (strchr(name, '.') == NULL) {
		/* '.' char MUST be in server name - we depend on it later */
		login_disconnect(handle, error_msg,
				 "invalid server name - domain is missing");
                return NULL;
	}

        /* check that the server doesn't already exist */
	server = ircnet_find_server(uplink->network, name);
	if (server != NULL) {
		login_disconnect(handle, error_msg,
				 t_strconcat(name, " already connected", NULL));
                return NULL;
	}

        /* check that flags parameter is correct */
	ptr = strchr(flags, '|');
	if (ptr == NULL) {
		login_disconnect(handle, error_msg, "missing '|' from flags");
                return NULL;
	}
        *ptr++ = '\0';

	if (*flags != '\0' && strcmp(flags, "IRC") != 0) {
		login_disconnect(handle, error_msg,
			t_strdup_printf("invalid protocol '%s' in flags",
					flags));
                return NULL;
	}

        /* figure out our capabilities */
	caps = version_get_caps(version);

        /* create the connection */
	if (caps & CAP_UID) {
		token = -1;

		if (ircnet_find_server_id(uplink->network, sid) != NULL) {
			login_disconnect(handle, error_msg,
				t_strdup_printf("SID '%s' is already in use",
						sid));
                        return NULL;
		}
	} else {
                token = atoi(sid);
		sid = NULL;
	}
	sconn = server_connect(uplink, handle, sid, name, info);
	sconn->inbuf->dump = TRUE; /* FIXME: debug! */

	server = sconn->server;
	server->hopcount = 1;

	server_conn_set_caps(sconn, caps);
        if ((sconn->caps & CAP_UID) == 0)
		server_conn_set_token(sconn, server, token);

	/* reply with our own PASS */
	optptr = myoptions;
	if (options != NULL) {
		if (strchr(options, 'Z') != NULL) {
			sconn->compressed = TRUE;
			*optptr++ = 'Z';
		}
		if (strchr(options, 'P') != NULL) {
			*optptr++ = 'P';
		}
	}
	*optptr = '\0';

	if (!remote) {
		server_send(server, NULL,
			    t_strdup_printf("PASS %s %s IRC|%s %s",
					    server_pass, IRCD_VERSION,
					    ircd_flags, myoptions));
	}

	server_log(server, "Server connected");

	/* notify others of the server connection */
        server_link = server_find_first_link(server);

	links = ircnet_links(server->network);
        parray_freeze(links);
	for (i = 0; i < links->last; i++) {
		Server *link = parray_index(links, i);

		if (link != NULL && link != server_link)
			server_send_serverjoin(link, server);
	}
        parray_thaw(links);

        /* burst it */
	server_send_burst(server, remote);
        return sconn;
}

static void server_send_downlinks(Server *dest_server, Server *server,
				  int skip_first)
{
        Server *link;
	int i;

        if (!skip_first)
		server_send_serverjoin(dest_server, server);

	if (server->downlinks == NULL)
                return;

	parray_freeze(server->downlinks);
	for (i = 0; i < server->downlinks->last; i++) {
		link = parray_index(server->downlinks, i);

		if (link != NULL && link != dest_server)
			server_send_downlinks(dest_server, link, FALSE);
	}
	parray_thaw(server->downlinks);
}

static void send_nickjoin(const char *nick_name, Nick *nick, Server *server)
{
	server_send_nickjoin(server, nick);
}

typedef struct {
	NetSendBuffer *sendbuf;
        int uids;
} NJoinData;

static int send_njoin(Nick *nick, ChannelNick *channick,
		      NJoinData *data)
{
        const char *status, *id;
	int idlen, statuslen;

	if (channick->status & CSTATUS_OWNER) {
		status = "@@";
                statuslen = 2;
	} else if (channick->status & CSTATUS_OP) {
		status = "@";
                statuslen = 1;
	} else if (channick->status & CSTATUS_VOICE) {
                status = "+";
                statuslen = 1;
	} else {
                status = NULL;
		statuslen = 0;
	}

	id = data->uids ? nick_get_id(nick) : nick->nick;
	idlen = strlen(id);
	if (!ircnet_sendbuf_check(data->sendbuf, idlen + statuslen + 1))
		ircnet_sendbuf_add(data->sendbuf, ",", 1);

        if (statuslen > 0)
		ircnet_sendbuf_add(data->sendbuf, status, statuslen);
	ircnet_sendbuf_add(data->sendbuf, id, idlen);

        return TRUE;
}

static void modelist_send(ServerConn *sconn, Channel *channel, char mode)
{
        SendOptions sopts;
	ModeBuffer *mode_buf;
	PArray *list;
	const char *mask;
        int i;

        list = chanmode_get_list(channel, mode);
	if (list == NULL || *list == NULL)
		return;

	mode_buf = mode_buffer_create(channel,
				      send_to_server(&sopts, sconn->server),
				      sconn->max_mode_args,
                                      sconn->max_line_length);

	for (i = 0; i < (*list)->last; i++) {
		mask = parray_index(*list, i);

		mode_buffer_append(mode_buf, TRUE, mode, mask);
	}
	mode_buffer_flush(mode_buf);
}

static void send_channeljoin(const char *channel_name, Channel *channel,
			     Server *server)
{
	SendOptions sopts;
        NJoinData njoin_data;
	ServerConn *sconn;
	NetSendBuffer *sendbuf;
        const char *prefix, *mode, *mode_args;

	if (CHANNEL_IS_LOCAL(channel))
                return;

	/* send NJOINs - FIXME: should we bother supporting JOINs?.. */
	prefix = t_strdup_printf("NJOIN %s :", channel->name);

	send_to_server(&sopts, server);
        sopts.timestamp = channel->created;

	sconn = server_conn_find(server);
	sendbuf = ircnet_sendbuf_create(server->network, &sopts, prefix,
					sconn->max_line_length);

	njoin_data.sendbuf = sendbuf;
        njoin_data.uids = sconn->caps & CAP_UID;
	tree_traverse(channel->nicks, (TreeTraverseFunc) send_njoin,
		      &njoin_data);
	ircnet_sendbuf_flush(sendbuf);

	/* send TOPIC */
	if ((sconn->caps & CAP_BURSTTOPIC) && channel->topic != NULL) {
		/* FIXME: send the "topic by" as well? */
		send_options_clear(&sopts);
		sopts.timestamp = channel->topic_time;

		server_send(server, &sopts,
			    t_strdup_printf("TOPIC %s :%s",
					    channel->name, channel->topic));
	}

	/* send MODE */
	mode = channel_get_mode(channel, &mode_args);
	if (mode[1] != '\0') {
		server_send(server, NULL,
			    t_strconcat("MODE ", channel->name, " ",
					mode, " ", mode_args, NULL));
	}

	/* send bans, ban exceptions and invites */
	modelist_send(sconn, channel, 'b');
        modelist_send(sconn, channel, 'e');
        modelist_send(sconn, channel, 'I');
}

typedef struct {
        Server *target;
	NetSendBuffer *sendbuf;
} EOBData;

static void send_downlinks_eob(Server *server, EOBData *data)
{
        const char *id;
	int i, idlen;

	id = server_get_id(server);
	idlen = strlen(id);

	if (!ircnet_sendbuf_check(data->sendbuf, idlen + 1))
		ircnet_sendbuf_add(data->sendbuf, ",", 1);
	ircnet_sendbuf_add(data->sendbuf, id, idlen);

	if (server->downlinks == NULL)
		return;

        /* add UID-capable downlinks as well */
	for (i = 0; i < server->downlinks->last; i++) {
		Server *link = parray_index(server->downlinks, i);
		if (link->uids && link != data->target)
                        send_downlinks_eob(link, data);
	}
}

static void send_eob(Server *server)
{
        SendOptions sopts;
	ServerConn *sconn;
	EOBData data;

	if (!server->uids)
		return;

        sconn = server_conn_find(server);
        data.target = server;
	data.sendbuf = ircnet_sendbuf_create(server->network,
					     send_to_server(&sopts, server),
					     "EOB ", sconn->max_line_length);
        send_downlinks_eob(server->network->server, &data);
	ircnet_sendbuf_flush(data.sendbuf);
}

void server_send_burst(Server *server, int remote)
{
        /* send list of all servers, first being our own server */
	server_send_downlinks(server, server->network->server, remote);

	if (server->leaf)
                return;

        /* send list of all users */
	hash_foreach(server->network->nicks,
		     (HashForeachFunc) send_nickjoin, server);


	/* send list of channels, nicks joined to them and their modes */
	hash_foreach(server->network->channels,
		     (HashForeachFunc) send_channeljoin, server);

        send_eob(server);
}
