#include "common.h"
#include "ioloop.h"
#include "server-connection.h"
#include "nick.h"
#include "channel.h"

#define SERVER_BUFFER_INITIAL_SIZE	4096

static Pool server_pool;
static int token_counter;

Server *server_create(IrcNetwork *net, Server *uplink, const char *id,
		      const char *name, const char *info, int hopcount)
{
	Server *server;

        return_val_if_fail(net != NULL, NULL);
	return_val_if_fail(uplink == NULL || uplink->network == net, NULL);
        return_val_if_fail(name != NULL, NULL);

	server = p_new(server_pool, Server, 1);
	server->pool = pool_create(t_strdup_printf("Server %s", name),
				   SERVER_BUFFER_INITIAL_SIZE, TRUE);
        server->token = ++token_counter;
	server->network = net;
	server->name = p_strdup(server->pool, name);
	server->info = p_strdup(server->pool, info);
        server->hopcount = hopcount;
	server->nicks = hash_create(server->pool, 100, irc_nick_hash,
				    (HashCompareFunc) irc_nick_cmp);
	server->downlinks = parray_create(server->pool, 10);

	if (id == NULL)
                server->output_id = p_strdup(server->pool, server->name);
	else {
		server->uids = TRUE;
		server->output_id = p_strconcat(server->pool, "\r", id, NULL);
	}

	if (uplink != NULL) {
		uplink->servers_count++;
		server->uplink = uplink;
		parray_alloc(uplink->downlinks, server);
	}

        net->servers_count++;
	hash_insert(net->servers, server->name, server);

        if (server->uids)
		hash_insert(net->server_ids, server_get_id(server), server);
        return server;
}

static void net_remove_nick(const char *nick_name, Nick *nick, IrcNetwork *net)
{
	channels_remove_nick(nick);

        if (nick->server->uids)
		hash_remove(net->nick_ids, nick_get_id(nick));
	hash_remove(net->nicks, nick->nick);

	nick_unref(nick);
}

void server_destroy(Server *server)
{
        ServerConn *sconn;
        Server *link;
	int i;

        return_if_fail(server != NULL);

        server->network->servers_count--;

        /* remove the server and nicks from network */
	hash_foreach(server->nicks, (HashForeachFunc) net_remove_nick,
		     server->network);
	hash_destroy_clean(server->nicks);

        /* remove the server from it's uplink */
	if (server->uplink != NULL) {
                server->uplink->servers_count--;
		i = parray_find(server->uplink->downlinks, server);
		if (i != -1) parray_free(server->uplink->downlinks, i);
	}

	if (server->downlinks != NULL) {
		/* remove all downlinks */
                parray_freeze(server->downlinks);
		for (i = 0; i < server->downlinks->last; i++) {
			link = parray_index(server->downlinks, i);

			if (link != NULL) {
				server->servers_count--;
				server_destroy(link);
			}
		}
                parray_thaw(server->downlinks);

                parray_destroy_clean(server->downlinks);
	}

	if (server->linked) {
                sconn = server_conn_find(server);
                if (sconn != NULL) sconn->server = NULL;
	}

	hash_remove(server->network->servers, server->name);
        if (server->uids)
		hash_remove(server->network->server_ids, server_get_id(server));

	/* update counters */
        server->network->users_count -= server->users_count;
	server->network->operators_count -= server->operators_count;
	server->network->services_count -= server->services_count;

        p_free_clean(server->pool, server->output_id);
        p_free_clean(server->pool, server->name);
        p_free_clean(server->pool, server->info);

        pool_unref(server->pool);
        p_free(server_pool, server);
}

void server_add_nick(Server *server, Nick *nick)
{
	const char *id;

        id = nick_get_id(nick);

	server->users_count++;
        server->network->users_count++;

	nick->server = server;
        hash_insert(server->nicks, nick->nick, nick);
        hash_insert(server->network->nicks, nick->nick, nick);
        hash_insert(server->network->nick_ids, id, nick);
}

void server_remove_nick(Nick *nick)
{
	const char *id;

	if (nick->removed)
                return;

        id = nick_get_id(nick);

	nick->server->users_count--;
        nick->server->network->users_count--;

	hash_remove(nick->server->nicks, nick->nick);
	hash_remove(nick->server->network->nicks, nick->nick);
        hash_remove(nick->server->network->nick_ids, id);

	channels_remove_nick(nick);

        nick->removed = TRUE;
        nick_unref(nick);
}

void server_rename_nick(Nick *nick, const char *old_nick)
{
	if (old_nick != NULL) {
		hash_remove(nick->server->network->nicks, old_nick);
		hash_remove(nick->server->nicks, old_nick);
	}

	hash_insert(nick->server->network->nicks, nick->nick, nick);
	hash_insert(nick->server->nicks, nick->nick, nick);
}

Server *server_find_first_link(Server *server)
{
	return_val_if_fail(!server->local, NULL);

	while (server->uplink != server->network->server)
		server = server->uplink;

        return server;
}

int server_leaf_match(Server *server, unsigned int leaf_servers)
{
        unsigned int leaf_id;

	if (leaf_servers == 0 || !server->leaf || !server->linked)
		return FALSE;

	leaf_id = server_conn_find(server)->leaf_id;
        return (leaf_servers & (1 << leaf_id)) != 0;
}

int server_match(Server *server, const char *mask)
{
	if (irc_id_cmp(server_get_id(server), mask) == 0)
		return TRUE;

        return mask_match(mask, server->name);
}

#define SPACE_CHECK(len, log_args) STMT_START { \
	if ((len) > space) { \
		server_log log_args; \
		return; \
	} \
	} STMT_END

void server_send(Server *server, SendOptions *sopts, const char *data)
{
        ServerConn *sconn;
        Nick *nick;
        time_t timestamp;
	unsigned char *buffer, *p;
	const char *cmd_id, *origin_str;
        char uid[UID_MAX_LENGTH+1];
	int i, val, len, space, uids, uidpos, negative;

        return_if_fail(server != NULL);

	while (!server->linked && !server->local)
		server = server->uplink;

	if (!server->linked) {
		i_error("server_send() to local: %s", data);
                return;
	}

        sconn = server_conn_find(server);
	if (sconn->disconnected)
                return;

        /* check that wanted capabilities match */
	if (sopts != NULL &&
	    ((sconn->caps & sopts->caps_set) != sopts->caps_set ||
	     (sconn->caps & sopts->caps_unset) != 0))
                return;

	/* origin -> string */
        uids = (sconn->caps & CAP_UID) == CAP_UID;
	if (sopts == NULL)
		origin_str = NULL;
	else if (sopts->origin_nick != NULL) {
		origin_str = uids ? nick_get_id(sopts->origin_nick) :
			sopts->origin_nick->nick;
	} else if (sopts->origin_server != NULL) {
		origin_str = uids ? server_get_id(sopts->origin_server) :
			sopts->origin_server->name;
	} else if (sopts->origin != NULL)
                origin_str = sopts->origin;
	else
                origin_str = NULL;

	if (sopts == NULL) {
		cmd_id = NULL;
		timestamp = 0;
	} else {
		/* Command ID */
		if ((sconn->caps & CAP_CID) == 0)
			cmd_id = NULL;
		else if (sopts->origin_server != NULL &&
			 !sopts->origin_server->local) {
			ServerConn *origin;

			origin = server_conn_link_find(sopts->origin_server);
			cmd_id = origin->current_cmd_id;
		} else if (sopts->origin_nick != NULL &&
			   sopts->origin_nick->server->local) {
			Client *client;

                        client = client_find(sopts->origin_nick->nick);
			cmd_id = client->current_cmd_id;
		} else {
                        cmd_id = NULL;
		}

		/* timestamp */
		timestamp = (sconn->caps & CAP_TS) == 0 ||
			sopts->timestamp == 0 ? 0 :
			ioloop_time - sopts->timestamp;
	}

	/* save the line directly to transmit buffer, so if it needs to be
	   buffered there won't be any memcpy()ing. */
        buffer = p = tbuffer_get_space(sconn->outbuf, sconn->max_line_length+2);
	space = sconn->max_line_length;

	if (buffer == NULL) {
		server_disconnect(sconn, "Output buffer full");
		return;
	}

	while (origin_str != NULL || timestamp != 0 || cmd_id != NULL) {
                /* need to add prefix */
		*p++ = ':'; space--;

		if (origin_str != NULL) {
			/* origin given */
			len = strlen(origin_str);
			SPACE_CHECK(len, (server, "server_send(): origin too "
					  "long :%s %s", origin_str, data));
			memcpy(p, origin_str, len); p += len;
			space -= len;
		}

		if (timestamp == 0 && cmd_id == NULL)
			break;

		*p++ = ':'; space--;
		if (timestamp != 0) {
			/* timestamp given */
			SPACE_CHECK(9, (server, "server_send(): origin+ts "
					"too long :%s %s", origin_str, data));

			negative = timestamp < 0;
			if (negative) {
				timestamp = -timestamp;
                                *p++ = '-';
			}

			for (i = 28; i > 0; i -= 4)
				if ((timestamp >> i) != 0)
					break;

			while (i >= 0) {
				val = (timestamp >> i) & 0xf;
				*p++ = val < 10 ? val + '0' : val - 10 + 'a';
                                space--;
                                i -= 4;
			}
		}

		if (cmd_id == NULL)
			break;

                len = strlen(cmd_id);
		SPACE_CHECK(len, (server,
				  "server_send(): cmdid too long :%s::%s %s",
				  origin_str, cmd_id, data));
		memcpy(p, cmd_id, len); p += len;
		space -= len;

		break;
	}

	SPACE_CHECK(10, (server, "server_send(): not enough space :%s::%s %s",
                         origin_str, cmd_id, data));
	if (p != buffer) {
		*p++ = ' ';
		space--;
	}

	/* add the actual data - replace all \rUIDs */
	while (space > 0 && *data != '\0') {
		if (*data != '\r') {
			*p++ = *data++;
			space--;
                        continue;
		}
		data++;

		if (uids) {
                        /* UIDs in use, just remove \r */
                        continue;
		}

		/* replace UID with nick */
                uidpos = 0;
		while (*data != '\0' && *data != ' ' && *data != ',' &&
		       uidpos < UID_MAX_LENGTH) {
                        uid[uidpos++] = *data++;
		}
                uid[uidpos] = '\0';

		nick = ircnet_find_nick_id(sconn->network, uid);
		if (nick == NULL) {
			server_log(sconn->server,
				   "server_send(): Unknown UID '%s'", uid);
                        return;
		}

                len = strlen(nick->nick);
		if (len > space) {
			server_log(sconn->server, "server_send(): Nick too "
				   "long '%s'", nick->nick);
                        return;
		}

		memcpy(p, nick->nick, len);
                p += len; space -= len;
	}

        /* CR + LF */
	*p++ = '\r';
        *p++ = '\n';

	if (tbuffer_send_buffer(sconn->outbuf, (int) (p-buffer)) < 0)
		server_disconnect(sconn, strerror(errno));
}

void server_send_all(IrcNetwork *network, SendOptions *sopts, const char *data)
{
	Server *server;
        PArray links;
        int i;

        links = ircnet_links(network);
	if (links == NULL)
		return;

        parray_freeze(links);
	for (i = 0; i < links->last; i++) {
		server = parray_index(links, i);

		if (server != NULL &&
		    (sopts == NULL || server != sopts->skip_server))
			server_send(server, sopts, data);
	}
        parray_thaw(links);
}

void server_send_nick_update(Nick *nick, SendOptions *sopts, const char *data)
{
        Server *server, *skip_server;
	PArray links;
	int i;

        skip_server = NULL;
	if (sopts != NULL) {
		if (sopts->skip_server != NULL)
			skip_server = sopts->skip_server;
                else if (sopts->origin_server != NULL &&
			 !sopts->origin_server->local) {
			skip_server = sopts->origin_nick != NULL ?
				sopts->origin_nick->server :
				sopts->origin_server;
			skip_server = server_find_first_link(skip_server);
		}
	}

        links = ircnet_links(nick->server->network);
	if (links == NULL)
		return;

        parray_freeze(links);
	for (i = 0; i < links->last; i++) {
		server = parray_index(links, i);

		if (server == NULL || server == skip_server)
			continue;

		if (server->leaf &&
		    !server_leaf_match(server, nick->leaf_servers))
                        continue;

		server_send(server, sopts, data);
	}
        parray_thaw(links);
}

void server_send_serverjoin(Server *dest_server, Server *server)
{
        SendOptions sopts;
	const char *data;

	return_if_fail(dest_server->linked);

        t_push();
	if ((server_conn_find(dest_server)->caps & CAP_UID) == 0) {
                /* no UIDs */
		data = t_strdup_printf("SERVER %s %d %d :%s",
				       server->name, server->hopcount+1,
				       server->token, server->info);
	} else if (server->uids) {
                /* both servers support UIDs */
		data = t_strdup_printf("SERVER %s %d %s :%s",
				       server->name, server->hopcount+1,
				       server_get_id(server), server->info);
	} else {
		/* dest supports UIDs, but the server we're
		   introducing doesn't */
		data = t_strdup_printf("SERVER %s %d $%d :%s",
				       server->name, server->hopcount+1,
				       server->token, server->info);
	}

	send_options_clear(&sopts);
        sopts.timestamp = server->created;

	server_send(dest_server, &sopts, data);
        t_pop();
}

void server_send_nickjoin(Server *server, Nick *nick)
{
        SendOptions sopts;
	const char *data, *umode;
        char ip_str[MAX_IP_LEN];

	return_if_fail(server->linked);

	if (server->leaf) {
		/* set the leaf server to be notified about
		   changes in this nick */
		nick->leaf_servers |= 1 << server_conn_find(server)->leaf_id;
	}

	umode = nick_get_umode_str(nick);
	if (*umode == '\0') umode = "+";

        t_push();
	if (server->uids && nick->server->uids) {
                net_ip2host(&nick->ip, ip_str);
		data = t_strdup_printf("UNICK %s %s %s %s %s %s :%s",
				       nick->nick, nick_get_id(nick),
				       nick->user, nick->host, ip_str,
				       umode, nick->real_name);
	} else {
		data = t_strdup_printf("NICK %s %d %s %s %d %s :%s",
				       nick->nick, nick->hopcount+1,
				       nick->user, nick->host,
				       nick->server->token,
				       umode, nick->real_name);
	}

	send_origin_server(&sopts, nick->server);
        sopts.timestamp = nick->created;

	server_send(server, &sopts, data);
        t_pop();
}

void server_send_nickjoins(Nick *nick)
{
        Server *server, *link_server;
	PArray links;
        int i;

	link_server = nick->server->local ? NULL :
		server_find_first_link(nick->server);

	links = ircnet_links(nick->server->network);
	if (links == NULL)
		return;

        parray_freeze(links);
	for (i = 0; i < links->last; i++) {
		server = parray_index(links, i);

		if (server == NULL || server == link_server)
			continue;

		if (server->leaf && server->linked &&
		    server_leaf_find_unknown_nick(server, nick->nick)) {
                        /* leaf server wanted to know about this nick */
			server_leaf_remove_unknown_nick(server, nick->nick);
			server_send_nickjoin(server, nick);
		} else if (!server->leaf) {
                        /* always send to non-leafs */
			server_send_nickjoin(server, nick);
		}
	}
        parray_thaw(links);
}

int server_check_route(Server *server, Server *link)
{
	return_val_if_fail(link->linked, FALSE);

	if (server_find_first_link(server) == link)
                return TRUE;

	server_log(link, "invalid route to %s", server->name);
        return FALSE;
}

int server_check_noroute(Server *server, Server *link)
{
	return_val_if_fail(link->linked, FALSE);

	if (server_find_first_link(server) != link)
                return TRUE;

	server_log(link, "invalid route to %s", server->name);
        return FALSE;
}

int server_name_cmp(const void *p1, const void *p2)
{
        return irc_case_cmp(((Server *)p1)->name, ((Server *)p2)->name);
}

void servers_init(void)
{
	server_pool = pool_create("Servers", 128, TRUE);
}

void servers_deinit(void)
{
        pool_unref(server_pool);
}
