#ifndef __SERVER_H
#define __SERVER_H

#include "send.h"

struct _Server {
	IrcNetwork *network;
        char *output_id; /* \rSID, or server name if it doesn't support UIDs */

	char *name;
        char *info;
	int token; /* unique token which we use to reference this server
		to our downlinks that don't understand SIDs - for uplinks
		you need to use the ServerConn->tokens table for converting. */
	int hopcount; /* 0 = ourself, 1+ = other servers. */
	time_t created; /* server start time, or if not known,
		the record creation time */

	unsigned int last_id; /* last command ID counter we received */

	unsigned int users_count, operators_count, channels_count;
	unsigned int services_count, servers_count;

	unsigned int local:1; /* local server == that's us */
	unsigned int linked:1; /* we're linked directly to this server */
	unsigned int leaf:1; /* this is a leaf server */
	unsigned int uids:1; /* server supports UIDs */
	unsigned int masked:1; /* behind server mask */
	unsigned int eob:1; /* received EOB from server */

	/* up- and downlinks are relative to us, we'll find the path to
	   the server just by following the uplinks. */
        struct _Server *uplink;
	PArray downlinks;

        ServerConn *sconn; /* use server_conn_find() instead of this directly */
	HashTable *nicks; /* nick name -> Nick mappings */
	Pool pool;
};

/* Add a new server to IRC network. NOTE: if id == NULL, the server is
   assumed not to support UIDs. */
Server *server_create(IrcNetwork *net, Server *uplink, const char *id,
		      const char *name, const char *info, int hopcount);
/* Remove server from network. Removes all it's downlinks as well. */
void server_destroy(Server *server);

/* Add nick to network in specified server. */
void server_add_nick(Server *server, Nick *nick);
/* Remove nick from network. */
void server_remove_nick(Nick *nick);
/* Rename nick. Use client_set_nick() for local clients instead of this. */
void server_rename_nick(Nick *nick, const char *old_nick);

/* Returns the server ID. */
#define server_get_id(server) \
	((server)->output_id[0] == '\r' ? \
	(server)->output_id+1 : (server)->output_id)
/* Returns the server's string ID that can be used in all
   strings to reference the server (\rSID) */
#define server_get_outputid(server) ((server)->output_id)

/* Find the first server we're connected to in the path to the server */
Server *server_find_first_link(Server *server);
/* Returns TRUE if the specified server is in leaf_servers bitmask. */
int server_leaf_match(Server *server, unsigned int leaf_servers);
/* Returns TRUE if the server's name matches for the mask, or if the
   mask is the same as server's ID. */
int server_match(Server *server, const char *mask);

/* Send data to server from specified nick/server.
   The prefix is added if needed, and CR+LF appended. */
void server_send(Server *server, SendOptions *sopts, const char *data);
/* Send data to all connected servers. */
void server_send_all(IrcNetwork *network, SendOptions *sopts, const char *data);
/* Send data to all servers that need to know about changes in the nick. */
void server_send_nick_update(Nick *nick, SendOptions *sopts, const char *data);

/* Send information of server to specified server (SERVER command) */
void server_send_serverjoin(Server *dest_server, Server *server);
/* Send information of nick to specified server (NICK command). */
void server_send_nickjoin(Server *server, Nick *nick);
/* Send information of nick to all servers except the link where the
   nick came from. */
void server_send_nickjoins(Nick *nick);

/* Returns TRUE if the given server is routed through given link.
   Besides returning FALSE, the error is also logged. */
int server_check_route(Server *server, Server *link);
/* Like server_check_route(), but the server should NOT be routed
   through given link. */
int server_check_noroute(Server *server, Server *link);

/* Server->name comparing for hashes/trees */
int server_name_cmp(const void *p1, const void *p2);

void servers_init(void);
void servers_deinit(void);

#endif
