#include "lib.h"
#include "internal.h"

#include <stdio.h>

#define STATE_NAME	0x01
#define STATE_VALUE	0x02
#define STATE_SET	0x10

struct _ConfigParser {
	Config *config;

	ConfigBlock *block;
	ConfigItem *last_item, *current_item;

        char *key;
	ConfigList *name;

        int state;
	int last_token_type;
        int line;

	unsigned int error:1;
	unsigned int warnings:1;
};

static void config_parse_error(ConfigParser *parser, const char *error)
{
	/* FIXME: save the error somewhere */
	parser->error = TRUE;

	fprintf(stderr, "Line %d: %s\n", parser->line, error);
}

static void config_parse_error_not_expect(ConfigParser *parser,
					  const char *token)
{
	/* FIXME: save the error somewhere */
	parser->error = TRUE;

	fprintf(stderr, "Line %d: not expecting '%s'\n", parser->line, token);
}

static void config_parse_warning_expect(ConfigParser *parser, const char *token)
{
	/* FIXME: save the warnings somewhere */
        parser->warnings = TRUE;

	fprintf(stderr, "Line %d: missing '%s', ignoring\n",
		parser->line, token);
}

static ConfigItem *config_item_find(ConfigParser *parser, const char *key)
{
	ConfigItem *item;

	for (item = parser->block->items; item != NULL;
	     item = item->next) {
		if (item->type == CONFIG_ITEM_TYPE_VALUE &&
		    strcasecmp(item->key, parser->key) == 0)
			return item;
	}

        return item;
}

static ConfigItem *config_item_add(ConfigParser *parser, int type)
{
	ConfigItem *item;

	if (type == CONFIG_ITEM_TYPE_VALUE) {
		/* if item with same key already exists, use it */
		item = config_item_find(parser, parser->key);
		if (item != NULL) {
			p_free(parser->config->pool, parser->key);
			parser->key = NULL;
			parser->current_item = item;
                        return item;
		}
	}

	item = p_new(parser->config->pool, ConfigItem, 1);
	item->type = type;

	item->key = parser->key;
        parser->key = NULL;

        if (parser->last_item != NULL)
		parser->last_item->next = item;
	else
		parser->block->items = item;

	parser->current_item = parser->last_item = item;
        return item;
}

static void config_block_add(ConfigParser *parser)
{
	ConfigBlock *block;
	ConfigItem *item;

	item = config_item_add(parser, CONFIG_ITEM_TYPE_BLOCK);

	block = p_new(parser->config->pool, ConfigBlock, 1);
	block->parent = parser->block;
	item->value.block = block;

        block->name = parser->name;
        parser->name = NULL;

        parser->block = block;
	parser->last_item = NULL;
}

static void config_parse_parent(ConfigParser *parser)
{
	parser->block = parser->block->parent;

	parser->last_item = parser->block->items;
	if (parser->last_item != NULL) {
		while (parser->last_item->next != NULL)
                        parser->last_item = parser->last_item->next;
	}
}

static void config_parse_add_list(ConfigParser *parser, ConfigList **list,
				  const char *value)
{
	ConfigList *rec, **tmp;

	rec = p_new(parser->config->pool, ConfigList, 1);
	rec->data = p_strdup(parser->config->pool, value);

	if (*list == NULL)
		*list = rec;
	else {
		for (tmp = list; *tmp != NULL; tmp = &(*tmp)->next) ;
                *tmp = rec;
	}
}

static void config_parse_reset_state(ConfigParser *parser)
{
        p_free(parser->config->pool, parser->key);
	config_list_destroy(parser->config, parser->name);

        parser->key = NULL;
	parser->name = NULL;
	parser->state = 0;
}

void config_parse_token(ConfigParser *parser, int token_type, const char *token)
{
	ConfigItem *item;
        int len;

	if (parser->error) {
                /* fatal error */
		return;
	}

	switch (token_type) {
	case TOKEN_TYPE_NEWLINE:
                parser->line++;
		if (parser->last_token_type != TOKEN_TYPE_NEWLINE)
			break;
                /* fall through - treat empty lines as comments */
	case TOKEN_TYPE_COMMENT:
		item = config_item_add(parser, CONFIG_ITEM_TYPE_COMMENT);
                item->value.comment = p_strdup(parser->config->pool, token);
		break;
	case TOKEN_TYPE_STRING:
		if ((parser->state & (STATE_VALUE|STATE_SET)) ==
		    (STATE_VALUE|STATE_SET)) {
			config_parse_warning_expect(parser, ";");
                        config_parse_reset_state(parser);
		}

                t_push();
		if (*token == '"') {
			/* quoted string, remove them */
			token++;
			len = strlen(token);
			if (len < 1 || token[len-1] != '"')
				config_parse_warning_expect(parser, "\"");
			else
                                token = t_strndup(token, len-1);
		}

		if (parser->state == 0) {
			/* key */
			parser->key = p_strdup(parser->config->pool, token);
                        parser->state = STATE_NAME;
		} else if (parser->state == STATE_NAME) {
			/* name */
                        parser->state |= STATE_SET;
			config_parse_add_list(parser, &parser->name, token);
		} else if (parser->state == STATE_VALUE) {
                        /* value */
                        parser->state |= STATE_SET;
			config_parse_add_list(parser,
					      &parser->current_item->value.list,
					      token);
		} else {
                        /* error */
			config_parse_error_not_expect(parser, token);
		}
                t_pop();
		break;
	case TOKEN_TYPE_PUNCT:
		switch (*token) {
		case '{':
			if ((parser->state & STATE_NAME) == 0) {
				config_parse_error_not_expect(parser, "{");
                                break;
			}

			config_block_add(parser);
                        config_parse_reset_state(parser);
                        break;
		case '}':
			if (parser->state != 0) {
				if ((parser->state & STATE_VALUE) == 0 ||
				    (parser->state & STATE_SET) == 0) {
					config_parse_error_not_expect(parser,
								      "}");
					break;
				}

				/* foo = bar } */
				config_parse_warning_expect(parser, ";");
				config_parse_reset_state(parser);
			}

			if (parser->block->parent == NULL) {
				/* too many '}' */
				config_parse_error(parser,
						   "missing matching '{'");
                                break;
			}

                        config_parse_parent(parser);
                        break;
		case ',':
			if ((parser->state & (STATE_NAME|STATE_VALUE)) == 0)
				config_parse_error_not_expect(parser, ",");

                        parser->state &= ~STATE_SET;
			break;
		case '=':
			if ((parser->state & STATE_NAME) == 0) {
				config_parse_error_not_expect(parser, "=");
                                break;
			}

			config_item_add(parser, CONFIG_ITEM_TYPE_VALUE);
			parser->state = STATE_VALUE;
			break;
		case ';':
			if ((parser->state & STATE_VALUE) == 0 ||
			    (parser->state & STATE_SET) == 0) {
				config_parse_error_not_expect(parser, ";");
                                break;
			}

                        config_parse_reset_state(parser);
                        break;
		default:
                        i_error("shouldn't happen");
		}
		break;
	}

        parser->last_token_type = token_type;
}

int config_parse(Config *config)
{
	ConfigParser p;

	memset(&p, 0, sizeof(p));
	p.config = config;
	p.block = config->root;
        p.line = 1;

	if (!config_lex_parse(&p, config->fname))
                return -1;

	return p.error ? 0 : 1;
}
