#include "lib.h"
#include "internal.h"

#include <stdio.h>
#include <ctype.h>

#define INDENT_CHAR	'\t'
#define INDENT_LEVEL	1

typedef struct {
	FILE *f;

        int indent_char;
	int indent_pos;
	int indent_level;

	int error;
} ConfigWriter;

static void write_line(ConfigWriter *writer, const char *data)
{
	int i;

	if (writer->error)
                return;

	if (!is_empty_str(data)) {
		for (i = 0; i < writer->indent_pos; i++)
			fputc(writer->indent_char, writer->f);
		fputs(data, writer->f);
	}

	if (fputc('\n', writer->f) == EOF)
		writer->error = errno;
}

static int string_need_quotes(const char *str)
{
	/* quote only if there's any non-normal characters */
	while (*str != '\0') {
		if (!i_isalnum(*str) && *str != '!' &&
		    *str != '.' && *str != '/')
			return TRUE;
		str++;
	}

        return FALSE;
}

static const char *config_list_get_string(ConfigList *list)
{
        char *buffer;
        int datalen, len, alloc_len, quote;

	if (list == NULL)
		return "";

        len = 0; alloc_len = 1024;
	buffer = t_buffer_get(alloc_len);
	while (list != NULL) {
		datalen = strlen(list->data);
		if (len+datalen+4 > alloc_len) {
                        alloc_len += datalen+2;
			buffer = t_buffer_reget(buffer, alloc_len);
		}

		if (len != 0) {
                        memcpy(buffer+len, ", ", 2);
			len += 2;
		}

		quote = string_need_quotes(list->data);
		if (quote) buffer[len++] = '"';
		memcpy(buffer+len, list->data, datalen);
		len += datalen;
		if (quote) buffer[len++] = '"';

                list = list->next;
	}

	buffer[len] = '\0';
        t_buffer_alloc(len+1);
        return buffer;
}

static void config_write_items(ConfigWriter *writer, ConfigBlock *block)
{
        ConfigBlock *subblock;
	ConfigItem *item;
        const char *data;

	for (item = block->items; item != NULL; item = item->next) {
		switch (item->type) {
		case CONFIG_ITEM_TYPE_COMMENT:
			write_line(writer, item->value.comment);
			break;
		case CONFIG_ITEM_TYPE_BLOCK:
			subblock = item->value.block;

                        /* header */
			t_push();
			data = t_strdup_printf("%s %s {", item->key,
				config_list_get_string(subblock->name));
			write_line(writer, data);
                        t_pop();

                        /* items */
                        writer->indent_pos += writer->indent_level;
			config_write_items(writer, subblock);
			writer->indent_pos -= writer->indent_level;

                        /* footer */
			write_line(writer, "}");
                        break;
		case CONFIG_ITEM_TYPE_VALUE:
			t_push();
			data = t_strdup_printf("%s = %s;", item->key,
				config_list_get_string(item->value.list));
			write_line(writer, data);
                        t_pop();
			break;
		}
	}
}

int config_write(Config *config, const char *fname)
{
	ConfigWriter writer;

	if (fname == NULL)
                fname = config->fname;

	writer.f = fopen(fname, "w");
	if (writer.f == NULL)
		return errno;

        writer.indent_char = INDENT_CHAR;
	writer.indent_pos = 0;
	writer.indent_level = INDENT_LEVEL;
        writer.error = 0;

	config_write_items(&writer, config->root);

        fclose(writer.f);
        return writer.error;
}
