/*
   tbuffer.c : Input/output transmit buffer handling

    Copyright (c) 2002 Timo Sirainen

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "lib.h"
#include "ioloop.h"
#include "tbuffer.h"
#include "network.h"

#include <stdio.h>
#include <unistd.h>

TBuffer *tbuffer_create(int handle, Pool pool, int priority,
			unsigned int max_size)
{
	TBuffer *buf;

        return_val_if_fail(handle >= 0, NULL);
        return_val_if_fail(pool != NULL, NULL);

	buf = p_new(pool, TBuffer, 1);
	buf->handle = handle;
	buf->pool = pool;
	buf->priority = priority;
	buf->max_size = max_size;
	return buf;
}

TBuffer *tbuffer_create_file(int handle, Pool pool, unsigned int max_size)
{
	TBuffer *buf;

	buf = tbuffer_create(handle, pool, IO_PRIORITY_DEFAULT, max_size);
	buf->file = TRUE;
        return buf;
}

void tbuffer_destroy(TBuffer *buf)
{
	if (buf == NULL)
		return;

        if (buf->io != NULL)
		io_remove(buf->io);
        p_free(buf->pool, buf->buffer);
        p_free(buf->pool, buf);
}

void tbuffer_close(TBuffer *buf)
{
	if (buf == NULL)
		return;

        buf->closed = TRUE;
}

void tbuffer_reset(TBuffer *buf)
{
	buf->pos = buf->skip = buf->cr_lookup_pos = 0;
        buf->last_cr = FALSE;
}

TBuffer *tbuffer_set_pool(TBuffer *buf, Pool pool)
{
	TBuffer *newbuf;

        return_val_if_fail(buf != NULL, NULL);
        return_val_if_fail(pool != NULL, NULL);

	newbuf = p_new(pool, TBuffer, 1);
	memcpy(newbuf, buf, sizeof(TBuffer));

	newbuf->pool = pool;
	newbuf->buffer = p_malloc(pool, buf->size);
	memcpy(newbuf->buffer, buf->buffer + buf->skip,
	       buf->size - buf->skip);

	newbuf->cr_lookup_pos -= newbuf->skip;
        newbuf->pos -= newbuf->skip;
	newbuf->skip = 0;

        p_free(buf->pool, buf->buffer);
        p_free(buf->pool, buf);
        return newbuf;
}

void tbuffer_set_max_size(TBuffer *buf, unsigned int max_size)
{
        buf->max_size = max_size;
}

static int my_write(int handle, const void *buf, int size)
{
	int ret;

	ret = write(handle, buf, size);

	if (ret == 0)
		ret = -1; /* EOF */
	else if (ret < 0 && (errno == EINTR || errno == EAGAIN))
		ret = 0;

        return ret;
}

static int buf_send(TBuffer *buf)
{
	int ret;

	if (!buf->file) {
		ret = net_transmit(buf->handle, buf->buffer + buf->skip,
				   buf->pos - buf->skip);
	} else {
		ret = my_write(buf->handle, buf->buffer + buf->skip,
			       buf->pos - buf->skip);
	}

	if (ret == -1) {
		buf->closed = TRUE;
	} else {
		buf->transfd += ret;
		buf->skip += ret;
		if (buf->skip == buf->pos) {
			/* everything sent */
			buf->skip = buf->pos = 0;
		}
	}

	if (buf->closed || buf->pos == 0) {
		io_remove(buf->io);
                buf->io = NULL;
		return FALSE;
	}

        return TRUE;
}

static void buffer_alloc_more(TBuffer *buf, int len)
{
	buf->size = buf->pos+len;
	buf->size = buf->size <= TBUFFER_MIN_SIZE ? TBUFFER_MIN_SIZE :
		nearest_power(buf->size);

	if (buf->max_size > 0 && buf->size > buf->max_size)
		buf->size = buf->max_size;

	buf->buffer = p_realloc(buf->pool, buf->buffer, buf->size);
	if (buf->buffer == NULL) {
		/* pool limit exceeded */
		buf->pos = buf->size = 0;
	}
}

static inline void tbuffer_compress(TBuffer *buf)
{
	memmove(buf->buffer, buf->buffer + buf->skip,
		buf->pos - buf->skip);
	buf->pos -= buf->skip;

	if (buf->skip > buf->cr_lookup_pos)
		buf->cr_lookup_pos = 0;
	else
		buf->cr_lookup_pos -= buf->skip;

	buf->skip = 0;
}

int tbuffer_send(TBuffer *buf, const void *data, int len)
{
	int ret;

	return_val_if_fail(!buf->receive, 0);
        return_val_if_fail(data != NULL, 0);
	buf->transmit = TRUE;

	if (buf->closed)
                return -1;

        if (buf->dump)
		printf("--> %.*s\n", len-2, (const char *) data);

	if (buf->pos == 0) {
		/* buffer is empty, try to send the data immediately */
		ret = buf->file ? my_write(buf->handle, data, len) :
			net_transmit(buf->handle, data, len);
		if (ret == -1) {
			/* disconnected */
			buf->closed = TRUE;
			return -1;
		}

		buf->transfd += ret;
		data = (const char *) data + ret;
                len -= ret;
	}

	if (len <= 0)
		return 1;

	if (tbuffer_get_space(buf, len) == NULL)
		return -2;

	/* add to buffer */
	memcpy(buf->buffer + buf->pos, data, len);
	buf->pos += len;

	if (buf->io == NULL) {
		buf->io = io_add_priority(buf->handle, buf->priority, IO_WRITE,
					  (IOFunc) buf_send, buf);
	}
        return 1;
}

void tbuffer_send_flush(TBuffer *buf)
{
	return_if_fail(!buf->receive);

	if (buf->closed || buf->io == NULL)
                return;

        /* set the socket blocking while doing this */
        if (!buf->file) net_set_nonblock(buf->handle, FALSE);
	while (buf_send(buf)) ;
        if (!buf->file) net_set_nonblock(buf->handle, TRUE);
}

int tbuffer_read_max(TBuffer *buf, int size)
{
	int ret;

	return_val_if_fail(!buf->transmit, 0);
	buf->receive = TRUE;

	if (buf->closed)
                return -1;

	if (buf->pos == buf->size) {
		if (buf->skip > 0) {
			/* remove the unused bytes from beginning of buffer */
                        tbuffer_compress(buf);
		} else if (buf->max_size == 0 || buf->size < buf->max_size) {
			/* buffer is full - grow it */
			buffer_alloc_more(buf, TBUFFER_READ_GROW_SIZE);
		}

		if (buf->pos == buf->size)
                        return -2; /* buffer full */
	}

        /* fill the buffer */
	if (size == -1 || buf->size-buf->pos < (unsigned int) size)
		size = buf->size - buf->pos;
	if (!buf->file) {
		ret = net_receive(buf->handle, buf->buffer + buf->pos,
				  buf->size - buf->pos);
	} else {
                ret = read(buf->handle, buf->buffer + buf->pos,
			   buf->size - buf->pos);
		if (ret == 0)
			ret = -1; /* EOF */
		else if (ret < 0 && (errno == EINTR || errno == EAGAIN))
                        ret = 0;
	}

	if (ret == -1) {
		/* disconnected */
                return -1;
	}

        buf->transfd += ret;
	buf->pos += ret;
        return ret;
}

int tbuffer_read(TBuffer *buf)
{
        return tbuffer_read_max(buf, -1);
}

/* skip the first LF, if it exists */
static inline void tbuffer_skip_lf(TBuffer *buf)
{
	if (!buf->last_cr || buf->skip >= buf->pos)
		return;

	if (buf->buffer[buf->skip] == 10) {
		if (buf->skip == buf->cr_lookup_pos)
			buf->cr_lookup_pos++;
		buf->skip++;
	}
	buf->last_cr = FALSE;
}

unsigned char *tbuffer_next_line(TBuffer *buf)
{
	unsigned char *ret_buf;
        unsigned int i;

        return_val_if_fail(buf != NULL, NULL);

	tbuffer_skip_lf(buf);
	if (buf->skip >= buf->pos)
		return NULL;

	ret_buf = NULL;
	for (i = buf->cr_lookup_pos; i < buf->pos; i++) {
		if (buf->buffer[i] == 13 || buf->buffer[i] == 10) {
			/* got it */
                        buf->last_cr = buf->buffer[i] == 13;
			buf->buffer[i] = '\0';
			ret_buf = buf->buffer + buf->skip;

			i++;
			buf->skip = i;
                        break;
		}
	}

	if (buf->dump && ret_buf != NULL)
		printf("<-- %s\n", ret_buf);

	buf->cr_lookup_pos = i;
        return ret_buf;
}

unsigned char *tbuffer_get_data(TBuffer *buf, int *size)
{
	tbuffer_skip_lf(buf);
	if (buf->skip >= buf->pos)
		return NULL;

        *size = buf->pos - buf->skip;
        return buf->buffer + buf->skip;
}

unsigned char *tbuffer_get_space(TBuffer *buf, int size)
{
	unsigned int needed_size;

	return_val_if_fail(!buf->receive, 0);
	buf->transmit = TRUE;

        /* make sure we have enough space in buffer */
	needed_size = buf->pos + size;
	if (needed_size > buf->size && buf->skip > 0) {
		/* remove the unused bytes from beginning of buffer */
		tbuffer_compress(buf);
		needed_size = buf->pos + size;
	}

	if (needed_size > buf->size &&
	    (buf->max_size == 0 || (needed_size <= buf->max_size))) {
		/* allocate more space */
                buffer_alloc_more(buf, size);
	}

	if (needed_size > buf->size)
                return NULL;

        return buf->buffer + buf->pos;
}

int tbuffer_send_buffer(TBuffer *buf, int size)
{
	int ret;

	return_val_if_fail(!buf->receive, 0);

        if (buf->dump)
		printf("--> %.*s\n", size-2, (const char *) buf->buffer);

	if (buf->pos == 0) {
		/* buffer is empty, try to send the data immediately */
		ret = buf->file ? my_write(buf->handle, buf->buffer, size) :
			net_transmit(buf->handle, buf->buffer, size);
		if (ret == -1) {
			/* disconnected */
                        buf->closed = TRUE;
			return -1;
		}

		buf->transfd += ret;
		if (ret == size) {
                        /* all sent */
			return 1;
		}

		buf->skip += ret;
	}

	buf->pos += size;

	if (buf->io == NULL) {
		buf->io = io_add_priority(buf->handle, buf->priority, IO_WRITE,
					  (IOFunc) buf_send, buf);
	}

        return 1;
}

int tbuffer_set_data(TBuffer *buf, const void *data, int len)
{
	tbuffer_reset(buf);

	if ((int) buf->size < len) {
		buffer_alloc_more(buf, len);
		if ((int) buf->size < len)
                        return -2;
	}

	memcpy(buf->buffer, data, len);
	buf->pos = len;
        buf->transfd += len;
        return 1;
}

int tbuffer_is_empty(TBuffer *buf)
{
        return buf->skip >= buf->pos;
}
