#ifndef __TBUFFER_H
#define __TBUFFER_H

#define TBUFFER_MIN_SIZE	512
#define TBUFFER_READ_GROW_SIZE	256

struct _TBuffer {
	int handle;
        IO io;

	Pool pool;
        int priority;

	unsigned char *buffer;

        unsigned int cr_lookup_pos; /* used only when reading a line */
	unsigned int pos, skip;
	unsigned int size, max_size;
        unsigned int transfd;

	unsigned int dump:1; /* DEBUG: dump input/output to stdout */
	unsigned int file:1; /* reading/writing a file */
	unsigned int closed:1; /* all further read/writes will return 0 */
	unsigned int transmit:1; /* this is a transmit buffer */
	unsigned int receive:1; /* this is a receive buffer */
	unsigned int last_cr:1; /* we're expecting a LF to be skipped in
		next call to tbuffer_next_line() */
};

/* Create a buffer. It can be used for either sending or receiving data,
   NEVER BOTH AT SAME TIME. If max_size is 0, there's no limit. */
TBuffer *tbuffer_create(int handle, Pool pool, int priority,
			unsigned int max_size);
/* Same as tbuffer_create(), but specify that we're reading/writing file. */
TBuffer *tbuffer_create_file(int handle, Pool pool, unsigned int max_size);
/* Destroy a buffer. */
void tbuffer_destroy(TBuffer *buf);
/* Mark the buffer closed. Any sends/reads after this will return -1.
   The data already in buffer can be used, and the remaining output buffer
   will be sent. */
void tbuffer_close(TBuffer *buf);
/* Reset all pointers so that the buffer looks empty, the actual data is
   not touched and can be used. */
void tbuffer_reset(TBuffer *buf);

/* Change the memory pool used by the buffer. Data already in
   buffer will be transferred to new buffer. */
TBuffer *tbuffer_set_pool(TBuffer *buf, Pool pool);
/* Change the maximum size for buffer to grow. */
void tbuffer_set_max_size(TBuffer *buf, unsigned int max_size);

/* Returns 1 if all was ok, -1 if disconnected, -2 if buffer is full */
int tbuffer_send(TBuffer *buf, const void *data, int len);
/* Flush the output buffer, blocks until all is sent. */
void tbuffer_send_flush(TBuffer *buf);

/* Returns number of bytes read if read was ok,
   -1 if disconnected, -2 if the buffer is full */
int tbuffer_read(TBuffer *buf);
/* Like tbuffer_read(), but don't read more than specified size. */
int tbuffer_read_max(TBuffer *buf, int size);
/* Returns the next line from input buffer, or NULL if more data is needed
   to make a full line. NOTE: call to tbuffer_read() invalidates the
   returned data. */
unsigned char *tbuffer_next_line(TBuffer *buf);
/* Returns pointer to beginning of unused data in buffer,
   or NULL if there's no data. */
unsigned char *tbuffer_get_data(TBuffer *buf, int *size);

/* Returns a pointer to buffer wanted amount of space,
   or NULL if size is too big. */
unsigned char *tbuffer_get_space(TBuffer *buf, int size);
/* Send data saved to buffer from tbuffer_get_space().
   Returns -1 if disconnected. */
int tbuffer_send_buffer(TBuffer *buf, int size);

/* Put data to buffer as if it was received.
   Returns 1 if successful, -2 if buffer isn't big enough. */
int tbuffer_set_data(TBuffer *buf, const void *data, int len);
/* Returns TRUE if there's nothing in buffer. */
int tbuffer_is_empty(TBuffer *buf);

#endif
