#include "common.h"
#include "numeric-strings.h"
#include "client-commands.h"

static Pool pool;
static CommandList *client_commands;

void client_command_register(const char *name, int min_params,
			     ClientCommandFunc func)
{
        command_register(client_commands, name, min_params, func);
}

void client_command_unregister(const char *name, ClientCommandFunc func)
{
        command_unregister(client_commands, name, func);
}

void client_command_parse(Client *client, const char *name,
			  int argc, const char *args[])
{
        Command *cmd;
	ClientCommandFunc func;

        cmd = command_find(client_commands, name);
	if (cmd == NULL) {
		client_send_numeric(client, ERR_UNKNOWNCOMMAND, name);
                return;
	}

	if (cmd->min_params > 0 && args[cmd->min_params-1] == NULL) {
		client_send_numeric(client, ERR_NEEDMOREPARAMS, name);
                return;
	}
	func = cmd->func;

	func(client, name, argc, args);
}

CLIENT_COMMAND(cmd_self)
{
	server_send(client->server, client->current_cmd_line);
}

static void client_restricted(Client *client)
{
	const char *data;

	data = t_strdup_printf("NOTICE %s :You're allowed to talk only in %s - "
			       "all other channels and private messages are "
			       "disabled to prevent abuse. If you wish to talk "
			       "with someone privately, ask him/her in %s "
			       "to DCC CHAT you.",
			       client->nick, client->network->channel,
			       client->network->channel);
	client_send(client, data);
}

CLIENT_COMMAND(cmd_talk)
{
	int len;

	if (strcasecmp(args[0], client->network->address) == 0) {
		len = strlen(args[1]);
		if (len > 10 && strncasecmp(args[1], "\001VERSION", 8) == 0) {
			/* reply to our ctcp version query */
			client_set_version(client, t_strndup(args[1]+9, len-10));
		}
		return;
	}

	if (strcasecmp(args[0], client->network->channel) != 0) {
		client_restricted(client);
		return;
	}

	server_send(client->server, client->current_cmd_line);
}

CLIENT_COMMAND(cmd_join)
{
	if (strcasecmp(args[0], client->network->channel) != 0) {
		client_send_numeric(client, ERR_TOOMANYCHANNELS, args[0]);
		client_restricted(client);
		return;
	}

	server_send(client->server, client->current_cmd_line);
}

void client_commands_init(void)
{
	pool = pool_create("Client commands", 1024, TRUE);
        client_commands = command_list_create(pool);

        client_command_register("QUIT", 0, cmd_self);
        client_command_register("NICK", 1, cmd_self);
        client_command_register("PRIVMSG", 2, cmd_talk);
        client_command_register("NOTICE", 2, cmd_talk);
        client_command_register("AWAY", 0, cmd_self);

	client_command_register("PING", 0, cmd_self);
	client_command_register("PONG", 0, cmd_self);
        client_command_register("WHOIS", 1, cmd_self);
        client_command_register("WHOWAS", 1, cmd_self);
	client_command_register("LUSERS", 0, cmd_self);

	client_command_register("ISON", 1, cmd_self);
	client_command_register("USERHOST", 1, cmd_self);

        client_command_register("JOIN", 1, cmd_join);
        client_command_register("PART", 1, cmd_self);
        client_command_register("MODE", 1, cmd_self);
        client_command_register("TOPIC", 1, cmd_self);
        client_command_register("WHO", 1, cmd_self);
        client_command_register("NAMES", 1, cmd_self);
}

void client_commands_deinit(void)
{
        client_command_unregister("QUIT", cmd_self);
        client_command_unregister("NICK", cmd_self);
        client_command_unregister("PRIVMSG", cmd_talk);
        client_command_unregister("NOTICE", cmd_talk);
        client_command_unregister("AWAY", cmd_self);

	client_command_unregister("PING", cmd_self);
	client_command_unregister("PONG", cmd_self);
        client_command_unregister("WHOIS", cmd_self);
        client_command_unregister("WHOWAS", cmd_self);
	client_command_unregister("LUSERS", cmd_self);

	client_command_unregister("ISON", cmd_self);
	client_command_unregister("USERHOST", cmd_self);

        client_command_unregister("JOIN", cmd_join);
        client_command_unregister("PART", cmd_self);
        client_command_unregister("MODE", cmd_self);
        client_command_unregister("TOPIC", cmd_self);
        client_command_unregister("WHO", cmd_self);
        client_command_unregister("NAMES", cmd_self);

	command_list_destroy(client_commands);
        pool_unref(pool);
}
