#include "common.h"
#include "ioloop.h"
#include "irc.h"

#include "iauth-connection.h"
#include "client-commands.h"

#define MAX_INBUF_SIZE 1024
#define MAX_OUTBUF_SIZE (1024*100)

static Pool pool;
static Client *clients;

static void client_parse_input(Client *client);

Client *client_create(IrcNetwork *network, int handle, IPADDR *ip,
		      const char *nick, const char *user,
		      const char *host, const char *realname)
{
	Client *rec;

        return_val_if_fail(ip != NULL, NULL);

	rec = p_new(pool, Client, 1);
	rec->network = network;
	rec->handle = handle;

	rec->nick = p_strdup(pool, nick);
	rec->user = p_strdup_printf(pool, "%08x", htonl(ip->ip.s_addr));
	rec->realname = p_strdup_printf(pool, "(%s@%s) %s",
					user, host, realname);
	rec->version = p_strdup(pool, "unknown");

	rec->io = io_add(handle, IO_READ, (IOFunc) client_parse_input, rec);

	if (clients != NULL)
                clients->prev = rec;

        rec->next = clients;
	clients = rec;

	server_create(rec);
	return rec;
}

static void client_remove_from_lists(Client *client)
{
	if (client->next != NULL)
		client->next->prev = client->prev;

	if (client->prev != NULL)
		client->prev->next = client->next;
	else
		clients = client->next;
}

int client_destroy(Client *client)
{
	if (!client->disconnected)
		client_disconnect(client, NULL);

	if (client->server != NULL)
		server_destroy(client->server);

	if (client->iauth != NULL) {
		hash_remove(client->iauth->clients,
			    INT_TO_POINTER(client->iauth_uid));
	}

	client_remove_from_lists(client);

	if (client->handle != -1)
		net_disconnect(client->handle);

        if (client->io != NULL)
		io_remove(client->io);

	tbuffer_destroy(client->inbuf);
        tbuffer_destroy(client->outbuf);

        client->prev = client->next = NULL;

        p_free(pool, client->nick);
        p_free(pool, client->user);
        p_free(pool, client->realname);
        p_free(pool, client->version);
	p_free(pool, client);
        return FALSE;
}

static void client_parse_next_line(Client *client)
{
        const char *cmd, *args[COMMAND_MAX_ARGS+1], *id;
        char *line, *prefix;
        int timestamp, argc;

        /* read the line */
	line = (char *) tbuffer_next_line(client->inbuf);
	if (line == NULL || *line == '\0')
		return;

        /* parse the line, prefix and timestamp are ignored */
	line = irc_parse_prefix(line, &prefix, &timestamp, &id);
	if (*line == '\0')
                return;

	cmd = irc_parse_command((char *) t_strdup(line), &argc, args);
	if (cmd == NULL)
                return;

	client->current_cmd_line = line;
	client_command_parse(client, cmd, argc, args);
	client->current_cmd_line = NULL;
}

static void client_parse_input(Client *client)
{
	if (client->disconnected) {
		client_destroy(client);
                return;
	}

	if (client->inbuf == NULL) {
		client->inbuf = tbuffer_create(client->handle, pool,
					       IO_PRIORITY_DEFAULT,
					       MAX_INBUF_SIZE);
	}

	switch (tbuffer_read(client->inbuf)) {
	case -1:
                /* disconnected */
                client_disconnect_errno(client);
		break;
	case -2:
                /* input buffer is full */
		client_disconnect(client, "Excess flood");
		break;
	default:
		client_parse_next_line(client);
	}

        if (client->disconnected)
		client_destroy(client);
}

void client_send(Client *client, const char *data)
{
	unsigned char *buffer;
	int len;

	if (client->disconnected)
                return;

	if (client->outbuf == NULL) {
		client->outbuf = tbuffer_create(client->handle, pool,
						IO_PRIORITY_DEFAULT,
						MAX_OUTBUF_SIZE);
	}

	len = strlen(data);
	buffer = tbuffer_get_space(client->outbuf, len+2);
	if (buffer == NULL) {
		client_disconnect(client, "Output buffer full");
                return;
	}

	memcpy(buffer, data, len);
	buffer[len++] = '\r';
	buffer[len++] = '\n';

	if (tbuffer_send_buffer(client->outbuf, len) < 0)
		client_disconnect_errno(client);
}

void client_login(Client *client, Server *server)
{
	char buf[512];

	client->server = server;

	i_snprintf(buf, sizeof(buf), "NICK %s", client->nick);
	server_send(client->server, buf);

	i_snprintf(buf, sizeof(buf), "USER %s - - :%s",
		   client->user, client->realname);
	server_send(client->server, buf);
}

void client_set_logged_in(Client *client)
{
	char buf[512];

	client->logged_in = TRUE;

	i_snprintf(buf, sizeof(buf), "MODE %s +r", client->nick);
	server_send(client->server, buf);

	i_snprintf(buf, sizeof(buf), "JOIN %s", client->network->channel);
	server_send(client->server, buf);

	/* ask for client version */
	i_snprintf(buf, sizeof(buf), ":%s!%s PRIVMSG %s :\001VERSION\001",
		   client->network->address, client->network->address,
		   client->nick);
	client_send(client, buf);
}

void client_set_nick(Client *client, const char *nick)
{
	p_strdup_replace(pool, &client->nick, nick);
}

void client_set_version(Client *client, const char *version)
{
	p_strdup_replace(pool, &client->version, version);
}

void client_disconnect(Client *client, const char *msg)
{
	if (client->disconnected)
		return;

	client->disconnected = TRUE;
	client_send(client, t_strconcat("ERROR :", msg, NULL));
	server_send(client->server, t_strconcat("QUIT :", msg, NULL));
}

void client_disconnect_errno(Client *client)
{
	client_disconnect(client, strerror(errno));
}

void clients_disconnect(const char *msg)
{
	Client *client, *next;

	for (client = clients; client != NULL; client = next) {
		next = client->next;

		client_disconnect(client, msg);
		client_destroy(client);
	}
}

void clients_init(void)
{
	pool = pool_create("Clients", sizeof(Client)*64, TRUE);
	clients = NULL;
}

void clients_deinit(void)
{
	Client *client, *next;
	int leaks;

        leaks = 0;
	for (client = clients; client != NULL; client = next) {
		next = client->next;

		if (!client->disconnected)
			leaks++;

		client_destroy(client);
	}

	if (leaks > 0)
                i_warning("leaked %d non-disconnected clients", leaks);

	pool_unref(pool);
}
