#include "common.h"
#include "ioloop.h"
#include "fdpass.h"
#include "iauth-connection.h"
#include "version.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#define IAUTH_PATH "../iauth/iauth"
#define IAUTH_BUFFER_SIZE 4096

#define IAUTH_IS_CLOSED(iauth) \
        ((iauth)->cmdin == -1)

void iauth_parse_command(IAuth *iauth, char *line);

static Pool pool;
static IAuth *iauths;
static Timeout iauth_to;

static void iauth_start(IAuth *iauth);
static void iauth_close(IAuth *iauth, int restart);

static void iauth_log(IAuth *iauth, const char *msg, ...)
{
	va_list va;

	va_start(va, msg);
	fprintf(stderr, "iauth(%s): ", iauth->network->name);
	vfprintf(stderr, msg, va);
        fputc('\n', stderr);
        va_end(va);
}

static void iauth_check_restarts(IAuth *iauth)
{
	if (iauth->pid != -1) {
		/* check if iauth has died */
		if (waitpid(iauth->pid, NULL, WNOHANG) == iauth->pid)
			iauth->pid = -1;
	}

	if (iauth->pid == -1)
		iauth_start(iauth);
}

static void iauth_create_client(IAuth *iauth)
{
	IPADDR ip;
	IAuthClientInfo *info;
        Client *client;

        info = &iauth->clientinfo;
	net_host2ip(info->ip, &ip);

	client = client_create(iauth->network, iauth->client_handle, &ip,
			       info->nick, info->user, info->host,
			       info->realname);
	client->iauth = iauth;
	client->iauth_uid = info->id;

        hash_insert(iauth->clients, INT_TO_POINTER(info->id), client);
}

static void iauth_client_error(IAuth *iauth, const char *msg)
{
	const char *client_msg;

	iauth_log(iauth, "%s", msg);

        client_msg = "bug in iauth";
        net_transmit(iauth->client_handle, client_msg, strlen(client_msg));
        net_transmit(iauth->client_handle, "\n", 1);
        net_disconnect(iauth->client_handle);
}

/* create user/server connection */
static void iauth_create_connection(IAuth *iauth)
{
        IAuthClientInfo *info;

        net_set_nonblock(iauth->client_handle, TRUE);

        info = &iauth->clientinfo;
	if (info->user[0] == '\0') {
                iauth_client_error(iauth, "user field can't be empty");
                return;
	}
	if (info->host[0] == '\0') {
                iauth_client_error(iauth, "host field can't be empty");
                return;
	}

	iauth_create_client(iauth);
}

/* I/O: command from iauth */
static void iauth_input(IAuth *iauth)
{
	unsigned char *line;

	switch (tbuffer_read(iauth->cmdinbuf)) {
	case -1:
                /* disconnected */
		iauth_log(iauth, "command input fd closed (%s)",
			  strerror(errno));
		iauth_close(iauth, TRUE);
		break;
	case -2:
		/* input buffer is full - shoudn't happen unless the
		   iauth don't send us any line breaks */
                iauth_log(iauth, "command input buffer full");
		iauth_close(iauth, TRUE);
		break;
	default:
		while ((line = tbuffer_next_line(iauth->cmdinbuf)) != NULL)
			iauth_parse_command(iauth, (char *) line);
	}
}

/* I/O: error from iauth */
static void iauth_error(IAuth *iauth)
{
	unsigned char *line;

	switch (tbuffer_read(iauth->errbuf)) {
	case -1:
                /* disconnected */
                iauth_log(iauth, "error fd closed (%s)", strerror(errno));
		iauth_close(iauth, TRUE);
		break;
	case -2:
		/* input buffer is full - shoudn't happen unless the
		   iauth don't send us any line breaks */
                iauth_log(iauth, "input error buffer full");
		iauth_close(iauth, TRUE);
		break;
	default:
		while ((line = tbuffer_next_line(iauth->errbuf)) != NULL)
			iauth_log(iauth, "%s", (char *) line);
	}
}

/* I/O: fd data from iauth */
static void iauth_data(IAuth *iauth)
{
        int ret;

	if (iauth->data_left == 0) {
                /* read file handle */
		if (fd_read(iauth->data, &iauth->clientinfo,
			    sizeof(iauth->clientinfo), &iauth->client_handle) <
		    sizeof(iauth->clientinfo)) {
                        iauth_close(iauth, TRUE);
			return;
		}

		if (iauth->client_handle == -1) {
			iauth_log(iauth, "fd_read() failed: %s",
				  strerror(errno));
                        iauth_close(iauth, TRUE);
			return;
		}

		iauth->data_left = iauth->clientinfo.extra_input_size;
                if (iauth->data_left == 0)
			iauth_create_connection(iauth);
                return;
	}

	ret = tbuffer_read_max(iauth->databuf, iauth->data_left);
	switch (ret) {
	case -1:
                /* disconnected */
                iauth_log(iauth, "data fd closed (%s)", strerror(errno));
		iauth_close(iauth, TRUE);
		break;
	case -2:
		/* data buffer is full - shoudn't happen unless the
		   iauth is broken */
                iauth_log(iauth, "data buffer full");
		iauth_close(iauth, TRUE);
		break;
	default:
		iauth->data_left -= ret;
		if (iauth->data_left == 0)
			iauth_create_connection(iauth);
	}
}

static void iauth_open(IAuth *iauth)
{
        const char *data;
	char ip_str[MAX_IP_LEN];

        iauth->clients = hash_create(pool, 256, NULL, NULL);

	iauth->io_cmdin =
		io_add(iauth->cmdin, IO_READ, (IOFunc) iauth_input, iauth);
	iauth->io_err =
		io_add(iauth->err, IO_READ, (IOFunc) iauth_error, iauth);
	iauth->io_data =
		io_add(iauth->data, IO_READ, (IOFunc) iauth_data, iauth);

	iauth->cmdinbuf =
		tbuffer_create_file(iauth->cmdin, pool, IAUTH_BUFFER_SIZE);
	iauth->cmdoutbuf =
		tbuffer_create_file(iauth->cmdout, pool, IAUTH_BUFFER_SIZE);
	iauth->errbuf =
		tbuffer_create_file(iauth->err, pool, IAUTH_BUFFER_SIZE);
	iauth->databuf =
		tbuffer_create_file(iauth->data, pool, IAUTH_BUFFER_SIZE);

	if (iauth->network->my_ip.family == 0 ||
	    net_ip2host(&iauth->network->my_ip, ip_str) == -1)
		ip_str[0] = '\0';

	iauth_send(iauth, t_strdup_printf("%c %s", IAUTH_CMD_MODULE_ADD,
                                          iauth->modules));

	/*iauth_send(iauth, t_strdup_printf("%c %s %s %s %s :%s", IAUTH_CMD_MYINFO,
					  "proxy",
					  IRCD_VERSION, ircd_flags, "000",
					  "proxy"));*/

	data = t_strdup_printf("%c %d %s", IAUTH_CMD_LISTEN_ADD,
			       iauth->network->my_port, ip_str);
	iauth_send(iauth, data);
}

static void remove_iauth(void *id, Client *client, IAuth *iauth)
{
	if (client->iauth == iauth)
                client->iauth = NULL;
}

static void iauth_remove_clients(IAuth *iauth)
{
        hash_foreach(iauth->clients, (HashForeachFunc) remove_iauth, iauth);
}

static void iauth_close(IAuth *iauth, int restart)
{
	if (IAUTH_IS_CLOSED(iauth))
                return;

	iauth_remove_clients(iauth);
	hash_destroy(iauth->clients);

        tbuffer_destroy(iauth->cmdinbuf);
        tbuffer_destroy(iauth->cmdoutbuf);
        tbuffer_destroy(iauth->errbuf);
        tbuffer_destroy(iauth->databuf);

        io_remove(iauth->io_cmdin);
        io_remove(iauth->io_err);
        io_remove(iauth->io_data);

        close(iauth->cmdin);
        close(iauth->cmdout);
        close(iauth->err);
        close(iauth->data);
	iauth->cmdin = -1;

	if (restart)
		iauth_check_restarts(iauth);
}

static void iauth_start(IAuth *iauth)
{
	char *iauth_args[2];
        int pipes[3][2], socks[2], i;

        iauth_close(iauth, FALSE);

	/* create pipes for stdin/out/err communication */
	for (i = 0; i < 3; i++) {
		if (pipe(pipes[i]) == -1) {
			i_panic("iauth_start(): pipe() failed: %s",
				  strerror(errno));
		}
	}

        /* create socket for passing fds */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, socks) == -1) {
		i_panic("iauth_start(): socketpair() failed: %s",
			strerror(errno));
	}

	iauth->pid = fork();
	if (iauth->pid == -1)
		i_panic("iauth_start(): fork() failed: %s", strerror(errno));

	if (iauth->pid > 0) {
		iauth->cmdin = pipes[1][0]; close(pipes[1][1]);
		iauth->cmdout = pipes[0][1]; close(pipes[0][0]);
		iauth->err = pipes[2][0]; close(pipes[2][1]);
		iauth->data = socks[0]; close(socks[1]);
                iauth_open(iauth);
		return;
	}

	/* set stdin/out/err */
	close(0); close(1); close(2);
	dup2(pipes[0][0], 0); close(pipes[0][1]);
	dup2(pipes[1][1], 1); close(pipes[1][0]);
	dup2(pipes[2][1], 2); close(pipes[2][0]);

	/* set data socket */
	dup2(socks[1], 3); close(socks[0]);

	/* execute iauth */
        iauth_args[0] = IAUTH_PATH;
        iauth_args[1] = NULL;
	execvp(iauth_args[0], iauth_args);

	fprintf(stderr, "Error executing iauth (%s): %s\n",
		iauth_args[0], strerror(errno));
        _exit(99);
}

IAuth *iauth_create(IrcNetwork *network, const char *modules)
{
	IAuth *rec;

	rec = p_new(pool, IAuth, 1);
	rec->network = network;
        rec->modules = p_strdup(pool, modules);

	rec->pid = -1;
	rec->cmdin = -1;

        rec->next = iauths;
        iauths = rec;

	iauth_start(rec);
	return rec;
}

void iauth_destroy(IAuth *iauth)
{
	IAuth **tmp;

	for (tmp = &iauths; *tmp != NULL; tmp = &(*tmp)->next) {
		if (*tmp == iauth) {
			*tmp = iauth->next;
                        break;
		}
	}

	if (iauth->pid != -1) {
		iauth_send(iauth, "X");

		/*FIXME: sleep(1);
		kill(iauth->pid, SIGKILL);
                waitpid(iauth->pid, NULL, WNOHANG);*/
	}

	p_free(pool, iauth->modules);
	iauth_close(iauth, FALSE);
        p_free(pool, iauth);
}

void iauth_send(IAuth *iauth, const char *data)
{
        unsigned char *buffer;
	int len;

	if (IAUTH_IS_CLOSED(iauth))
		return;

	len = strlen(data);
	buffer = tbuffer_get_space(iauth->cmdoutbuf, len+1);
	if (buffer == NULL) {
                iauth_log(iauth, "command output buffer full");
		iauth_close(iauth, TRUE);
                return;
	}

	memcpy(buffer, data, len);
        buffer[len] = '\n';

	if (tbuffer_send_buffer(iauth->cmdoutbuf, len+1) < 0) {
                /* disconnected */
		iauth_log(iauth, "command output fd closed (%s)",
			  strerror(errno));
		iauth_close(iauth, TRUE);
	}
}

void iauth_connect_server(const char *address, int port)
{
	if (iauths == NULL)
		return;

	iauth_send(iauths, t_strdup_printf("%c %s %d", IAUTH_CMD_CONNECT_SERVER,
					   address, port));
}

static void iauth_timeout(void)
{
	IAuth *iauth;

	for (iauth = iauths; iauth != NULL; iauth = iauth->next)
                iauth_check_restarts(iauth);
}

void iauth_init(void)
{
	pool = pool_create("IAuth", 128 + IAUTH_BUFFER_SIZE*4, TRUE);

	iauths = NULL;
        iauth_to = timeout_add(10000, (TimeoutFunc) iauth_timeout, NULL);
}

void iauth_deinit(void)
{
	while (iauths != NULL) {
		IAuth *iauth = iauths;
		iauths = iauths->next;

                iauth_destroy(iauth);
	}
	timeout_remove(iauth_to);
        pool_unref(pool);
}
