#include "common.h"
#include "ioloop.h"
#include "network.h"

#include "irc.h"
#include "irc-limits.h"

#define MAX_INBUF_SIZE (1024*100)
#define MAX_OUTBUF_SIZE 1024

static Pool pool;

static int server_parse_next_line(Server *server)
{
	const char *line, *cmd, *args[COMMAND_MAX_ARGS+1], *id;
	const char *data;
        char *prefix, *p;
        int timestamp, argc;

	line = tbuffer_next_line(server->inbuf);
	if (line == NULL || *line == '\0')
		return FALSE;

	data = irc_parse_prefix((char *) t_strdup(line),
				&prefix, &timestamp, &id);
	if (*data == '\0')
                return TRUE;

	cmd = irc_parse_command((char *) t_strdup(data), &argc, args);
	if (cmd == NULL)
                return TRUE;

	if (argc > 0 && strcasecmp(cmd, "NICK") == 0)
		client_set_nick(server->client, args[1]);
	else if (strcasecmp(cmd, "001") == 0)
                client_set_logged_in(server->client);
	else if (argc > 1 && strcasecmp(cmd, "PRIVMSG") == 0 &&
		 strcasecmp(args[0], server->client->nick) == 0 &&
		 strcasecmp(args[1], "\001VERSION\001") == 0) {
		/* someone ctcp versioned us - reply */
		p = strchr(prefix, '!');
		if (p != NULL) {
			*p = '\0';
			data = t_strdup_printf("NOTICE %s :\001VERSION %s\001",
					       prefix, server->client->version);
			server_send(server, data);
			return TRUE;
		}
	}

	client_send(server->client, line);
        return TRUE;
}

static void server_parse_input(Server *server)
{
	if (server->inbuf == NULL) {
		server->inbuf = tbuffer_create(server->handle, pool,
					       IO_PRIORITY_DEFAULT,
					       MAX_INBUF_SIZE);
	}

	switch (tbuffer_read(server->inbuf)) {
	case -1:
                /* disconnected */
                client_disconnect_errno(server->client);
		break;
	case -2:
                /* input buffer is full */
		client_disconnect(server->client, "Input buffer full");
		break;
	default:
		while (server_parse_next_line(server)) ;
	}
}

static void server_connected(Server *server)
{
	io_remove(server->io);
	server->io = NULL;

	server->io = io_add(server->handle, IO_READ,
			    (IOFunc) server_parse_input, server);

	client_login(server->client, server);
}

Server *server_create(Client *client)
{
	Server *rec;

        return_val_if_fail(client != NULL, NULL);

	rec = p_new(pool, Server, 1);
	rec->client = client;
	rec->handle = net_connect_ip(&client->network->ip,
				     client->network->port,
				     client->network->my_ip.family == 0 ? NULL :
				     &client->network->my_ip);
	rec->io = io_add(rec->handle, IO_READ | IO_WRITE,
			 (IOFunc) server_connected, rec);
	return rec;
}

void server_destroy(Server *server)
{
	if (server->handle != -1)
		net_disconnect(server->handle);

        if (server->io != NULL)
		io_remove(server->io);

	tbuffer_destroy(server->inbuf);
        tbuffer_destroy(server->outbuf);

	p_free(pool, server);
}

void server_send(Server *server, const char *data)
{
	unsigned char *buffer;
	int len;

	if (server->outbuf == NULL) {
		server->outbuf = tbuffer_create(server->handle, pool,
						IO_PRIORITY_DEFAULT,
						MAX_OUTBUF_SIZE);
	}

	len = strlen(data);
	buffer = tbuffer_get_space(server->outbuf, len+2);
	if (buffer == NULL) {
		client_disconnect(server->client, "Excess flood");
                return;
	}

	memcpy(buffer, data, len);
	buffer[len++] = '\r';
	buffer[len++] = '\n';

	if (tbuffer_send_buffer(server->outbuf, len) < 0)
		client_disconnect_errno(server->client);
}

void servers_init(void)
{
	pool = pool_create("Servers", 10240, TRUE);
}

void servers_deinit(void)
{
	pool_unref(pool);
}
